    //
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include "AutomaticHeapScope.hpp"
#include "DataStore.hpp"
#include "RockallNew.hpp"
#include "RockallDelete.hpp"
#include "SingleThreadedHeap.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
	//
	//  The 'DATA_STORE' class constructor.
	//
	//  Create a new database to manage a state table.
	//
	DATA_STORE::DATA_STORE( void )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope;

		//
		//  Create an instance of a database.
		//
		Active = (ROCKALL_NEW<DATABASE_HEAP>::New( & Scope,& Database ));
		Open = false;
		}

	//
	//  Create Data Store.
	//
	//  Create a new state table in the database.
	//
	bool DATA_STORE::CreateDataStore( char *DatabaseName,STATE_TABLE *StateTable )
		{
		//
		//  We need to make sure we were able to create the 
		//  an instance of the database but have not opened
		//  any database file.
		//
		if ( (Active) && (! Open) )
			{
			//
			//  Create a new database file.
			//
			if ( Database -> CreateFile( DatabaseName ) )
				{
				int MaxNextStates = (StateTable -> MaxNextStates * sizeof(NEXT_STATE));
				int MaxStates = (StateTable -> MaxStates * sizeof(STATE));
				int MaxWordSize = StateTable -> MaxWordSize;
				NEXT_STATE **NextStates = & StateTable -> NextStates;
				STATE **States = & StateTable -> States;

				//
				//  Create a new transaction.
				//
				if ( Database -> BeginTransaction() )
					{
					FILE_ADDRESS FileAddress1;
					FILE_ADDRESS FileAddress2;

					//
					//  Transactionally allocate space in the 
					//  database for the new state table.
					//
					if
							(
							(Database -> New( & FileAddress1,((void**) NextStates),MaxNextStates ))
								&&
							(Database -> New( & FileAddress2,((void**) States),MaxStates ))
								&&
							(Database -> SetUserValue( 0,MagicNumber ))
								&&
							(Database -> SetUserValue( 1,MaxWordSize ))
								&&
							(Database -> SetUserValue( 2,FileAddress1 ))
								&&
							(Database -> SetUserValue( 3,FileAddress2 ))
							)
						{ return (Open = true); }

					//
					//  We will 'Abort' if we have any problems.
					//
					Database -> EndTransaction( true );
					}

				//
				//  We close the file if we have any problems.
				//
				Database -> CloseFile(); 
				}
			}

		return false;
		}

	//
	//  Load Data Store.
	//
	//  Load a state table from the database.
	//
	bool DATA_STORE::LoadDataStore( char *DatabaseName,STATE_TABLE *StateTable )
		{
		//
		//  We need to make sure we were able to create the 
		//  an instance of the database but have not opened
		//  any database file.
		//
		if ( (Active) && (! Open) )
			{
			//
			//  Open an existing database file.
			//
			if ( Database -> OpenFile( DatabaseName,true ) )
				{
				FILE_ADDRESS MagicValue;
				FILE_ADDRESS MaxWordSize;
				FILE_ADDRESS FileAddress1;
				FILE_ADDRESS FileAddress2;

				//
				//  We load the user values and verify the magic
				//  value to ensure it is the correct type of
				//  database.
				//
				if
						(
						(Database -> GetUserValue( 0,& MagicValue ))
							&&
						(MagicNumber == MagicValue)
							&&
						(Database -> GetUserValue( 1,& MaxWordSize ))
							&&
						(Database -> GetUserValue( 2,& FileAddress1 ))
							&&
						(Database -> GetUserValue( 3,& FileAddress2 ))
							&&
						(Database -> BeginTransaction())
						)
					{
					void *NextStates;
					void *States;

					//
					//  We now load both parts of the 'StateTable'.
					//
					if
							(
							(Database -> View( FileAddress1,& NextStates ))
								&&
							(Database -> View( FileAddress2,& States ))
							)
						{
						//
						//  Now we have everything we need we can
						//  update the 'StateTable'.
						//
						StateTable -> MaxNextStates = 0;
						StateTable -> MaxStates = 0;
						StateTable -> MaxWordSize = ((int) MaxWordSize);
						StateTable -> NextStates = ((NEXT_STATE*) NextStates);
						StateTable -> States = ((STATE*) States);

						return (Open = true); 
						}

					//
					//  We will 'Abort' if we have any problems.
					//
					Database -> EndTransaction( true );
					}

				//
				//  We close the file if we have any problems.
				//
				Database -> CloseFile(); 
				}
			}

		return false;
		}

	//
	//  Close Data Store.
	//
	//  Close the data store when it's no longer needed.
	//
	bool DATA_STORE::CloseDataStore( void )
		{
		//
		//  We need to make sure we were able to create the 
		//  an instance of the database and opened a file.
		//
		if ( (Active) && (Open) )
			{
			//
			//  Mark the file as closed.
			//
			Open = false;

			//
			//  End the transaction and close the file.
			//
			Database -> EndTransaction();

			Database -> CloseFile(); 

			return true;
			}

		return false;
		}

	//
	//  The 'DATA_STORE' class destructor.
	//
	//  Delete an instance of the database.
	//
	DATA_STORE::~DATA_STORE( void )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope;

		//
		//  Close any open file.
		//
		if ( Open )
			{ CloseDataStore(); }

		//
		//  Delete any instance of a database.
		//
		if ( Active )
			{ ROCKALL_DELETE<DATABASE_HEAP>::Delete( & Scope,Database ); }

		//
		//  Zero the class members.
		//
		Database = NULL;

		Open = false;
		Active = false;
		}
	}