    //
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include <ctype.h>
#include <stdio.h>

#include "AutomaticHeapScope.hpp"
#include "Dictionary.hpp"
#include "SingleThreadedHeap.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
	//
	//  The 'DICTIONARY' class constructor.
	//
	//  Create a new instance of the dictionary.
	//
	DICTIONARY::DICTIONARY( void )
		{ Active = false; }

	//
	//  Create Dictionary.
	//
	//  Create a new dictionary in the database.
	//
	bool DICTIONARY::CreateDictionary( char *DatabaseName,char *FileName,STATE_TABLE *StateTable )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope;

		if ( ! Active )
			{
			FILE *File = fopen( FileName,"rb" );

			//
			//  Let's see if we were able to open the new text
			//  version of the dictionary.
			//
			if ( File != NULL )
				{
				CONTEXT Context = { 0 };
				COUNTERS Counters = { MinStates };
				STATES States;

				//
				//  We process all the characters in the file
				//  until we reach 'EOF'.
				//
				while ( Context.Character != EOF )
					{
					//
					//  We read the next character.
					//
					Context.Character = fgetc( File );

					//
					//  We convert any upper case characters
					//  to lower case.
					//
					if ( isupper( Context.Character ) )
						{ Context.Character = tolower( Context.Character ); }

					//
					//  We process each character as necessary.
					//
					switch ( Context.Character )
						{
						case 'a': case 'b': case 'c': case 'd':
						case 'e': case 'f': case 'g': case 'h':
						case 'i': case 'j': case 'k': case 'l':
						case 'm': case 'n': case 'o': case 'p':
						case 'q': case 'r': case 's': case 't':
						case 'u': case 'v': case 'w': case 'x':
						case 'y': case 'z':
							{
							KEY *Key = ((KEY*) & Context.NewState);

							//
							//  We update 'NewState' with the current
							//  state and the current character.  This
							//  will be the 'Key' for the 'States' tree.
							//
							Context.NewState.Character = Context.Character;
							Context.NewState.State = Context.NextState.State;

							//
							//  Let's see if we already have this 'Character'
							//  in this state.
							//
							if ( States.FindKey( & Scope,(*Key),& Context.NextState ) )
								{
								//
								//  We do have this 'Character' in the
								//  current state.
								//  
								if ( Context.NextState.State < MinStates )
									{
									//
									//  We sometimes delete 'States' trying
									//  to save space.  We now have to
									//  undo this as we need the state in
									//  this case.
									//
									Context.NextState.State = Counters.MaxStates ++;

									if ( ! States.UpdateKey( & Scope,(*Key),Context.NextState ) )
										{ fclose( File ); return false; }
									}
								}
							else
								{
								//
								//  We do *not* have this 'Character'
								//  in the current state.
								//  
								Context.NextState.State = Counters.MaxStates ++;
								Context.NextState.Character = Context.Character;
								Context.NextState.Score = 0;

								//
								//  We create a new state in the tree.
								//
								if ( ! States.NewKey( & Scope,(*Key),Context.NextState ) )
									{ fclose( File ); return false; }
								}

							//
							//  We count the number of characters in
							//  the current word.
							//
							Context.Length ++;

							break;
							}

						case '\r':
						case '\n':
						case EOF:
							{
							CONTEXT NewContext = { Context.Character };

							//
							//  We do not need to do anything for zero
							//  length words.
							//
							if ( Context.Length > 0 )
								{
								KEY *Key = ((KEY*) & Context.NewState);

								//
								//  We try to save space by deleting
								//  empty states at the end of the table.
								//
								if ( Context.NextState.State == (Counters.MaxStates-1) )
									{
									Context.NextState.State = 0;
									Counters.MaxStates --;
									}

								//
								//  We compute the 'Score' for the word
								//  based on its 'Length' according to
								//  the rules of the game.
								//
								switch ( Context.Length )
									{
									case 1: case 2:
										{
										Context.NextState.Score = 0;

										break;
										}

									case 3: case 4: case 5: case 6: case 7:
										{
										Context.NextState.Score = (Context.Length - 2);

										break;
										}

									default:
										{
										Context.NextState.Score = 11;

										break;
										}
									}

								//
								//  We now save the final adjustments
								//  to the current word.
								//
								if ( States.UpdateKey( & Scope,(*Key),Context.NextState ) )
									{
									//
									//  Update the 'MaxWords' and
									//  'MaxWordSize' counts.
									//
									if ( Context.Length > Counters.MaxWordSize )
										{ Counters.MaxWordSize = Context.Length; }

									Counters.MaxWords ++;
									}
								else
									{ fclose( File ); return false; }
								}

							//
							//  We can now reset the 'Context'
							//  ready for the next word.
							//
							Context = NewContext;

							break;
							}

						default:
							{
							printf
								( 
								"Warning: Skipping non-alphabetic character '%c'.\n",
								Context.Character 
								);

							break;
							}
						}
					}

				//
				//  We have now finished processing the text version
				//  of the dictionary so close the text file.
				//
				fclose( File );

				//
				//  Let's ensure that the text version of the
				//  dictionary was not empty.
				//
				if ( Counters.MaxStates > MinStates )
					{
					STATE_TABLE NewStateTable = { 0,Counters.MaxStates,Counters.MaxWordSize };
					FILE_SIZE MaxNextStates;

					//
					//  We now compute the size of the 'StateTable'
					//  and create space for it in the database.
					//
					if 
							(
							(States.Size( & MaxNextStates ))
								&&
							((NewStateTable.MaxNextStates = ((int) MaxNextStates)), true)
								&&
							(DataStore.CreateDataStore( DatabaseName,& NewStateTable )) 
							)
						{
						COPY_DATA CopyData = { 0,0,NewStateTable.NextStates,NewStateTable.States };

						//
						//  Walk all the 'States' and copy them
						//  into the database.
						//
						if 
								( 
								(States.Walk( & Scope,CopyStatesIntoDatabase,& CopyData ))
									&&
								(CopyData.MaxNextStates == NewStateTable.MaxNextStates)
									&&
								((CopyData.MaxStates+1) == NewStateTable.MaxStates)
								)
							{
							NEXT_STATE (*View1)[ ViewSize ] = ((NEXT_STATE(*)[ ViewSize ]) NewStateTable.NextStates);
							STATE (*View2)[ ViewSize ] = ((STATE(*)[ ViewSize ]) NewStateTable.States);

							//
							//  Now we have everything we need
							//  we can update the 'StateTable'.
							//
							(*StateTable) = NewStateTable;

							return (Active = true); 
							}
						}
					}
				}
			}

		return false;
		}

	//
	//  Copy States Into Database.
	//
	//  Copy all the states into the database.
	//
	bool DICTIONARY::CopyStatesIntoDatabase( KEY & Key,NEXT_STATE & NextState,void *UserData )
		{
		COPY_DATA *CopyData = ((COPY_DATA*) UserData);
		NEW_STATE NewState = (*((NEW_STATE*) & Key));

		//
		//  We only need to update 'Start' and 'MaxStates'
		//  when the 'State' value changes.  
		if
				(
				(NewState.State > CopyData -> MaxStates)
					||
				(CopyData -> MaxNextStates == 0)
				)
			{
			CopyData -> States[ NewState.State ].Start = CopyData -> MaxNextStates;

			CopyData -> MaxStates = NewState.State;
			}

		//
		//  Copy across the 'NextState' and update 'End' to
		//  include it in the current 'State'. 
		//
		CopyData -> NextStates[ (CopyData -> MaxNextStates ++) ] = NextState;

		CopyData -> States[ CopyData -> MaxStates ].End = CopyData -> MaxNextStates;

		return true;
		}

	//
	//  Load Dictionary.
	//
	//  Load the dictionary from the database.
	//
	bool DICTIONARY::LoadDictionary( char *DatabaseName,STATE_TABLE *StateTable )
		{
		//
		//  Let's ensure that we are not 'Active'.
		//
		if ( ! Active )
			{ return (Active = (DataStore.LoadDataStore( DatabaseName,StateTable ))); }

		return false;
		}

	//
	//  Close State Table.
	//
	//  Close the dictionary when it's no longer needed.
	//
	bool DICTIONARY::CloseDictionary( void )
		{
		//
		//  Close any 'Active' dictionary.
		//
		if ( Active )
			{
			Active = false;

			return (DataStore.CloseDataStore());
			}

		return false;
		}

	//
	//  The 'DICTIONARY' class destructor.
	//
	//  Delete an instance of the dictionary.
	//
	DICTIONARY::~DICTIONARY( void )
		{
		//
		//  Close any 'Active' state table.
		//
		if ( Active )
			{ CloseDictionary(); }

		//
		//  Zero the class members.
		//
		Active = false;
		}
	}