	//
	//  Header files.
	//
	//  The header files provide access to functionality in the
	//  Rockall-DB library and other classes.
	//
#include <stdio.h>
#include <memory.h>

#include "AsynchronousCallbacks.hpp"
#include "FindWords.hpp"

	//
	//  The 'BOOGLE' namespace contains all the related code.
	//
namespace BOGGLE
	{
	//
	//  The 'FIND_WORDS' class constructor.
	//
	//  Create a new instance of the class to find words.
	//
	FIND_WORDS::FIND_WORDS( void )
		{
		STATE_TABLE Zeros = { 0 };

		//
		//  Setup class members.
		//
		Active = false;

		StateTable = Zeros;
		}

	//
	//  Create Find Words.
	//
	//  Create a new dictionary in the database.
	//
	bool FIND_WORDS::CreateFindWords( char *DatabaseName,char *FileName )
		{
		//
		//  Let's ensure that we are not 'Active'.
		//
		if ( ! Active )
			{ return (Active = (Dictionary.CreateDictionary( DatabaseName,FileName,& StateTable ))); }

		return false;
		}

	//
	//  Load Find Words.
	//
	//  Load the dictionary from the database.
	//
	bool FIND_WORDS::LoadFindWords( char *DatabaseName )
		{
		//
		//  Let's ensure that we are not 'Active'.
		//
		if ( ! Active )
			{ return (Active = (Dictionary.LoadDictionary( DatabaseName,& StateTable ))); }

		return false;
		}

	//
	//  Find All Words.
	//
	//  Find all the words on the 'Board'.
	//
	bool FIND_WORDS::FindAllWords( char *Board,int Height,int Width,FUNCTION Function,void *UserData )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope;
		bool Processing = true;
		bool Result = true;
		UNIQUE Unique;

		//
		//  We need to be sure that 'Text' is large enough for
		//  the largest word in the dictionary.
		//
		if ( StateTable.MaxWordSize < (MaxTextSize-1) )
			{
			ASYNCHRONOUS_CALLBACKS<bool*> Timer;

			//
			//  We now 'Schedule()' the 'Timer' thread.
			//
			if ( Timer.Schedule( FindTimer,& Processing ) )
				{
				ASYNCHRONOUS_CALLBACKS<CONTEXT> Worker;
				int MaxThreads = (Worker.NumberOfCPUs() * 2);
				int MaxDimension = ((Height > Width) ? Height : Width);
				int MaxCallbacks = ((MaxDimension < MaxThreads) ? MaxDimension : MaxThreads);
				CONTEXT Context = { Board,this,Height,MaxThreads,& Processing,& Result,& Scope,0,"",0,& Unique,Width };
				int Count;

				//
				//  We will 'Schedule()' a number of 'Worker' threads.
				//
				for ( Count=0;Count < MaxCallbacks;Count ++ )
					{
					//
					//  We update the 'Context' with the 'ThreadID'.
					//
					Context.ThreadID = Count;

					//
					//  We now 'Schedule()' each 'Worker' thread.
					//
					if ( ! Worker.Schedule( FindAllWordsInParallel,Context ) )
						{ Result = false; }
					}

				//
				//  We wait for the 'Worker' threads to complete.
				//
				Worker.WaitAll();

				Processing = false;
				}
			else
				{ Result = false; }
			}
		else
			{ Result = false; }

		//
		//  We should have found all the words so now walk
		//  all the 'Unique' words.
		//
		if ( Result )
			{ Result = (Unique.Walk( & Scope,Function,UserData )); }

		return Result;
		}	

	//
	//  Find All Words In Parallel.
	//
	//  Find all words in parallel by searching multiple columns
	//  or rows at the same time.
	//
	void FIND_WORDS::FindAllWordsInParallel( CONTEXT *Context )
		{
		AUTOMATIC_HEAP_SCOPE<MULTI_THREADED_HEAP> Scope( Context -> Scope );
		int MaxSize = (Context -> Height * Context -> Width);
		char *NewBoard;

		//
		//  We need to make a copy of the 'Board' for each thread
		//  as we sometimes change letters while we are looking
		//  for new words. 
		//
		if ( Scope.New( ((void**) & NewBoard),MaxSize ) )
			{
			//
			//  We make a new copy of the 'Board'.
			//
			memcpy( NewBoard,Context -> Board,MaxSize );

			Context -> Board = NewBoard;

			//
			//  We make the inner loop operate on the longest
			//  dimension of the 'Board'.
			//
			if ( Context -> Height >= Context -> Width )
				{
				int Count1;

				//
				//  We search all the 'X' values in parallel in
				//  strides of 'MaxThreads'.
				//
				for 
						( 
						Count1=Context -> ThreadID;
						(Count1 < Context -> Width) && (*Context -> Processing) && (*Context -> Result);
						Count1 += Context -> MaxThreads 
						)
					{
					int Count2;

					//
					//  We search all the 'Y' values.
					//
					for 
							( 
							Count2=0;
							(Count2 < Context -> Height) && (*Context -> Processing) && (*Context -> Result);
							Count2 ++ 
							)
						{ 
						if ( ! Context -> Class -> FindSingle( Context,0,Count1,Count2 ) )
							{ (*Context -> Result) = False; }
						}
					}
				}
			else
				{
				int Count1;

				//
				//  We search all the 'Y' values in parallel in
				//  strides of 'MaxThreads'.
				//
				for 
						( 
						Count1=Context -> ThreadID;
						(Count1 < Context -> Height) && (*Context -> Processing) && (*Context -> Result);
						Count1 += Context -> MaxThreads 
						)
					{
					int Count2;

					//
					//  We search all the 'X' values.
					//
					for 
							( 
							Count2=0;
							(Count2 < Context -> Width) && (*Context -> Processing) && (*Context -> Result);
							Count2 ++ 
							)
						{ 
						if ( ! Context -> Class -> FindSingle( Context,0,Count2,Count1 ) )
							{ (*Context -> Result) = False; }
						}
					}
				}

			//
			//  We delete to copy of the 'Board'.
			//
			Scope.Delete( NewBoard );
			}
		else
			{ (*Context -> Result) = False; }
		}

	//
	//  Find Nearby.
	//
	//  Find all the nearby letters and see if any of them work
	//  with the current 'State'.
	//
	bool FIND_WORDS::FindNearby( CONTEXT *Context,int State,int X,int Y )
		{
		bool Result = true;
		int Count1;

		//
		//  We search the nearby 'X' values.
		//
		for ( Count1=(X-1);(Result) && (Count1 <= (X+1));Count1 ++ )
			{
			int Count2;

			//
			//  We search the nearby 'Y' values.
			//
			for ( Count2=(Y-1);(Result) && (Count2 <= (Y+1));Count2 ++ )
				{
				//
				//  We ensure that the new 'X' and 'Y' values
				//  are a valid location on the 'Board'.
				//
				if
						(
						((Count1 >= 0) && (Count1 < Context -> Width))
							&&
						((Count2 >= 0) && (Count2 < Context -> Height))
						)
					{ Result = FindSingle( Context,State,Count1,Count2 ); }
				}
			}

		return Result;
		}

	//
	//  Find Single.
	//
	//  Find a single letter that might be part of a word.
	//
	bool FIND_WORDS::FindSingle( CONTEXT *Context,int State,int X,int Y )
		{
		char *Current1 = & Context -> Board[ (Y * Context -> Width) + X ];
		STATE *Current2 = & StateTable.States[ State ];
		bool Result = true;
		int Count;

		//
		//  We search for the current letter in the 'StateTable'
		//  to see if it is part of a word.  You would think that
		//  a binary search would be faster but tests showed it
		//  was not in this case.  Quite surprising.  It is probably
		//  because most of the states are very small.
		//
		for ( Count=Current2 -> Start;(Result) && (Count < Current2 -> End);Count ++ )
			{
			NEXT_STATE *Current3 = & StateTable.NextStates[ Count ];

			//
			//  We look at each alternative one at a time.
			//
			if ( (*Current1) == Current3 -> Character )
				{
				int NewState = Current3 -> State;

				//
				//  We have found a suitable alternative so add it
				//  to the current 'Text' string.
				//
				Context -> Text[ (Context -> Size ++) ] = (*Current1);
				Context -> Text[ Context -> Size ] = '\0';

				//
				//  We must simulate a 'u' every time we see a 'q'
				//  so we replace 'q' with 'u' and do a recursive
				//  call later.  If not, we replace the letter with
				//  a '*' as we know that this can't match anything
				//  in the dictionary.
				//
				if ( Current3 -> Character == 'q' )
					{ (*Current1) = 'u'; }
				else
					{ (*Current1) = '*'; }

				//
				//  We will record the current word if it has a
				//  'Score'.
				//
				if ( Current3 -> Score > 0 )
					{
					STRING Key = Context -> Text;
					short Skip;

					//
					//  Let's see if we have alreday found the 
					//  new 'Key' before we try to add it.
					//  We use 'FindKey()' as it only needs a
					//  shareable lock whereas 'NewKey()' needs 
					//  an exclusive lock.
					//
					if ( ! Context -> Unique -> FindKey( Key,& Skip ) )
						{
						//
						//  We create a new 'Key' for every word
						//  and put it in 'Unique'.  This call
						//  will fail if the 'Key' alreday exists.
						//
						Context -> Unique -> NewKey( Key,Current3 -> Score );
						}
#ifdef ROCKALL_DB_EXAMPLES

					//
					//  Nasty Hack: We should never modify the 
					//  'StateTable' as it's supposed to be read
					//  only.  However, we zero 'Score' below as
					//  it stops us finding the same word again.
					//  Multiple threads still work because any
					//  duplicates or race conditions are resolved
					//  by 'Unique'.
					//  
					Current3 -> Score = 0;
#endif
					}

				//
				//  The 'NextState' will be zero if we have already
				//  found the longest word in this series.
				//
				if ( NewState != 0 )
					{
					//
					//  We do a recursive call if we are processing
					//  a 'q' to deal with the following 'u'.  If not,
					//  we examine the nearby letters.
					//
					if ( Current3 -> Character == 'q' )
						{ Result = (FindSingle( Context,NewState,X,Y )); }
					else
						{ Result = (FindNearby( Context,NewState,X,Y )); } 
					}

				//
				//  We have finished with this letter so remove it
				//  from the current 'Text' string.
				//
				(*Current1) = Context -> Text[ (-- Context -> Size) ];

				return Result;
				}
			}

		return Result;
		}

	//
	//  Find Timer.
	//
	//  The find timer resets the processing flag when its
	//  time to stop finding new words.
	//
	void FIND_WORDS::FindTimer( bool **Processing )
		{
		ASYNCHRONOUS_CALLBACKS<bool> Sleeping;
		int Count;

		//
		//  We 'Sleep()' until we finish or timeout.
		//
		for ( Count=0;(Count < MaxFindTime) && (**Processing);Count ++ )
			{ Sleeping.Sleep( 1000 ); }

		(**Processing) = false;
		}

	//
	//  Close Find Words.
	//
	//  Close the dictionary when it's no longer needed.
	//
	bool FIND_WORDS::CloseFindWords( void )
		{
		//
		//  Close any 'Active' dictionary.
		//
		if ( Active )
			{
			STATE_TABLE Zeros = { 0 };

			//
			//  Zero class members.
			//
			Active = false;

			StateTable = Zeros;

			//
			//  Close the dictionary.
			//
			return (Dictionary.CloseDictionary());
			}

		return false;
		}

	//
	//  The 'FIND_WORDS' class destructor.
	//
	//  Delete an instance of the class.
	//
	FIND_WORDS::~FIND_WORDS( void )
		{
		STATE_TABLE Zeros = { 0 };

		//
		//  Close any 'Active' dictionary.
		//
		if ( Active )
			{ CloseFindWords(); }

		//
		//  Zero class members.
		//
		StateTable = Zeros;

		Active = false;
		}
	}
