    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.
    //

#include <stdio.h>

#include "AutomaticHeapScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Database.hpp"
#include "RockallNew.hpp"
#include "TextFile.hpp"
#include "Words.hpp"

    //
    //  Database (Constructor).
    //
DATABASE::DATABASE( void )
    { Active = False; }

    //
    //  Create.
    //
    //  Create and build a new database.
    //
BOOLEAN DATABASE::Create( char *FileName1,char *FileName2 )
    {
    AUTO FILE_SIZE FileSize;
    AUTO TEXT_FILE TextFile;

    //
    //  Open the text file.
    //
    if ( (! Active) && (TextFile.Open( FileName2,& FileSize )) )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database( & Heap );

        //
        //  Create the database.
        //
        if ( Database.CreateFile( FileName1 ) )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database;

            //
            //  Start a transaction.
            //
            if ( Database.BeginTransaction() )
                {
                AUTO FILE_ADDRESS RootAddress;
                ROOT *Root;
                char *Text;

                //
                //  Create space for the book and read it into
                //  the database.
                //
                if
                        (
                        (ROCKALL_NEW<ROOT>::New( & Database,& RootAddress,& Root ))
                            &&
                        (ROCKALL_NEW<char>::New( & Database,& Root -> Book.Text,& Text,(FileSize+1) ))
                            &&
                        (TextFile.Read( Text,FileSize ))
                            &&
                        (Database.SetUserValue( 0,RootAddress ))
                        )
                    {
                    AUTO CONTEXT Context = { this,NULL,0,0,Root,0,0 };
                    AUTO WORDS Words;

                    //
                    //  Update related data.
                    //
                    Root -> Book.Size = FileSize;
                    Text[ FileSize ] = NULL;

                    //
                    //  Analyse the entire book and build an index of
                    //  all the words in it.
                    //
                    if ( Words.Analyse( & Context,NewParagraph,NewWord,FileSize,Text ) )
                        {
                        //
                        //  Output the totals.
                        //
                        printf
                            (
                            "Summary: %d words (%d unique and %d refs).\n",
                            ((int) Context.TotalWords),
                            ((int) Context.UniqueWords),
                            ((int) Context.References)
                            );

                        //
                        //  Close the text file.
                        //
                        TextFile.Close();

                        return (Active = True);
                        }
                    }

                //
                //  Abort the transaction.
                //
                Database.EndTransaction( True );
                }

            //
            //  Close the database.
            //
            Database.CloseFile();
            }

        //
        //  Close the text file.
        //
        TextFile.Close();
        }

    return False;
    }

    //
    //  Open.
    //
    //  Open an existing database.
    //
BOOLEAN DATABASE::Open( char *FileName )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database( & Heap );

    //
    //  Open the database.
    //
    return (Active = Database.OpenFile( FileName ));
    }

    //
    //  Find.
    //
    //  Find all the related references in the database.
    //
BOOLEAN DATABASE::Find( DISPLAY_REFERENCE Display,int Size,char *Words[] )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database( & Heap );

    if ( (Active) && (Database.BeginTransaction()) )
        {
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope1( & Database );
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope2( & Database );
        AUTO FILE_ADDRESS RootAddress;
        AUTO ROOT *Root;
        AUTO char *Text;

        //
        //  Find the 'ROOT' structure and then create a view on
        //  it and the text of the book.
        //
        if
                (
                (Database.GetUserValue( 0,& RootAddress ))
                    &&
                (RootAddress != NULL)
                    &&
                (AutomaticViewScope1.View( RootAddress,((VOID**) & Root) ))
                    &&
                (AutomaticViewScope2.View( Root -> Book.Text,((VOID**) & Text) ))
                )
            {
            AUTO BOOLEAN Result = True;
            AUTO SETS Set( & Database );
            AUTO FILE_SIZE ResultSize;
            AUTO FILE_SIZE Count;

            //
            //  We look up each search word in the index and add
            //  the associated references into the result set.
            //
            for ( Count=0;(Result) && (Count < Size);Count ++ )
                {
                AUTO FILE_ADDRESS SetAddress;
                AUTO STRING Word = Words[ Count ];

                Word.Lower();

                //
                //  Find the search word in the index.
                //
                if ( Root -> Index.FindKey( & Database,Word,& SetAddress ) )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope3( & Database );
                    AUTO SETS *NewSet;

                    //
                    //  Create a view on the associated row set and
                    //  then join it to the final result set.
                    //
                    if ( (Result = (AutomaticViewScope3.View( SetAddress,((VOID**) & NewSet) ))) )
                        {
                        if ( Count == 0 )
                            { Result = (Set.Copy( & Database,NewSet )); }
                        else
                            { Result = (Set.Intersect( & Database,NewSet )); }
                        }
                    }
                else
                    { Result = (Set.DeleteAll( & Database )); }
                }

            if ( Result )
                {
                //
                //  We now have the final result set to output it to
                //  the user.
                //
                if ( (Set.Size( & Database,& ResultSize )) && (ResultSize > 0) )
                    {
                    AUTO FILE_ADDRESS *ResultSet;

                    Set.Touch( & Database );

                    //
                    //  Create a view on all the rows in the result set.
                    //
                    if ( (Result = (Set.View( & Database,& ResultSet ))) )
                        {
                        AUTO FILE_SIZE Count1;
                        AUTO FILE_SIZE Count2;

                        //
                        //  Process the references one at a time.
                        //
                        for ( Count1=0,Count2=0;(Result) && (Count1 < ResultSize);Count1 ++ )
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope( & Database );
                            AUTO REFERENCE *Reference;

                            //
                            //  Create a view on each reference in the
                            //  result set.
                            //
                            if ( (Result = (AutomaticViewScope.View( ResultSet[ Count1 ],((VOID**) & Reference) ))) )
                                {
                                //
                                //  Output the reference to the user.
                                //
                                Display
                                    (
                                    ((Count1 == 0) ? ResultSize : 0),
                                    & Text[ Reference -> ParagraphStart ],
                                    Reference -> ParagraphSize,
                                    & Text[ Reference -> TitleStart ],
                                    ((Count2 != Reference -> TitleStart) ? Reference -> TitleSize : 0)
                                    );

                                if ( Count2 != Reference -> TitleStart )
                                    { Count2 = Reference -> TitleStart; }
                                }
                            }

                        //
                        //  End the view on the result set.
                        //
                        Result = (Set.EndView( & Database ));
                        }
                    }
                else
                    { Display( ResultSize,NULL,0,NULL,0 ); }

                return Result;
                }
            }
        }

    return False;
    }

    //
    //  New Paragraph.
    //
    //  Convert the static call into a call to the associated member
    //  function.
    //
BOOLEAN DATABASE::NewParagraph( void *Context,REFERENCE *Reference )
    {
    AUTO CONTEXT *FullContext = ((CONTEXT*) Context);

    return (FullContext -> Database -> NewParagraph( FullContext,Reference ));
    }

    //
    //  New Paragraph.
    //
    //  Create a new reference in the database.
    //
BOOLEAN DATABASE::NewParagraph( CONTEXT *Context,REFERENCE *Reference1 )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database;
    AUTO REFERENCE *Reference2;

    if ( (ROCKALL_NEW<REFERENCE>::New( & Database,& Context -> Reference,& Reference2 )) )
        {
        (*Reference2) = (*Reference1);

        return True;
        }

    return False;
    }

    //
    //  New Word.
    //
    //  Convert the static call into a call to the associated member
    //  function.
    //
BOOLEAN DATABASE::NewWord( void *Context,FILE_SIZE Size,char *Word )
    {
    AUTO CONTEXT *FullContext = ((CONTEXT*) Context);

    return (FullContext -> Database -> NewWord( FullContext,Size,Word ));
    }

    //
    //  New Word.
    //
    //  Create a new word to the database.
    //
BOOLEAN DATABASE::NewWord( CONTEXT *Context,FILE_SIZE Size,char *Word )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database;
    AUTO FILE_ADDRESS SetAddress;
    AUTO SETS *Set;
    AUTO STRING String;

    //
    //  Update the current count of words.
    //
    if ( (Context -> TotalWords % 1000) == 0 )
        {
        //
        //  Output the totals.
        //
        printf
            (
            "Summary: %d words (%d unique and %d refs).\r",
            ((int) Context -> TotalWords),
            ((int) Context -> UniqueWords),
            ((int) Context -> References)
            );
        }

    //
    //  Store the current word in a string.
    //
    String.Assignment( Word,Size );
    String.Lower();

    //
    //  Find the current word in the index.
    //
    if ( Context -> Root -> Index.FindKey( & Database,String,& SetAddress ) )
        {
        //
        //  If the word already exists in the index we create
        //  a view on the existing row set and add the new
        //  reference to it.  If the new reference is a duplicate
        //  we silently ignore the failure.
        //
        if ( Database.View( SetAddress,((VOID**) & Set) ) )
            {
            if ( Set -> NewValue( & Database,Context -> Reference ) )
                { Context -> References ++; }

            Context -> TotalWords ++;

            return True;
            }
        }
    else
        {
        //
        //  If the word does not exist in the index we create
        //  a new row set and add the new reference to it and
        //  then add it into the index.
        //
        if
                (
                (ROCKALL_NEW<SETS>::New( & Database,& SetAddress,& Set ))
                    &&
                (Set -> NewValue( & Database,Context -> Reference ))
                    &&
                (Context -> Root -> Index.NewKey( & Database,String,SetAddress ))
                )
            {
            Context -> References ++;
            Context -> TotalWords ++;
            Context -> UniqueWords ++;

            return True;
            }
        }

    return False;
    }

    //
    //  Walk.
    //
    //  Walk the index in the database.
    //
BOOLEAN DATABASE::Walk( DISPLAY_WORD Display )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database( & Heap );

    if ( (Active) && (Database.BeginTransaction()) )
        {
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope1( & Database );
        AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope2( & Database );
        AUTO FILE_ADDRESS RootAddress;
        AUTO ROOT *Root;
        AUTO char *Text;

        //
        //  Find the 'ROOT' structure and then create a view on
        //  it and the text of the book.
        //
        if
                (
                (Database.GetUserValue( 0,& RootAddress ))
                    &&
                (RootAddress != NULL)
                    &&
                (AutomaticViewScope1.View( RootAddress,((VOID**) & Root) ))
                    &&
                (AutomaticViewScope2.View( Root -> Book.Text,((VOID**) & Text) ))
                )
            {
            AUTO CONTEXT Context = { this,Display,0,0,Root,0,0 };
            AUTO INDEX *Index = & Root -> Index;

            //
            //  Walk the index and output every word.
            //
            if ( Index -> Walk( & Database,WalkFunction,((void*) & Context) ) )
                {
                //
                //  Output the totals.
                //
                printf
                    (
                    "\nSummary: %d unique words (%d refs).\n",
                    ((int) Context.UniqueWords),
                    ((int) Context.References)
                    );

                return True;
                }
            }
        }

    return False;
    }


    //
    //  Walk Function.
    //
    //  Walk callback function.
    //
BOOLEAN DATABASE::WalkFunction( STRING & String,FILE_ADDRESS & Address,VOID *UserValue )
    {
    AUTO AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> AutomaticHeapScope;
    AUTO AUTOMATIC_VIEW_SCOPE<HEAP_TYPE> AutomaticViewScope( & AutomaticHeapScope );
    AUTO CONTEXT *Context = ((CONTEXT*) UserValue);
    AUTO SETS *Set;

    //
    //  Create a view on the associated row set so we can
    //  get the size.
    //
    if ( AutomaticViewScope.View( Address,((VOID**) & Set) ) )
        {
        AUTO FILE_SIZE SetSize;

        //
        //  Get the size of the row set.
        //
        if ( Set -> Size( & AutomaticHeapScope,& SetSize ) )
            {
            AUTO CHAR *Text;

            //
            //  We get a view on the string.
            //
            if ( String.View( & AutomaticHeapScope,& Text ) )
                {
                AUTO BOOLEAN Newline = (((Context -> UniqueWords+1) % 3) == 0);

                //
                //  Display the current word.
                //
                ((DISPLAY_WORD) Context -> Display)( Text,SetSize,Newline );

                //
                //  Update the counters.
                //
                Context -> References += SetSize;
                Context -> UniqueWords ++;

                //
                //  End the view on the string.
                //
                String.EndView();

                return True;
                }
            }
        }

    return False;
    }

    //
    //  Close.
    //
    //  Close the database.
    //
BOOLEAN DATABASE::Close( void )
    {
    if ( Active )
        {
        AUTOMATIC_HEAP_SCOPE<HEAP_TYPE> Database( & Heap );

        if ( Database.CloseFile() )
            { return ((Active = False), True); }
        }

    return False;
    }

    //
    //  Database (Destructor).
    //
DATABASE::~DATABASE( void )
    {
    if ( Active )
        { Close(); }

    Active = False;
    }