#ifndef _DATABASE_HPP_
#define _DATABASE_HPP_ 1
    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.  A number of alternative structures
    //  are available to allow the example to be built in a
    //  variety of ways.
    //
#include "DatabaseHeap.hpp"
#include "FixedString.hpp"
#include "NoLock.hpp"
#include "RockallTypes.hpp"
#include "RowSet.hpp"
#include "Tree.hpp"
#include "Words.hpp"

#ifdef ALTERNATIVE_STRUCTURES
    //
    //  Any of the alternative heaps below may be selected and
    //  used in the example.
    //
#include "DatabaseHeap.hpp"
#include "MultiThreadedHeap.hpp"
#include "SingleThreadedHeap.hpp"
#include "TransactionalHeap.hpp"

    //
    //  Any of the alternative strings below may be selected
    //  and used in the example.
    //
#include "FixedString.hpp"
#include "FlexibleString.hpp"

    //
    //  Any of the alternative data structures below may be
    //  selected and used in the example.
    //
#include "Hash.hpp"
#include "Tree.hpp"
#endif

    //
    //  Database.
    //
    //  The database class manages all the interaction with
    //  Rockall-DB.
    //
class DATABASE
    {
        //
        //  Private constants.
        //
        STATIC CONST int MinText            = 32;

        //
        //  Private types.
        //
        typedef DATABASE_HEAP HEAP_TYPE;
        typedef ROW_SET<HEAP_TYPE,NO_LOCK> SETS;
        typedef FIXED_STRING<HEAP_TYPE,CHAR,MinText> STRING;
        typedef TREE<HEAP_TYPE,NO_LOCK,STRING,FILE_ADDRESS> INDEX;

#ifdef ALTERNATIVE_STRUCTURES
        //
        //  Any of the alternative heaps below may be selected
        //  and used in the example.
        //
        typedef DATABASE_HEAP HEAP_TYPE;
        typedef MULTI_THREADED_HEAP HEAP_TYPE;
        typedef SINGLE_THREADED_HEAP HEAP_TYPE;
        typedef TRANSACTIONAL_HEAP HEAP_TYPE;

        //
        //  Any of the alternative strings below may be selected
        //  and used in the example.
        //
        typedef FIXED_STRING<HEAP_TYPE,CHAR,MinText> STRING;
        typedef FLEXIBLE_STRING<HEAP_TYPE,CHAR> STRING;

        //
        //  Any of the alternative data structures below may
        //  be selected and used in the example.
        //
        typedef HASH<HEAP_TYPE,NO_LOCK,STRING,FILE_ADDRESS> INDEX;
        typedef TREE<HEAP_TYPE,NO_LOCK,STRING,FILE_ADDRESS> INDEX;
#endif

        typedef struct
            {
            FILE_SIZE                       Size;
            FILE_ADDRESS                    Text;
            }
        BOOK;

        typedef struct
            {
            BOOK                            Book;
            INDEX                           Index;
            }
        ROOT;

        typedef struct
            {
            DATABASE                        *Database;
            void                            *Display;
            FILE_ADDRESS                    Reference;
            FILE_SIZE                       References;
            ROOT                            *Root;
            FILE_SIZE                       TotalWords;
            FILE_SIZE                       UniqueWords;
            }
        CONTEXT;

        //
        //  Function prototypes.
        //
        typedef void (*DISPLAY_REFERENCE)
            (
            FILE_SIZE                       MaxSize,
            char                            *Paragraph,
            FILE_SIZE                       Size1,
            char                            *Title,
            FILE_SIZE                       Size12
            );

        typedef void (*DISPLAY_WORD)
            (
            char                            *Word,
            FILE_SIZE                       References,
            BOOLEAN                         Newline
            );

        //
        //  Private data.
        //
        BOOLEAN                             Active;
        HEAP_TYPE                           Heap;

    public:
        //
        //  Public functions.
        //
        DATABASE( void );

        BOOLEAN Create( char *FileName1,char *FileName2 );

        BOOLEAN Open( char *FileName );

        BOOLEAN Find( DISPLAY_REFERENCE Display,int Size,char *Words[] );

        BOOLEAN Walk( DISPLAY_WORD Display );

        BOOLEAN Close( void );

        ~DATABASE( void );

    private:
        //
        //  Private functions.
        //
        BOOLEAN NewParagraph( CONTEXT *Context,REFERENCE *Reference );

        BOOLEAN NewWord( CONTEXT *Context,FILE_SIZE Size,char *Text );

        //
        //  Private static functions.
        //
        STATIC BOOLEAN NewParagraph( void *Context,REFERENCE *Reference );

        STATIC BOOLEAN NewWord( void *Context,FILE_SIZE Size,char *Word );

        STATIC BOOLEAN WalkFunction( STRING & String,FILE_SIZE & Address,void *UserValue );

        //
        //  Disabled functionality.
        //
        DATABASE( const DATABASE & NoCopies );

        void operator=( const DATABASE & NoCopies );
    };
#endif