#ifdef ROCKALL_DB_EXAMPLES
#define main TestDump
#endif
    //
    //  Header files.
    //
    //  The header files provide access to functionality in the
    //  Rockall-DB library.
    //
#include <stdio.h>
#include <ctype.h>

#include "DatabaseHeap.hpp"
#include "Failure.hpp"
#include "RockallTypes.hpp"

    //
    //  Configuration and setup.
    //
    //  The following constants and types enable the program
    //  to be targeted as needed. The 'Ascii' value is the
    //  size of the ASCII output in bytes.  The hexadecimal
    //  value is the size of the hexadecimal output in bytes.
    //  A value of zero indicates no output of this type is
    //  needed.  The 'STRIDE' type specifies whether values
    //  should be formatted as 1, 2, 4 or 8 byte values.
    //
STATIC CONST int                                Ascii = 32;
STATIC CONST int                                Hexadecimal = 72;
typedef ROCKALL::UBIT32                         STRIDE;

    //
    //  Types.
    //
    //  The walk data provides basic information about
    //  the progress of the dump.
    //
typedef struct
    {
    FILE_SIZE                                   ActiveSpace;
    FILE_SIZE                                   Allocated;
    FILE_SIZE                                   Free;
    FILE_SIZE                                   FreeSpace;
    FILE_SIZE                                   IdleSpace;
    FILE_SIZE                                   Unallocated;
    }
WALK_DATA;

    //
    //  Heap Walk.
    //
    //  The heap walk function is called for each allocation
    //  in the associated file.  It outputs a line for each
    //  of these allocations containing its address, size
    //  and contents.
    //
BOOLEAN HeapWalk
        (
        BOOLEAN                                 Active,
        FILE_ADDRESS                            Address,
        VOID                                    *Data,
        BOOLEAN                                 Destroy,
        FILE_SIZE                               Space,
        VOID                                    *UserValue
        )
    {
    AUTO WALK_DATA *WalkData = ((WALK_DATA*) UserValue);
    STATIC CONST int Width = (sizeof(STRIDE)*2);

    //
    //  We only output information for allocated space.  All
    //  unallocated space is simply counted and output in a
    //  summary.
    //
    if ( Active )
        {
        AUTO int Count1;
        AUTO int Count2;
        AUTO int Count3;

        //
        //  When we find allocated space we output a summary
        //  of any previous unallocated space.
        //
        if ( WalkData -> Free > 0 )
            {
            AUTO FILE_SIZE Size = (WalkData -> FreeSpace / WalkData -> Free);

            if ( (WalkData -> Free * Size) == WalkData -> FreeSpace )
                {
                printf
                    (
                    "Free space: %lld * %lld = %lld bytes\n",
                    WalkData -> Free,
                    Size,
                    WalkData -> FreeSpace
                    );
                }
            else
                {
                printf
                    (
                    "Free space: %lld allocations (%lld bytes)\n",
                    WalkData -> Free,
                    WalkData -> FreeSpace
                    );
                }
            }

        //
        //  Output the titles if necessary.
        //
        if ( WalkData -> Allocated == 0 )
            {
            printf
                (
                "      File\n"
                "    Address        Size\n"
                );
            }

        //
        //  Update all the related statistics.
        //
        WalkData -> ActiveSpace += Space;
        WalkData -> Allocated ++;
        WalkData -> Free = 0;
        WalkData -> FreeSpace = 0;

        //
        //  We cycle round outputing the contents of the allocation
        //  using as many lines of output as is necessary.
        //
        for ( Count1=0,Count2=0;Count1 < Space;Count1+=Count2,Count2=0 )
            {
            //
            //  Output the basic allocation information.
            //
            if ( Count1 == 0 )
                { printf( "0x%012llx  %8lld",Address,Space ); }
            else
                {
                if ( (Ascii > 0) || (Hexadecimal > 0) )
                    {
                    for ( Count2=0;Count2 < 24;Count2 ++ )
                        { printf( " " ); }
                    }
                }

            //
            //  Output a seperator if necessary.
            //
            if ( (Ascii > 0) || (Hexadecimal > 0) )
                { printf( " : " ); }

            //
            //  Output the contents of the allocation in hexadecimal.
            //
            if ( Hexadecimal > 0 )
                {
                AUTO int MaxSpace = ((int) (Space-sizeof(STRIDE)));
                STATIC CONST int MaxSize = (Hexadecimal-(Width+1));

                //
                //  Output the prefix.
                //
                printf( "0x" );

                //
                //  Output the contents in hexadecimal.
                //
                for
                        (
                        Count2=0,Count3=0;
                        ((Count1 + Count2) <= MaxSpace) && (Count3 <= MaxSize);
                        Count2 += sizeof(STRIDE),Count3 += (Width+1)
                        )
                    {
                    printf
                        (
                        " %0*llx",
                        Width,
                        (*((STRIDE*) & ((CHAR*) Data)[ (Count1+Count2) ]))
                        );
                    }
                }

            //
            //  Add padding and output a seperator if necessary.
            //
            if ( (Ascii > 0) && (Hexadecimal > 0) )
                {
                for ( /* void */;Count3 < Hexadecimal;Count3 ++ )
                    { printf( " " ); }

                printf( " : " );
                }

            //
            //  Output the contents of the allocation in ASCII.
            //
            if ( Ascii > 0 )
                {
                AUTO int MaxSize = ((Count2 == 0) ? (Ascii-1) : (Count2-1));
                AUTO int MaxSpace = ((int) (Space-1));

                //
                //  Output the contents in ASCII.
                //
                for
                        (
                        Count2=0;
                        ((Count1 + Count2) <= MaxSpace) && (Count2 <= MaxSize);
                        Count2 += sizeof(CHAR)
                        )
                    {
                    AUTO int Value = ((((char*) Data)[ (Count1+Count2) ]) & 0xff);

                    if ( isprint( Value ) )
                        { putchar( Value ); }
                    else
                        { putchar( '?' ); }
                    }
                }

            //
            //  Output a new line.
            //
            putchar( '\n' );
            }
        }
    else
        {
        //
        //  Update any related statistics.
        //
        WalkData -> Free ++;
        WalkData -> FreeSpace += Space;
        WalkData -> IdleSpace += Space;
        WalkData -> Unallocated ++;
        }

    return True;
    }

    //
    //  Main.
    //
    //  The main function opens the file and processes it.
    //
int main( int Count,char *Argument[] )
    {
    DATABASE_HEAP Heap;

    //
    //  We make sure we have the correct number of parameters.
    //
    if ( Count == 2 )
        {
        //
        //  Open the file.
        //
        if ( Heap.OpenFile( Argument[1],True ) )
            {
            AUTO WALK_DATA WalkData = { 0 };

            //
            //  Walk the file.
            //
            if ( ! Heap.Walk( HeapWalk,& WalkData ) )
                { FAILURE( "Heap walk fails" ); }

            //
            //  Output a final summary.
            //
            printf
                (
                "Totals: Active %lld (%lld bytes), "
                "Idle %lld (%lld bytes), "
                "Total %lld (%lld bytes)\n",
                WalkData.Allocated,
                WalkData.ActiveSpace,
                WalkData.Unallocated,
                WalkData.IdleSpace,
                (WalkData.Allocated + WalkData.Unallocated),
                (WalkData.ActiveSpace + WalkData.IdleSpace)
                );

            //
            //  Close the file.
            //
            Heap.CloseFile();
            }
        else
            { printf( "Error: unable to open file '%s'\n",Argument[1] ); }
        }
    else
        { printf( "Usage: Dump <file name>\n" ); }

    return 0;
    }
