#ifndef _AUTOMATIC_LOCK_SCOPE_HPP_
#define _AUTOMATIC_LOCK_SCOPE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "ReservedWords.hpp"
#include "Failure.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Automatic Lock Scope.                                                                             */
        /*                                                                                                    */
        /*  An automated lock scope helps manage lock usage by ensuring a lock is claimed at the start        */
        /*  of a block and released at the end of a block (no matter how the block is exited).                */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Public types.                                                                             */
            /*                                                                                            */
            /*  The public types specify structures and function prototypes that are useful to            */
            /*  external code.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    typedef enum
        {
        AutomaticExclusiveLockType,
        AutomaticShareableLockType,
        EndOfAutomaticLockTypes
        }
    AUTOMATIC_LOCK_TYPES;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic Lock Scope.                                                                     */
            /*                                                                                            */
            /*  An automated lock scope is a fairly simple mechanism and is available for all of          */
            /*  the lock types.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class AUTOMATIC_LOCK_SCOPE
        {
            //
            //  Private data.
            //
            BOOLEAN                             Active;
            AUTOMATIC_LOCK_TYPES                ActiveLockType;
            TYPE                                *Spinlock;

        public:
            //
            //  Public inline functions.
            //
            AUTOMATIC_LOCK_SCOPE( AUTOMATIC_LOCK_TYPES NewActiveLockType,TYPE *NewSpinlock )
                {
                switch ( (ActiveLockType = NewActiveLockType) )
                    {
                    case AutomaticExclusiveLockType:
                        { Active = True; ((Spinlock = NewSpinlock) -> ClaimExclusiveLock()); break; }

                    case AutomaticShareableLockType:
                        { Active = True; ((Spinlock = NewSpinlock) -> ClaimSharedLock()); break; }

                    default:
                        { FAILURE( "Unknown lock type" ); }
                    }
                }

            INLINE BOOLEAN Reclaim( BOOLEAN Wait = True )
                {
                if ( ! Active )
                    {
                    switch ( ActiveLockType )
                        {
                        case AutomaticExclusiveLockType:
                            { return (Active = (Spinlock -> ClaimExclusiveLock( Wait ))); }

                        case AutomaticShareableLockType:
                            { return (Active = (Spinlock -> ClaimSharedLock( Wait ))); }

                        default:
                            { FAILURE( "Unknown lock type" ); }
                        }
                    }

                return False;
                }

            INLINE VOID Release( VOID )
                {
                if ( Active )
                    {
                    switch ( ActiveLockType )
                        {
                        case AutomaticExclusiveLockType:
                            { Active = False; (Spinlock -> ReleaseExclusiveLock()); break; }

                        case AutomaticShareableLockType:
                            { Active = False; (Spinlock -> ReleaseSharedLock()); break; }

                        default:
                            { FAILURE( "Unknown lock type" ); }
                        }
                    }

                }

            ~AUTOMATIC_LOCK_SCOPE( VOID )
                { Release(); }

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_LOCK_SCOPE( CONST AUTOMATIC_LOCK_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_LOCK_SCOPE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic Exclusive Lock Scope.                                                           */
            /*                                                                                            */
            /*  An automated exclusive lock scope is a fairly simple mechanism and is available for       */
            /*  the all the lock types.                                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class AUTOMATIC_EXCLUSIVE_LOCK_SCOPE : public AUTOMATIC_LOCK_SCOPE<TYPE>
        {
            //
            //  Private constants.
            //
            STATIC CONST AUTOMATIC_LOCK_TYPES Type  = AutomaticExclusiveLockType;

        public:
            //
            //  Public inline functions.
            //
            AUTOMATIC_EXCLUSIVE_LOCK_SCOPE( TYPE *NewSpinlock ) : AUTOMATIC_LOCK_SCOPE( Type,NewSpinlock )
                { /* void */ }

            ~AUTOMATIC_EXCLUSIVE_LOCK_SCOPE( VOID )
                { /* void */ }

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_EXCLUSIVE_LOCK_SCOPE( CONST AUTOMATIC_EXCLUSIVE_LOCK_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_EXCLUSIVE_LOCK_SCOPE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic Shareable Lock Scope.                                                           */
            /*                                                                                            */
            /*  An automated shareable lock scope is a fairly simple mechanism and is available for       */
            /*  the all the lock types.                                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class AUTOMATIC_SHAREABLE_LOCK_SCOPE  : public AUTOMATIC_LOCK_SCOPE<TYPE>
        {
            //
            //  Private constants.
            //
            STATIC CONST AUTOMATIC_LOCK_TYPES Type  = AutomaticShareableLockType;

        public:
            //
            //  Public inline functions.
            //
            AUTOMATIC_SHAREABLE_LOCK_SCOPE( TYPE *NewSpinlock ) : AUTOMATIC_LOCK_SCOPE( Type,NewSpinlock )
                { /* void */ }

            ~AUTOMATIC_SHAREABLE_LOCK_SCOPE( VOID )
                { /* void */ }

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_SHAREABLE_LOCK_SCOPE( CONST AUTOMATIC_SHAREABLE_LOCK_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_SHAREABLE_LOCK_SCOPE & NoCopies );
        };
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define AUTOMATIC_LOCK_SCOPE ROCKALL::AUTOMATIC_LOCK_SCOPE
#define AUTOMATIC_EXCLUSIVE_LOCK_SCOPE ROCKALL::AUTOMATIC_EXCLUSIVE_LOCK_SCOPE
#define AUTOMATIC_MULTI_USE_LOCK_SCOPE ROCKALL::AUTOMATIC_MULTI_USE_LOCK_SCOPE
#define AUTOMATIC_SHAREABLE_LOCK_SCOPE ROCKALL::AUTOMATIC_SHAREABLE_LOCK_SCOPE
#endif
#endif