#ifndef _AUTOMATIC_HEAP_SCOPE_HPP_
#define _AUTOMATIC_HEAP_SCOPE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "ExportThreadLocalStore.hpp"
#include "Error.hpp"
#include "Failure.hpp"
#include "ReservedWords.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Automatic Heap Scope.                                                                             */
        /*                                                                                                    */
        /*  An automatic heap scope will change the heap, region and various other settings at the start      */
        /*  of a block and then change them back as needed at the end of the block.  It will also skip        */
        /*  region related calls if there is no support for them in the underlying heap.                      */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic Heap Scope.                                                                     */
            /*                                                                                            */
            /*  An automatic heap scope will change the heap, region and various other settings at        */
            /*  the start of a block and then change them back as needed at the end of the block.         */
            /*  It  will also skip region related calls if there is no support for them in the            */
            /*  underlying heap.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> class AUTOMATIC_HEAP_SCOPE : private VIRTUAL_HEAP
        {
            //
            //  Private data.
            //
            HEAP                                *Heap;
            AUTOMATIC_HEAP_SCOPE<HEAP>          *Parent;
            SBIT32                              Region;
            BOOLEAN                             Suppress;
            FILE_ADDRESS                        TransactionID;
            TRANSACTION_LOCK<HEAP>              TransactionLock;
            SBIT32                              Transactions;
            BOOLEAN                             Truncate;
            SBIT32                              Undo;

            //
            //  Static private data.
            //
            STATIC EXPORT_THREAD_LOCAL_STORE    Parents;

        public:
            //
            //  Public functions.
            //
            AUTOMATIC_HEAP_SCOPE
                (
                HEAP                            *NewHeap = NULL,
                SBIT32                          NewRegion = NoRegion,
                SBIT16                          NewTruncate = NoValue
                );

            AUTOMATIC_HEAP_SCOPE
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                HEAP                            *NewHeap = NULL,
                SBIT32                          NewRegion = NoRegion,
                SBIT16                          NewTruncate = NoValue
                );

            //
            //  The Rockall-DB core functionality.
            //
            BOOLEAN Delete( FILE_ADDRESS Address,BOOLEAN NoDestroy = False,BOOLEAN Zero = False );

            BOOLEAN Delete( VOID *Data,BOOLEAN NoDestroy = False,BOOLEAN Zero = False );

            BOOLEAN DeleteAll( BOOLEAN NoDestroy = False,BOOLEAN Zero = False );

            BOOLEAN Details( FILE_ADDRESS Address,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL );

            BOOLEAN Details( VOID *Data,BOOLEAN *Destroy = NULL,FILE_SIZE *Space = NULL );

            SBIT32 Feature( VOID );

            BOOLEAN Feature( SBIT32 Features );

            BOOLEAN GetError( SBIT32 *ErrorNumber );

            BOOLEAN New
                (
                FILE_ADDRESS                    *Address,
                VOID                            **Data,
                FILE_SIZE                       Size,
                FILE_SIZE                       *Space = NULL,
                BOOLEAN                         Destroy = False,
                BOOLEAN                         Zero = False
                );

            BOOLEAN Resize
                (
                FILE_ADDRESS                    *Address,
                VOID                            **Data,
                FILE_SIZE                       NewSize,
                FILE_SIZE                       *Space = NULL,
                BOOLEAN                         Zero = False
                );

            BOOLEAN SetError( SBIT32 ErrorNumber );

            BOOLEAN Size( FILE_SIZE *Space );

            BOOLEAN Walk( HEAP_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            //
            //  The Rockall-DB transactional functionality.
            //
            BOOLEAN CreateFile
                (
                CHAR                            *FileName = NULL,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                );

            BOOLEAN CreateFile
                (
                WCHAR                           *FileName = NULL,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                );

            BOOLEAN CreateRegion( SBIT32 *ID );

            BOOLEAN BeginTransaction( FILE_ADDRESS *NewTransactionID = NULL,EXPORT_MEMORY_LOCK *UserLock = NULL );

            BOOLEAN ChangeRegion( SBIT32 NewID,SBIT32 *OldID = NULL );

            BOOLEAN ExclusiveView( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL );

            BOOLEAN EndExclusiveView( FILE_ADDRESS Address,BOOLEAN Reset = True );

            BOOLEAN JoinTransaction( FILE_ADDRESS NewTransactionID,EXPORT_MEMORY_LOCK *UserLock = NULL );

            BOOLEAN LeaveTransaction( VOID );

            BOOLEAN Update( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL );

            BOOLEAN View( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL );

            BOOLEAN EndView( FILE_ADDRESS Address,BOOLEAN Reset = True );

            BOOLEAN EndTransaction( BOOLEAN Abort = False,BOOLEAN Commit = False );

            BOOLEAN DeleteRegion( SBIT32 ID );

            BOOLEAN CloseFile( VOID );

            //
            //  The Rockall-DB database functionality.
            //
            BOOLEAN CreateShadowFile( CHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True );

            BOOLEAN CreateShadowFile( WCHAR *FileName,SBIT32 *ID = NULL,BOOLEAN Wait = True );

            BOOLEAN CreateSnapshotFile( CHAR *FileName );

            BOOLEAN CreateSnapshotFile( WCHAR *FileName );

            BOOLEAN OpenFile
                (
                CHAR                            *FileName,
                BOOLEAN                         ReadOnly = False,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                );

            BOOLEAN OpenFile
                (
                WCHAR                           *FileName,
                BOOLEAN                         ReadOnly = False,
                ENCRYPTION_FUNCTION             Function = NULL,
                VOID                            *UserValue = NULL
                );

            BOOLEAN Commit( VOID );

            BOOLEAN FileSize( FILE_SIZE *Size );

            BOOLEAN GetUserValue( SBIT32 ID,FILE_ADDRESS *Value );

            BOOLEAN SetUserValue( SBIT32 ID,FILE_ADDRESS Value );

            BOOLEAN Touch( FILE_ADDRESS Address );

            BOOLEAN CloseShadowFile( SBIT32 ID = 0 );

            ~AUTOMATIC_HEAP_SCOPE( VOID );

            //
            //  Public inline functions.
            //
            INLINE HEAP *GetHeap( VOID )
                { return Heap; }

            INLINE BOOLEAN GetTruncate( VOID )
                { return Truncate; }

            INLINE BOOLEAN New
                    (
                    VOID                        **Data,
                    FILE_SIZE                   Size,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Destroy = False,
                    BOOLEAN                     Zero = False
                    )
                { return (Heap -> New( NULL,Data,Size,Space,Destroy,Zero )); }

            INLINE BOOLEAN Resize
                    (
                    VOID                        **Data,
                    FILE_SIZE                   NewSize,
                    FILE_SIZE                   *Space = NULL,
                    BOOLEAN                     Zero = False
                    )
                { return (Heap -> Resize( NULL,Data,NewSize,Space,Zero )); }

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_HEAP_SCOPE( CONST AUTOMATIC_HEAP_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_HEAP_SCOPE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> EXPORT_THREAD_LOCAL_STORE AUTOMATIC_HEAP_SCOPE<HEAP>::Parents;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_HEAP_SCOPE'.                                                   */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_HEAP_SCOPE' prepares the class for use.                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_HEAP_SCOPE<HEAP>::AUTOMATIC_HEAP_SCOPE
            (
            HEAP                                *NewHeap,
            SBIT32                              NewRegion,
            SBIT16                              NewTruncate
            )
            //
            //  Call any special member constructors.
            //
            : TransactionLock( NewHeap )
        {
        //
        //  We inherit various class member values from our immediate 'Parent'.
        //
        if ( (Parent = ((AUTOMATIC_HEAP_SCOPE*) Parents.Get())) != NULL )
            {
            //
            //  Setup the class members based on our 'Parent'.
            //
            TransactionID = NoValue;
            Transactions = 0;
            Undo = NoRegion;

            //
            //  We will use 'NewHeap' in preference to any inherited 'Heap'.
            //
            if ( NewHeap == NULL )
                { Heap = (Parent -> Heap); }
            else
                { Heap = NewHeap; }

            //
            //  We only update the transactional information if necessary.
            //
            if ( Heap -> Feature( RockallTransactionalSupport ) )
                {
                //
                //  We check that the 'NewRegion' is different from our 'Parent'.
                //  If not, we suppress the region change.
                //
                if ( NewRegion == (Region = Parent -> Region) )
                    { NewRegion = NoRegion; }

                //
                //  We will 'Suppress' creating any new transactions if we already have
                //  any 'Transactions' in an outer scope.
                //
                Suppress = ((Heap == Parent -> Heap) && ((Parent -> Suppress) || (Parent -> Transactions > 0)));

                //
                //  When we 'Suppress' creating any new transactions we copy any information
                //  about any existing 'Transactions' into the current scope so we can return
                //  it if necessary.
                //
                if ( Suppress )
                    {
                    TransactionID = Parent -> TransactionID;
                    TransactionLock.RegisterLock( (Parent -> TransactionLock.GetLock()) );
                    }
                }
            else
                {
                //
                //  We just zero the unused values.
                //
                Region = NoRegion;
                Suppress = False;
                }

            //
            //  We will use 'NewTruncate' in preference to any inherited 'Truncate'.
            //
            if ( NewTruncate == NoValue )
                { Truncate = (Parent -> Truncate); }
            else
                { Truncate = (NewTruncate == ((SBIT16) True)); }
            }
        else
            {
            //
            //  We must be provided a 'NewHeap' as the outer most scope.  If not, it is
            //  pointless creating the instance as we will not be able to make any calls.
            //
            if ( NewHeap != NULL )
                {
                //
                //  Setup the class members if we have no 'Parent'.
                //
                Heap = NewHeap;
                Region = NoRegion;
                Suppress = False;
                TransactionID = NoValue;
                Transactions = 0;
                Undo = NoRegion;

                //
                //  We will use 'NewTruncate' in preference to the default.
                //
                if ( NewTruncate == NoValue )
                    { Truncate = False; }
                else
                    { Truncate = (NewTruncate == ((SBIT16) True)); }
                }
            else
                { FAILURE( "A 'NewHeap' must be supplied at the outer most scope" ); }
            }

        //
        //  We can now change to a 'NewRegion' if needed.
        //
        if ( (Heap -> Feature( RockallTransactionalSupport )) && (NewRegion != NoRegion) )
            {
            if ( Heap -> ChangeRegion( NewRegion,& Undo ) )
                { Region = NewRegion; }
            else
                { FAILURE( "Unable to change to new region" ); }
            }

        //
        //  Now we have set up all the members we can now set 'Parents'.
        //
        if ( ! Parents.Set( this ) )
            { FAILURE( "Unable to set thread local store" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_HEAP_SCOPE'.                                                   */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_HEAP_SCOPE' prepares the class for use.                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_HEAP_SCOPE<HEAP>::AUTOMATIC_HEAP_SCOPE
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            HEAP                                *NewHeap = NULL,
            SBIT32                              NewRegion = NoRegion,
            SBIT16                              NewTruncate = NoValue
            )
            //
            //  Call any special member constructors.
            //
            : TransactionLock( AutomaticHeapScope )
        {
        //
        //  Setup the class members based on the supplied 'AutomaticHeapScope'.
        //
        Parent = ((AUTOMATIC_HEAP_SCOPE*) Parents.Get());
        TransactionID = NoValue;
        Transactions = 0;
        Undo = NoRegion;

        //
        //  We will use 'NewHeap' in preference to any inherited 'Heap'.
        //
        if ( NewHeap == NULL )
            { Heap = (AutomaticHeapScope -> Heap); }
        else
            { Heap = NewHeap; }

        //
        //  We only update the transactional information if necessary.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We check that the 'NewRegion' is different from our parent.
            //  If not, we suppress the region change.
            //
            if ( NewRegion == (Region = AutomaticHeapScope -> Region) )
                { NewRegion = NoRegion; }

            //
            //  We will 'Suppress' creating any new transactions if we already have
            //  any 'Transactions' in an outer scope.
            //
            Suppress =
                (
                (Heap == AutomaticHeapScope -> Heap)
                    &&
                ((AutomaticHeapScope -> Suppress) || (AutomaticHeapScope -> Transactions > 0))
                );

            //
            //  When we 'Suppress' creating any new transactions we copy any information
            //  about any existing 'Transactions' into the current scope so we can return
            //  it if necessary.
            //
            if ( Suppress )
                {
                TransactionID = AutomaticHeapScope -> TransactionID;
                TransactionLock.RegisterLock( (AutomaticHeapScope -> TransactionLock.GetLock()) );
                }
            }
        else
            {
            //
            //  We just zero the unused values.
            //
            Region = NoRegion;
            Suppress = False;
            }

        //
        //  We will use 'NewTruncate' in preference to any inherited 'Truncate'.
        //
        if ( NewTruncate == NoValue )
            { Truncate = (AutomaticHeapScope -> Truncate); }
        else
            { Truncate = (NewTruncate == ((SBIT16) True)); }

        //
        //  We can now change to a 'NewRegion' if needed.
        //
        if ( (Heap -> Feature( RockallTransactionalSupport )) && (NewRegion != NoRegion) )
            {
            if ( Heap -> ChangeRegion( NewRegion,& Undo ) )
                { Region = NewRegion; }
            else
                { FAILURE( "Unable to change to new region" ); }
            }

        //
        //  Now we have set up all the members we can now set 'Parents'.
        //
        if ( ! Parents.Set( this ) )
            { FAILURE( "Unable to set thread local store" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create File.                                                                              */
            /*                                                                                            */
            /*  Create a file.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateFile
            (
            CHAR                                *FileName,
            ENCRYPTION_FUNCTION                 Function,
            VOID                                *UserValue
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> CreateFile( FileName,Function,UserValue )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create File.                                                                              */
            /*                                                                                            */
            /*  Create a file.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateFile
            (
            WCHAR                               *FileName,
            ENCRYPTION_FUNCTION                 Function,
            VOID                                *UserValue
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> CreateFile( FileName,Function,UserValue )); }
        else
            { return True; }

        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Shadow File.                                                                       */
            /*                                                                                            */
            /*  Create a shadow file.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateShadowFile
            (
            CHAR                                *FileName,
            SBIT32                              *ID,
            BOOLEAN                             Wait
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> CreateShadowFile( FileName,ID,Wait )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Shadow File.                                                                       */
            /*                                                                                            */
            /*  Create a shadow file.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateShadowFile
            (
            WCHAR                               *FileName,
            SBIT32                              *ID,
            BOOLEAN                             Wait
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> CreateShadowFile( FileName,ID,Wait )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Snapshot File.                                                                     */
            /*                                                                                            */
            /*  Create a snapshot file.                                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateSnapshotFile( CHAR *FileName )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> CreateSnapshotFile( FileName )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Snapshot File.                                                                     */
            /*                                                                                            */
            /*  Create a snapshot file.                                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateSnapshotFile( WCHAR *FileName )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> CreateSnapshotFile( FileName )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Open File.                                                                                */
            /*                                                                                            */
            /*  Open a file.                                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::OpenFile
            (
            CHAR                                *FileName,
            BOOLEAN                             ReadOnly,
            ENCRYPTION_FUNCTION                 Function,
            VOID                                *UserValue
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> OpenFile( FileName,ReadOnly,Function,UserValue )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Open File.                                                                                */
            /*                                                                                            */
            /*  Open a file.                                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::OpenFile
            (
            WCHAR                               *FileName,
            BOOLEAN                             ReadOnly,
            ENCRYPTION_FUNCTION                 Function,
            VOID                                *UserValue
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> OpenFile( FileName,ReadOnly,Function,UserValue )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Create Region.                                                                            */
            /*                                                                                            */
            /*  Create a region.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CreateRegion( SBIT32 *ID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> CreateRegion( ID )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Begin Transaction.                                                                        */
            /*                                                                                            */
            /*  Begin a transaction.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::BeginTransaction
            (
            FILE_ADDRESS                        *NewTransactionID,
            EXPORT_MEMORY_LOCK                  *UserLock
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will call 'BeginTransaction()' if this is the first call, we do not need
            //  to 'Suppress' the call as we have a transaction in an outer scope and we have
            //  support for transactions.
            //
            if ( ((Transactions ++) == 0) && (! Suppress) )
                {
                //
                //  We can now call 'BeginTransaction()' which we would expect to work.  If not,
                //  we decrement 'Transactions' and exit.
                //
                if ( Heap -> BeginTransaction( & TransactionID,& TransactionLock ) )
                    {
                    //
                    //  Update 'NewTransactionID' if required.
                    //
                    if ( NewTransactionID != NULL )
                        { (*NewTransactionID) = TransactionID; }

                    //
                    //  Update 'UserLock' if required.
                    //
                    if ( UserLock != NULL )
                        { UserLock -> RegisterLock( (TransactionLock.GetLock()) ); }

                    return True;
                    }
                else
                    {
                    AUTO SBIT32 Error;

                    //
                    //  Nasty: the call to 'BeginTransaction()' was not successsful.  It
                    //  maybe because somebody has alreday created a transaction on this
                    //  'Heap'.  If so, we 'Suppress' all new transcations and assume that
                    //  a suitable call to 'EndTransaction()' will be made.  If not, we
                    //  decrement 'Transactions' and exit.
                    //
                    if ( (Heap -> GetError( & Error )) && (Error == ErrorActiveTransaction) )
                        {
                        //
                        //  Update 'NewTransactionID' if required.
                        //
                        if ( NewTransactionID != NULL )
                            { (*NewTransactionID) = TransactionID; }

                        //
                        //  Update 'UserLock' if required.
                        //
                        if ( UserLock != NULL )
                            { UserLock -> RegisterLock( (TransactionLock.GetLock()) ); }

                        return (Suppress = True);
                        }
                    else
                        { Transactions --; }
                    }
                }
            else
                {
                //
                //  Update 'NewTransactionID' if required.
                //
                if ( NewTransactionID != NULL )
                    { (*NewTransactionID) = TransactionID; }

                //
                //  Update 'UserLock' if required.
                //
                if ( UserLock != NULL )
                    { UserLock -> RegisterLock( (TransactionLock.GetLock()) ); }

                return True;
                }
            }
        else
            { return True; }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Change Region.                                                                            */
            /*                                                                                            */
            /*  Change the region.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::ChangeRegion( SBIT32 NewID,SBIT32 *OldID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> ChangeRegion( NewID,OldID )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Commit.                                                                                   */
            /*                                                                                            */
            /*  Commit a transaction.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Commit( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> Commit()); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete.                                                                                   */
            /*                                                                                            */
            /*  Delete an allocation.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Delete
            (
            FILE_ADDRESS                        Address,
            BOOLEAN                             NoDestroy,
            BOOLEAN                             Zero
            )
        { return (Heap -> Delete( Address,NoDestroy,Zero )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete.                                                                                   */
            /*                                                                                            */
            /*  Delete an allocation.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Delete
            (
            VOID                                *Data,
            BOOLEAN                             NoDestroy,
            BOOLEAN                             Zero
            )
        { return (Heap -> Delete( Data,NoDestroy,Zero )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all allocations.                                                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::DeleteAll( BOOLEAN NoDestroy,BOOLEAN Zero )
        { return (Heap -> DeleteAll( NoDestroy,Zero )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Deatils.                                                                                  */
            /*                                                                                            */
            /*  Compute an allocations actual size.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Details
            (
            FILE_ADDRESS                        Address,
            BOOLEAN                             *Destroy,
            FILE_SIZE                           *Space
            )
        { return (Heap -> Details( Address,Destroy,Space )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Details.                                                                                  */
            /*                                                                                            */
            /*  Compute an allocations actual size.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Details
            (
            VOID                                *Data,
            BOOLEAN                             *Destroy,
            FILE_SIZE                           *Space
            )
        { return (Heap -> Details( Data,Destroy,Space )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End Exclusive View.                                                                       */
            /*                                                                                            */
            /*  End an exclusive view.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::EndExclusiveView( FILE_ADDRESS Address,BOOLEAN Reset )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> EndExclusiveView( Address,Reset )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End Transaction.                                                                          */
            /*                                                                                            */
            /*  End a transaction.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::EndTransaction( BOOLEAN Abort,BOOLEAN Commit )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will exit immediately if we have no 'Transactions'.
            //
            if ( Transactions > 0 )
                {
                //
                //  We will call 'EndTransaction()' if this is the last of the 'Transactions',
                //  we do not need to 'Suppress' the call due to a transaction in an outer
                //  scope and we have support for transactions.
                //
                if ( ((-- Transactions) == 0) && (! Suppress) )
                    {
                    //
                    //  We can now call 'EndTransaction()' which we really expect to work.
                    //
                    if ( Heap -> EndTransaction( Abort,Commit ) )
                        { return ((TransactionID = NoValue), True); }
                    else
                        { Transactions ++; }
                    }
                else
                    { return True; }
                }
            }
        else
            { return True; }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End View.                                                                                 */
            /*                                                                                            */
            /*  End a view.                                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::EndView( FILE_ADDRESS Address,BOOLEAN Reset )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> EndView( Address,Reset )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Exclusive View.                                                                           */
            /*                                                                                            */
            /*  An exclusive view of an area.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::ExclusiveView
            (
            FILE_ADDRESS                        Address,
            VOID                                **Data,
            FILE_SIZE                           *Space
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> ExclusiveView( Address,Data,Space )); }
        else
            {
            //
            //  When the 'Heap' does not have support for transactions we may need to call
            //  'Details()' if we need 'Space'.  Regardless, we need to cast 'Address'
            //  to 'Data', as they are known to be the same in this case.
            //
            if ( (Space == NULL) || (Heap -> Details( Address,NULL,Space )) )
                {
                if ( Data != NULL )
                    { (*Data) = ((VOID*) Address); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Feature.                                                                                  */
            /*                                                                                            */
            /*  Ensure a heap has a feature.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> SBIT32 AUTOMATIC_HEAP_SCOPE<HEAP>::Feature( VOID )
        { return (Heap -> Feature()); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Feature.                                                                                  */
            /*                                                                                            */
            /*  Ensure a heap has a feature.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Feature( SBIT32 Features )
        { return (Heap -> Feature( Features )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  File Size.                                                                                */
            /*                                                                                            */
            /*  Compute the file size.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::FileSize( FILE_SIZE *Size )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> FileSize( Size )); }
        else
            { return (((*Size) = 0), True); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Get Error.                                                                                */
            /*                                                                                            */
            /*  Get the last error.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::GetError( SBIT32 *ErrorNumber )
        { return (Heap -> GetError( ErrorNumber )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Get User Value.                                                                           */
            /*                                                                                            */
            /*  Get a user value.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::GetUserValue( SBIT32 ID,FILE_ADDRESS *Value )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> GetUserValue( ID,Value )); }
        else
            { return (((*Value) = 0), True); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Join Transaction.                                                                         */
            /*                                                                                            */
            /*  Join a transaction.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::JoinTransaction
            (
            FILE_ADDRESS                        NewTransactionID,
            EXPORT_MEMORY_LOCK                  *UserLock
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will call 'JoinTransaction()' if this is the first call, we do not need
            //  to 'Suppress' the call as we have a transaction in an outer scope and we have
            //  support for transactions.
            //
            if ( ((Transactions ++) == 0) && (! Suppress) )
                {
                //
                //  We can now call 'JoinTransaction()' which we would expect to work.  If not,
                //  we decrement 'Transactions' and exit.
                //
                if ( Heap -> JoinTransaction( NewTransactionID,& TransactionLock ) )
                    {
                    //
                    //  Update the 'TransactionID'.
                    //
                    TransactionID = NewTransactionID;

                    //
                    //  Update 'UserLock' if required.
                    //
                    if ( UserLock != NULL )
                        { UserLock -> RegisterLock( (TransactionLock.GetLock()) ); }

                    return True;
                    }
                else
                    {
                    AUTO SBIT32 Error;

                    //
                    //  Nasty: the call to 'JoinTransaction()' was not successsful.  It
                    //  maybe because somebody has alreday created a transaction on this
                    //  'Heap'.  If so, we 'Suppress' all new transcations and assume that
                    //  a suitable call to 'EndTransaction()' will be made.  If not, we
                    //  decrement 'Transactions' and exit.
                    //
                    if ( (Heap -> GetError( & Error )) && (Error == ErrorActiveTransaction) )
                        { return (Suppress = True); }
                    else
                        { Transactions --; }
                    }
                }
            else
                {
                //
                //  Update 'UserLock' if required.
                //
                if ( UserLock != NULL )
                    { UserLock -> RegisterLock( (TransactionLock.GetLock()) ); }

                return True;
                }
            }
        else
            { return True; }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Leave Transaction.                                                                        */
            /*                                                                                            */
            /*  Join a transaction.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::LeaveTransaction( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will exit immediately if we have no 'Transactions'.
            //
            if ( Transactions > 0 )
                {
                //
                //  We will call 'LeaveTransaction()' if this is the last of the 'Transactions',
                //  we do not need to 'Suppress' the call due to a transaction in an outer
                //  scope and we have support for transactions.
                //
                if ( ((-- Transactions) == 0) && (! Suppress) )
                    {
                    //
                    //  We can now call 'EndTransaction()' which we really expect to work.
                    //
                    if ( Heap -> LeaveTransaction() )
                        { return ((TransactionID = NoValue), True); }
                    else
                        { Transactions ++; }
                    }
                else
                    { return True; }
                }
            }
        else
            { return True; }

        return False;
        }
            /**********************************************************************************************/
            /*                                                                                            */
            /*  New.                                                                                      */
            /*                                                                                            */
            /*  A new allocation.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::New
            (
            FILE_ADDRESS                        *Address,
            VOID                                **Data,
            FILE_SIZE                           Size,
            FILE_SIZE                           *Space,
            BOOLEAN                             Destroy,
            BOOLEAN                             Zero
            )
        { return (Heap -> New( Address,Data,Size,Space,Destroy,Zero )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Resize.                                                                                   */
            /*                                                                                            */
            /*  Resize an allocation.                                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Resize
            (
            FILE_ADDRESS                        *Address,
            VOID                                **Data,
            FILE_SIZE                           Size,
            FILE_SIZE                           *Space,
            BOOLEAN                             Zero
            )
        { return (Heap -> Resize( Address,Data,Size,Space,Zero )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Set Error.                                                                                */
            /*                                                                                            */
            /*  Set the error.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::SetError( SBIT32 ErrorNumber )
        { return (Heap -> SetError( ErrorNumber )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Set User Value.                                                                           */
            /*                                                                                            */
            /*  Set a user value.                                                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::SetUserValue( SBIT32 ID,FILE_ADDRESS Value )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> SetUserValue( ID,Value )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Compute the size of the heap.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Size( FILE_SIZE *Space )
        { return (Heap -> Size( Space )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Touch.                                                                                    */
            /*                                                                                            */
            /*  Touch a page.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Touch( FILE_ADDRESS Address )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> Touch( Address )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update.                                                                                   */
            /*                                                                                            */
            /*  Update an area.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Update
            (
            FILE_ADDRESS                        Address,
            VOID                                **Data,
            FILE_SIZE                           *Space
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> Update( Address,Data,Space )); }
        else
            {
            //
            //  When the 'Heap' does not have support for transactions we may need to call
            //  'Details()' if we need 'Space'.  Regardless, we need to cast 'Address'
            //  to 'Data', as they are known to be the same in this case.
            //
            if ( (Space == NULL) || (Heap -> Details( Address,NULL,Space )) )
                {
                if ( Data != NULL )
                    { (*Data) = ((VOID*) Address); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View an area.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::View
            (
            FILE_ADDRESS                        Address,
            VOID                                **Data,
            FILE_SIZE                           *Space
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> View( Address,Data,Space )); }
        else
            {
            //
            //  When the 'Heap' does not have support for transactions we may need to call
            //  'Details()' if we need 'Space'.  Regardless, we need to cast 'Address'
            //  to 'Data', as they are known to be the same in this case.
            //
            if ( (Space == NULL) || (Heap -> Details( Address,NULL,Space )) )
                {
                if ( Data != NULL )
                    { (*Data) = ((VOID*) Address); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the heap.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::Walk
            (
            HEAP_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        { return (Heap -> Walk( WalkFunction,UserValue )); }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Region.                                                                            */
            /*                                                                                            */
            /*  Delete a region.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::DeleteRegion( SBIT32 ID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> DeleteRegion( ID )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Close Shadow File.                                                                        */
            /*                                                                                            */
            /*  Close a shadow file.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CloseShadowFile( SBIT32 ID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallFileSupport ) )
            { return (Heap -> CloseShadowFile( ID )); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Close File.                                                                               */
            /*                                                                                            */
            /*  Close the file.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_HEAP_SCOPE<HEAP>::CloseFile( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            { return (Heap -> CloseFile()); }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'AUTOMATIC_HEAP_SCOPE'.                                                    */
            /*                                                                                            */
            /*  The destructor for 'AUTOMATIC_HEAP_SCOPE' cleans up the class.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_HEAP_SCOPE<HEAP>::~AUTOMATIC_HEAP_SCOPE( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  If we have a transaction we call 'EndTransaction()'.
            //
            if ( (Transactions > 0) && (! Suppress) )
                {
                if ( Heap -> EndTransaction() )
                    { Transactions = 0; }
                else
                    { FAILURE( "Unable to 'EndTransaction()'" ); }
                }

            //
            //  We change back to the original region.
            //
            if ( (Undo != NoRegion) && (! Heap -> ChangeRegion( Undo )) )
                { FAILURE( "Unable to change to original region" ); }
            }

        //
        //  We now reset the thread local store to the 'Parent'.
        //
        if ( ! Parents.Set( Parent ) )
            { FAILURE( "Unable to set thread local store" ); }

        //
        //  Zero the class members.
        //
        Undo = NoRegion;
        Truncate = False;
        Transactions = 0;
        TransactionID = NoValue;
        Suppress = False;
        Region = NoRegion;
        Parent = NULL;
        Heap = NULL;
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define AUTOMATIC_HEAP_SCOPE ROCKALL::AUTOMATIC_HEAP_SCOPE
#endif
#endif