#ifndef _QUEUE_HPP_
#define _QUEUE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticLockScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Failure.hpp"
#include "FileLock.hpp"
#include "MemoryFunctions.hpp"
#include "MemoryLock.hpp"
#include "NoLock.hpp"
#include "PlacementNew.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Queue.                                                                                            */
        /*                                                                                                    */
        /*  A queue is a simple queuing data structure for storing values.                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Queue.                                                                                    */
            /*                                                                                            */
            /*  A queue is a simple queuing data structure for storing values.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> class QUEUE
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 CacheSize       = 64;
            STATIC CONST SBIT32 DefaultSize     = (CacheSize * 4);
            STATIC CONST SBIT32 ViewSize        = 8192;

            //
            //  The root signature.
            //
            STATIC CONST SBIT64 Signature1      = ((sizeof(LOCK) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature2      = (((Signature1 + sizeof(VALUE)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature       = (0x0000200000000000 + Signature2);

            //
            //  Private types.
            //
            typedef TRANSACTION_LOCK<HEAP> TRANSACTIONAL_LOCK;

            typedef struct
                {
                VALUE                           Values[ CacheSize ];
                SBIT32                          Used;
                }
            CACHE;

            typedef struct
                {
                //
                //  Data locks.
                //
                LOCK                            ClassLock;

                //
                //  Protected data.
                //
                FILE_SIZE                       First;
                FILE_SIZE                       Last;
                FILE_SIZE                       MaxSize;
                CACHE                           PopCache;
                CACHE                           PushCache;
                FILE_ADDRESS                    Queue;
                FILE_SIZE                       Signature;
                }
            ROOT;

            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        Address;

            //
            //  Private static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  Public types.
            //
            typedef BOOLEAN (*QUEUE_WALK_FUNCTION)
                (
                VALUE                           & Value,
                VOID                            *UserValue
                );

            //
            //  Public functions.
            //
            QUEUE( VOID );
            QUEUE( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteAll( VOID );
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN PeekValue( VALUE *Value );
            BOOLEAN PeekValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,VALUE *Value );

            BOOLEAN PopValue( VALUE *Value );
            BOOLEAN PopValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,VALUE *Value );

            BOOLEAN PushValue( CONST VALUE & Value );
            BOOLEAN PushValue( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST VALUE & Value );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN Walk( QUEUE_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            BOOLEAN Walk
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                QUEUE_WALK_FUNCTION             WalkFunction,
                VOID                            *UserValue = NULL
                );

            ~QUEUE( VOID );

        private:
            //
            //  Private functions.
            //
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root );

            BOOLEAN MergeCaches( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,ROOT *Root );

            VOID ZeroCache( CACHE *Cache );

            VOID ZeroValues( VALUE *Values,FILE_SIZE Size );

            //
            //  Disabled functionality.
            //
            QUEUE( CONST QUEUE & NoCopies );

            VOID operator=( CONST QUEUE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> MEMORY_FUNCTIONS<SNATIVE> QUEUE<HEAP,LOCK,VALUE>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'QUEUE'.                                                                  */
            /*                                                                                            */
            /*  The constructor for 'QUEUE' prepares the class for use.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> QUEUE<HEAP,LOCK,VALUE>::QUEUE( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope.Feature( (LOCK::Feature()) ) )
            {
            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticHeapScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( & AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the 'Queue'.
                    //
                    Root -> First = 0;
                    Root -> Last = 0;
                    Root -> MaxSize = 0;
                    Root -> Queue = NULL;
                    Root -> Signature = Signature;

                    //
                    //  Zero the caches.
                    //
                    ZeroCache( & Root -> PopCache );
                    ZeroCache( & Root -> PushCache );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'QUEUE'.                                                                  */
            /*                                                                                            */
            /*  The constructor for 'QUEUE' prepares the class for use.                                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> QUEUE<HEAP,LOCK,VALUE>::QUEUE
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        //
        //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
        //  we fail to highlight the issue.
        //
        if ( AutomaticHeapScope -> Feature( (LOCK::Feature()) ) )
            {
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  Where necessary we automatically wrap all the changes within a new transaction.
            //
            if ( AutomaticTransactionScope.BeginTransaction() )
                {
                AUTO ROOT *Root;

                //
                //  We create a new 'Root' on the heap and store the 'Address'.
                //
                if ( ROCKALL_NEW<ROOT>::New( AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                    {
                    //
                    //  We now setup the 'Root' of the 'Queue'.
                    //
                    Root -> First = 0;
                    Root -> Last = 0;
                    Root -> MaxSize = 0;
                    Root -> Queue = NULL;
                    Root -> Signature = Signature;

                    //
                    //  Zero the caches.
                    //
                    ZeroCache( & Root -> PopCache );
                    ZeroCache( & Root -> PushCache );

                    Active = True;
                    }
                }
            }
        else
            { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the queue values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::DeleteAll( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteAll( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the queue values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Delete all the values.
                //
                return (DeleteAll( AutomaticHeapScope,Root ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the queue values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root
            )
        {
        //
        //  We delete the 'Queue' if needed.
        //
        if ( Root -> MaxSize > 0 )
            {
            AUTO VALUE *Queue;

            //
            //  We will need to 'Update()' the 'Queue'.
            //
            if ( AutomaticHeapScope -> Update( Root -> Queue,((VOID**) & Queue) ) )
                {
                AUTO FILE_SIZE Count;

                //
                //  Call the extra destructors.
                //
                for ( Count=(Root -> MaxSize-1);Count >= DefaultSize;Count -- )
                    { PLACEMENT_DELETE( & Queue[ Count ],VALUE ); }

                //
                //  Delete the queue.
                //
                if ( ROCKALL_DELETE<VALUE>::Delete( AutomaticHeapScope,Root -> Queue,DefaultSize ) )
                    {
                    Root -> First = 0;
                    Root -> Last = 0;
                    Root -> MaxSize = 0;
                    Root -> Queue = NULL;

                    //
                    //  Finally, we zero the 'PopCache' and the 'PushCache'.
                    //
                    ZeroCache( & Root -> PopCache );
                    ZeroCache( & Root -> PushCache );

                    return True;
                    }
                else
                    { FAILURE( "Unable to delete the queue values" ); }
                }

            return False;
            }
        else
            {
            //
            //  When there is no 'Queue' all we need to do is zero the 'PopCache'
            //  and the 'PushCache'.
            //
            ZeroCache( & Root -> PopCache );
            ZeroCache( & Root -> PushCache );

            return True;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Merge Caches.                                                                             */
            /*                                                                                            */
            /*  Merge the caches into the queue.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::MergeCaches
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            ROOT                                *Root
            )
        {
        AUTO VALUE *Queue = NULL;
        AUTO FILE_SIZE Size = 0;

        //
        //  Compute the current 'Size' of the 'Queue'.
        //
        if ( Root -> First <= Root -> Last )
            { Size = (Root -> Last - Root -> First); }
        else
            { Size = ((Root -> MaxSize - Root -> First) + (Root -> Last)); }

        //
        //  We will not need to use the 'Queue' if we have enough space in the caches
        //  and the 'Queue' is currently empty.
        //
        if ( ((Root -> PopCache.Used + Root -> PushCache.Used) <= CacheSize) && (Size == 0) )
            {
            //
            //  We will fill the 'PopCache' if there is space and the 'PushCache' is
            //  not empty.
            //
            if ( Root -> PushCache.Used > 0 )
                {
                AUTO SBIT32 Size1 = Root -> PopCache.Used;
                AUTO SBIT32 Size2 = Root -> PushCache.Used;
                AUTO SBIT32 Count;

                //
                //  Move the values along in the 'PopCache' to make space for the values
                //  in the 'PushCache'.
                //
                for ( Count=(Size1-1);Count >= 0;Count -- )
                    {  Root -> PopCache.Values[ (Size2 + Count) ] = Root -> PopCache.Values[ Count ]; }

                //
                //  Copy all the values in the 'PushCache' into the 'PopCache'.
                //
                for ( Count=(Size2-1);Count >= 0;Count -- )
                    { Root -> PopCache.Values[ (Size2 - Count - 1) ] = Root -> PushCache.Values[ Count ]; }

                //
                //  We zero the 'PushCache' now it is empty.
                //
                ZeroCache( & Root -> PushCache );

                //
                //  Update the sizes of the caches.
                //
                Root -> PopCache.Used += Size2;
                Root -> PushCache.Used = 0;
                }

            return (Root -> PopCache.Used > 0);
            }

        //
        //  We will create a 'Queue' if one does not already exist.
        //
        if ( Root -> MaxSize <= 0 )
            {
            AUTO FILE_SIZE Space;

            //
            //  We create a new 'Queue'.
            //
            if ( ROCKALL_NEW<VALUE>::New( AutomaticHeapScope,& Root -> Queue,& Queue,DefaultSize,& Space,False,True ) )
                {
                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Queue);
                AUTO FILE_SIZE Count;

                //
                //  Compute the 'MaxSize' and zero 'First' and 'Last' just to be sure.
                //
                Root -> MaxSize = (Space / sizeof(VALUE));
                Root -> First = 0;
                Root -> Last = 0;

                //
                //  Call the extra constructors.
                //
                for ( Count=DefaultSize;Count < Root -> MaxSize;Count ++ )
                    { PLACEMENT_NEW( & Queue[ Count ],VALUE ); }
                }
            else
                { return False; }
            }

        //
        //  We will expand the 'Queue' if it does not have enough space.
        //
        if ( Root -> MaxSize <= (Size + Root -> PushCache.Used) )
            {
            AUTO FILE_SIZE MaxSize = (Root -> MaxSize);
            AUTO FILE_SIZE NewSize = ((MaxSize * GrowthFactor) * sizeof(VALUE));
            AUTO FILE_SIZE Space;

            //
            //  We 'Resize()' the existing 'Queue'.
            //
            if ( AutomaticHeapScope -> Resize( & Root -> Queue,((VOID**) & Queue),NewSize,& Space,True ) )
                {
                AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Queue);
                AUTO FILE_SIZE Count;

                //
                //  Compute the new 'MaxSize'.
                //
                Root -> MaxSize = (Space / sizeof(VALUE));

                //
                //  Call the extra constructors.
                //
                for ( Count=MaxSize;Count < Root -> MaxSize;Count ++ )
                    { PLACEMENT_NEW( & Queue[ Count ],VALUE ); }

                //
                //  When the 'Queue' has wrapped around we need to unwrap it.
                //
                if ( Root -> First > Root -> Last )
                    {
                    //
                    //  We copy the wrapped values to the correct place.
                    //
                    for ( Count=0;Count < Root -> Last;Count ++ )
                        { Queue[ (MaxSize + Count) ] = Queue[ Count ]; }

                    //
                    //  We zero values we have unwrapped.
                    //
                    ZeroValues( Queue,Root -> Last );

                    //
                    //  Now we need to update 'Last' as we have unwrapped the values.
                    //
                    Root -> Last += MaxSize;
                    }
                }
            else
                { return False; }
            }

        //
        //  We now merge the the 'PushCache' into he 'Queue' and extract the 'PopCache'
        //  from the 'Queue'.
        //
        if ( AutomaticHeapScope -> Update( Root -> Queue,((VOID**) & Queue) ) )
            {
            AUTO VALUE (*View)[ ViewSize ] = ((VALUE(*)[ ViewSize ]) Queue);
            AUTO FILE_SIZE Count;

            //
            //  We now merge the the 'PushCache' into he 'Queue'.
            //
            for ( Count=0;Count < Root -> PushCache.Used;Count ++ )
                {
                Queue[ Root -> Last ] = Root -> PushCache.Values[ Count ];

                Root -> Last = ((Root -> Last + 1) % Root -> MaxSize);
                }

            //
            //  We zero the 'PushCache' now it is empty.
            //
            ZeroCache( & Root -> PushCache );

            //
            //  We now need to recompute the current 'Size' of the 'Queue' as it has
            //  just changed.
            //
            if ( Root -> First <= Root -> Last )
                { Size = (Root -> Last - Root -> First); }
            else
                { Size = ((Root -> MaxSize - Root -> First) + (Root -> Last)); }

            //
            //  We will reduce 'Size' if it is greater than the free space in the
            //  'PopCache' to ensure we copy the correct amount from the 'Queue'.
            //
            if ( Size > (CacheSize - Root -> PopCache.Used) )
                { Size = (CacheSize - Root -> PopCache.Used); }

            //
            //  We will fill the 'PopCache' if there is space and the 'Queue' is
            //  not empty.
            //
            if ( Size > 0 )
                {
                //
                //  Move the values along in the 'PopCache' to make space for the values
                //  from the 'Queue'.
                //
                for ( Count=(Root -> PopCache.Used-1);Count >= 0;Count -- )
                    {  Root -> PopCache.Values[ (Size + Count) ] = Root -> PopCache.Values[ Count ]; }

                //
                //  Copy all the values from the 'Queue' into the 'PopCache'.
                //
                for ( Count=(Size-1);Count >= 0;Count -- )
                    {
                    AUTO VALUE *Current = & Queue[ Root -> First ];

                    Root -> First = ((Root -> First + 1) % Root -> MaxSize);
                    Root -> PopCache.Values[ Count ] = (*Current);

                    ZeroValues( Current,1 );
                    }

                //
                //  Update the sizes of the 'PopCache' cache.
                //
                Root -> PopCache.Used += ((SBIT32) Size);
                }

            return (Root -> PopCache.Used > 0);
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Peek Value.                                                                               */
            /*                                                                                            */
            /*  Peek the next value.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PeekValue( VALUE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PeekValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Peek Value.                                                                               */
            /*                                                                                            */
            /*  Peek the next value.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PeekValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()',
        //  'ExclusiveView()' and 'Update()' should be automatically optimized away for
        //  heaps that do not support this functionality by any decent compiler, due to
        //  the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'ExclusiveView()' the 'Root' as we need to view it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> ExclusiveView( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not empty or refill it.
                //
                if
                        (
                            (Root -> PopCache.Used > 0)
                                ||
                            (
                            (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                                &&
                            (MergeCaches( AutomaticHeapScope,Root ))
                            )
                        )
                    {
                    //
                    //  We peek at the top value in the 'PopCache' and return it.
                    //
                    (*Value) = Root -> PopCache.Values[ (Root -> PopCache.Used-1) ];

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorNoValueAvailable )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pop Value.                                                                                */
            /*                                                                                            */
            /*  Pop a value off the queue.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PopValue( VALUE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PopValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Pop Value.                                                                                */
            /*                                                                                            */
            /*  Pop a value off the queue.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PopValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not empty or refill it.
                //
                if ( (Root -> PopCache.Used > 0) || (MergeCaches( AutomaticHeapScope,Root )) )
                    {
                    AUTO VALUE *Current = & Root -> PopCache.Values[ (-- Root -> PopCache.Used) ];

                    //
                    //  We pop the top value on the 'PopCache' and return it.  We will also
                    //  zero the 'VALUE' just to be tidy.
                    //
                    (*Value) = (*Current);

                    ZeroValues( Current,1 );

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorNoValueAvailable )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Push Value.                                                                               */
            /*                                                                                            */
            /*  Push a value on the queue.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PushValue( CONST VALUE & Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (PushValue( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Push Value.                                                                               */
            /*                                                                                            */
            /*  Push a value on the queue.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::PushValue
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the queue.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Let's make sure the 'Cache' is not full or empty it.
                //
                if ( (Root -> PushCache.Used < CacheSize) || (MergeCaches( AutomaticHeapScope,Root )) )
                    {
                    //
                    //  We push the value on to the top of the 'Cache'.
                    //
                    Root -> PushCache.Values[ (Root -> PushCache.Used ++) ] = Value;

                    return True;
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToStoreValue )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the queue.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::Size( FILE_SIZE *Value )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Calculate the size of the queue.                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO FILE_SIZE Size = 0;

                //
                //  Compute the 'Size' of the 'Queue'.
                //
                if ( Root -> First <= Root -> Last )
                    { Size = (Root -> Last - Root -> First); }
                else
                    { Size = ((Root -> MaxSize - Root -> First) + (Root -> Last)); }

                //
                //  We also need to include the size of the caches.
                //
                (*Value) = (Size + Root -> PopCache.Used + Root -> PushCache.Used);

                return True;
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the queue.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::Walk
            (
            QUEUE_WALK_FUNCTION                 WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Walk( & AutomaticHeapScope,WalkFunction,UserValue ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk the queue.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> BOOLEAN QUEUE<HEAP,LOCK,VALUE>::Walk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            QUEUE_WALK_FUNCTION                 WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the queue.
            //
            if
                    (
                    (AutomaticViewScope1.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );

                //
                //  If the 'PopCache' is not empty then 'Walk()' all the values.
                //
                if ( Root -> PopCache.Used > 0 )
                    {
                    AUTO SBIT32 Count;

                    for ( Count=(Root -> PopCache.Used-1);Count >= 0;Count -- )
                        {
                        //
                        //  We simply don't trust user supplied code so we prepare for trouble.
                        //
#ifndef DISABLE_CRT_FEATURES
                        try
                            {
#endif
                            //
                            //  Execute the user supplied walk function.
                            //
                            if ( ! WalkFunction( Root -> PopCache.Values[ Count ],UserValue ) )
                                { return False; }
#ifndef DISABLE_CRT_FEATURES
                            }
                        catch ( ... )
                            {
                            //
                            //  We will catch and skip all errors detected in the user supplied
                            //  walk function.
                            //
                            return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                            }
#endif
                        }
                    }

                //
                //  If the 'MaxSize' is not zero then 'Walk()' all the values.
                //
                if ( Root -> MaxSize > 0 )
                    {
                    AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope ) ;
                    AUTO VALUE *Queue;

                    //
                    //  We will need to 'View()' the 'Queue'.
                    //
                    if ( AutomaticViewScope2.View( Root -> Queue,((VOID**) & Queue) ) )
                        {
                        AUTO FILE_SIZE Count;

                        //
                        //  Walk the 'Queue'.
                        //
                        for ( Count=Root -> First;Count != Root -> Last;Count=((Count + 1) %  Root -> MaxSize) )
                            {
                            //
                            //  We simply don't trust user supplied code so we prepare for trouble.
                            //
#ifndef DISABLE_CRT_FEATURES
                            try
                                {
#endif
                                //
                                //  Execute the user supplied walk function.
                                //
                                if ( ! WalkFunction( Queue[ Count ],UserValue ) )
                                    { return False; }
#ifndef DISABLE_CRT_FEATURES
                                }
                            catch ( ... )
                                {
                                //
                                //  We will catch and skip all errors detected in the user supplied
                                //  walk function.
                                //
                                return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                                }
#endif
                            }
                        }
                    }

                //
                //  If the 'PushCache' is not empty then 'Walk()' all the values.
                //
                if ( Root -> PushCache.Used > 0 )
                    {
                    AUTO SBIT32 Count;

                    for ( Count=0;Count < Root -> PushCache.Used;Count ++ )
                        {
                        //
                        //  We simply don't trust user supplied code so we prepare for trouble.
                        //
#ifndef DISABLE_CRT_FEATURES
                        try
                            {
#endif
                            //
                            //  Execute the user supplied walk function.
                            //
                            if ( ! WalkFunction( Root -> PushCache.Values[ Count ],UserValue ) )
                                { return False; }
#ifndef DISABLE_CRT_FEATURES
                            }
                        catch ( ... )
                            {
                            //
                            //  We will catch and skip all errors detected in the user supplied
                            //  walk function.
                            //
                            return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                            }
#endif
                        }
                    }

                return True;
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Values.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID QUEUE<HEAP,LOCK,VALUE>::ZeroCache( CACHE *Cache )
        {
        //
        //  We have to be careful as a 'VALUE' contains user types.  It is very hard to zero
        //  these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        ZeroValues( Cache -> Values,Cache -> Used );

        Cache -> Used = 0;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Values.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of values.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> VOID QUEUE<HEAP,LOCK,VALUE>::ZeroValues
            (
            VALUE                               *Values,
            FILE_SIZE                           Size
            )
        {
        AUTO FILE_SIZE Count;

        //
        //  We have to be careful as a 'VALUE' contains user types.  It is very hard to zero
        //  these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Values[ Count ],VALUE ); }

        MemoryFunctions.Zero( Values,((SNATIVE) (Size * sizeof(VALUE))) );

        for ( Count=0;Count < Size;Count ++ )
            { PLACEMENT_NEW( & Values[ Count ],VALUE ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'QUEUE'.                                                                   */
            /*                                                                                            */
            /*  The destructor for 'QUEUE' cleans up the class.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class VALUE> QUEUE<HEAP,LOCK,VALUE>::~QUEUE( VOID )
        {
        //
        //  We need to ensure that the class is 'Active'.
        //
        if ( Active )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( & AutomaticHeapScope );
            AUTO TRANSACTIONAL_LOCK TransactionLock( & AutomaticHeapScope );

            //
            //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
            //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
            //  should be automatically optimized away for heaps that do not support this
            //  functionality by any decent compiler, due to the way the code has been written.
            //
            if
                    (
                    (Active)
                        &&
                    (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO ROOT *Root;

                //
                //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
                //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
                //  above to prevent threads in the same transaction from accessing the queue.
                //
                if
                        (
                        (AutomaticHeapScope.Update( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );

                    //
                    //  We 'DeleteAll()' 'Stack' values.
                    //
                    if ( DeleteAll( & AutomaticHeapScope,Root ) )
                        {
                        //
                        //  Zero the 'Root' structure.
                        //
                        Root -> Signature = 0;
                        Root -> Queue = NULL;
                        Root -> MaxSize = 0;
                        Root -> Last = 0;
                        Root -> First = 0;

                        //
                        //  We are no longer 'Active'.
                        //
                        Active = False;

                        //
                        //  Release the lock on the 'Root'.
                        //
                        AutomaticExclusiveLock2.Release();

                        //
                        //  Delete the 'Root' structure.
                        //
                        if ( ROCKALL_DELETE<ROOT>::Delete( & AutomaticHeapScope,Address ) )
                            {
                            //
                            //  Zero class members.
                            //
                            Address = NULL;
                            Active = False;
                            }
                        else
                            { FAILURE( "Unable to delete the root" ); }
                        }
                    else
                        { FAILURE( "Unable to delete the root structure" ); }
                    }
                else
                    { FAILURE( "Unable to update the root structure" ); }
                }
            else
                { FAILURE( "Unable to update the root structure" ); }
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define QUEUE ROCKALL::QUEUE
#endif
#endif