#ifndef _MEMORY_FUNCTIONS_HPP_
#define _MEMORY_FUNCTIONS_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "CompilerFlags.hpp"
#include "PublicConstants.hpp"
#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Memory Functions.                                                                                 */
        /*                                                                                                    */
        /*  A significant amount of effort has been undertaken to remove all external dependancies to         */
        /*  make the code highly portable.  This has lead to the reinvention of of basic compare and          */
        /*  copy related functionality within this module.                                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Memory Functions.                                                                         */
            /*                                                                                            */
            /*  There is no real need for a class for the generic block compare and copy related          */
            /*  functionality.  Nonetheless, we include one just for consistency as there are no          */
            /*  significant performance costs.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class MEMORY_FUNCTIONS
        {
            //
            //  Self referencing friend class.
            //
            friend class MEMORY_FUNCTIONS;

        public:
            //
            //  Public functions.
            //
            MEMORY_FUNCTIONS( VOID ) { /* void */ }

            STATIC SNATIVE Compare( VOID *First,VOID *Second,SNATIVE Size );

            STATIC VOID Copy( VOID *Source,VOID *Destination,SNATIVE Size );

            STATIC SBIT32 QuickChecksum( VOID *Source,SNATIVE Size );

            STATIC SBIT32 SlowChecksum( VOID *Source,SNATIVE Size );

            STATIC BOOLEAN Verify( VOID *Source,SNATIVE Size,SNATIVE Value = 0 );

            STATIC VOID Zero( VOID *Destination,SNATIVE Size,SNATIVE Value = 0 );

            ~MEMORY_FUNCTIONS( VOID ) { /* void */ }

        private:
            //
            //  Private functions.
            //
            STATIC SNATIVE PrivateCompare( TYPE *First,TYPE *Second,SNATIVE Size );

            STATIC VOID PrivateCopy( TYPE *Source,TYPE *Destination,SNATIVE Size );

            STATIC SBIT32 PrivateQuickChecksum( TYPE *Source,SNATIVE Size );

            STATIC SBIT32 PrivateSlowChecksum( TYPE *Source,SNATIVE Size );

            STATIC BOOLEAN PrivateVerify( TYPE *Source,SNATIVE Size,TYPE Value );

            STATIC VOID PrivateZero( TYPE *Destination,SNATIVE Size,TYPE Value );

            //
            //  Disabled functionality.
            //
            MEMORY_FUNCTIONS( CONST MEMORY_FUNCTIONS & NoCopies );

            VOID operator=( CONST MEMORY_FUNCTIONS & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Compare.                                                                          */
            /*                                                                                            */
            /*  A raw compare does a straight compare of a series of values based on the supplied         */
            /*  size regardless of any zero terminators.                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE MEMORY_FUNCTIONS<TYPE>::PrivateCompare( TYPE *First,TYPE *Second,SNATIVE Size )
        {
        //
        //  A simple fast compare.  Most compilers manage to optimise the following into a loop with 5 or 6
        //  instructions.  In some cases this proves to be faster than hardware instructions.
        //
        for ( /* void */;Size > 0;Size -= sizeof(TYPE) )
            {
            REGISTER TYPE Current1 = (*(First ++));
            REGISTER TYPE Current2 = (*(Second ++));

            //
            //  We compare each value and exit if there is no match.
            //
            if ( Current1 != Current2 )
                { return ((Current1 - Current2) > 0 ? 1 : -1); }
            }

        return 0;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  A raw compare does a straight compare of a series of values based on the supplied         */
            /*  size regardless of any zero terminators.                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE MEMORY_FUNCTIONS<TYPE>::Compare( VOID *First,VOID *Second,SNATIVE Size )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) First) | ((SNATIVE) Second) | Size) & 0x7);

        //
        //  Automatically select a copy size based on the aligment of the compare data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                return (MemoryFunctions.PrivateCompare( ((SBIT64*) First),((SBIT64*) Second),Size ));
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                return (MemoryFunctions.PrivateCompare( ((SBIT32*) First),((SBIT32*) Second),Size ));
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                return (MemoryFunctions.PrivateCompare( ((SBIT16*) First),((SBIT16*) Second),Size ));
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                return (MemoryFunctions.PrivateCompare( ((SBIT8*) First),((SBIT8*) Second),Size ));
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Copy.                                                                             */
            /*                                                                                            */
            /*  A raw copy does a straight copy of a series of values based on the supplied size          */
            /*  regardless of any zero terminators.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> VOID MEMORY_FUNCTIONS<TYPE>::PrivateCopy( TYPE *Source,TYPE *Destination,SNATIVE Size )
        {
        //
        //  A simple fast copy.  Most compilers manage to optimise the following into a loop with 3 or 4
        //  instructions.  In some cases this proves to be faster than hardware instructions.
        //
        for ( /* void */;Size > 0;Size -= sizeof(TYPE) )
            { (*(Destination ++)) = (*(Source ++)); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A raw copy does a straight copy of a series of values based on the supplied size          */
            /*  regardless of any zero terminators.                                                       */
            /*                                                                                            */
            /**********************************************************************************************/


    template <class TYPE> VOID MEMORY_FUNCTIONS<TYPE>::Copy
            (
            VOID                                *Source,
            VOID                                *Destination,
            SNATIVE                             Size
            )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) Source) | ((SNATIVE) Destination) | Size) & 0x7);

        //
        //  Automatically select a copy size based on the aligment of the copy data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                MemoryFunctions.PrivateCopy( ((SBIT64*) Source),((SBIT64*) Destination),Size );

                return;
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                MemoryFunctions.PrivateCopy( ((SBIT32*) Source),((SBIT32*) Destination),Size );

                return;
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                MemoryFunctions.PrivateCopy( ((SBIT16*) Source),((SBIT16*) Destination),Size );

                return;
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                MemoryFunctions.PrivateCopy( ((SBIT8*) Source),((SBIT8*) Destination),Size );

                return;
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Quick Checksum.                                                                   */
            /*                                                                                            */
            /*  A quick checksum is a fast and dirty checksum which is better than nothing but not        */
            /*  a lot better.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SBIT32 MEMORY_FUNCTIONS<TYPE>::PrivateQuickChecksum( TYPE *Source,SNATIVE Size )
        {
        REGISTER SBIT32 HashValue = 0;
        REGISTER SNATIVE Count;

        //
        //  We compute the hash value using every byte in the area to ensure a smooth distribution even
        //  when the values are very similar.
        //
        for ( Count=0;Count < Size;Count += sizeof(TYPE) )
            {
            //
            //  Compute the hash value for the current value and add it the the current total.
            //
            HashValue = ((SBIT32) (((HashValue | 1) + (*(Source ++))) * PrimeSBit32));
            HashValue &= HashMaskSBit32;
            }

        return HashValue;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Quick Checksum.                                                                           */
            /*                                                                                            */
            /*  A quick checksum is a fast and dirty checksum which is better than nothing but not        */
            /*  a lot better.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SBIT32 MEMORY_FUNCTIONS<TYPE>::QuickChecksum( VOID *Source,SNATIVE Size )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) Source) | Size) & 0x7);

        //
        //  Automatically select a checksum size based on the aligment of the source data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                return(MemoryFunctions.PrivateQuickChecksum( ((SBIT64*) Source),Size ));
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                return(MemoryFunctions.PrivateQuickChecksum( ((SBIT32*) Source),Size ));
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                return(MemoryFunctions.PrivateQuickChecksum( ((SBIT16*) Source),Size ));
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                return(MemoryFunctions.PrivateQuickChecksum( ((SBIT8*) Source),Size ));
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Slow Checksum.                                                                    */
            /*                                                                                            */
            /*  A slow checksum is a time consuming but gives a good evenly spread result.                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SBIT32 MEMORY_FUNCTIONS<TYPE>::PrivateSlowChecksum( TYPE *Source,SNATIVE Size )
        {
        REGISTER SBIT32 HashValue = 0;
        REGISTER SNATIVE Count;

        //
        //  We compute the hash value using every byte in the area to ensure a smooth distribution even
        //  when the values are very similar.
        //
        for ( Count=0;Count < Size;Count += sizeof(TYPE) )
            {
            //
            //  Compute the hash value for the current value and add it the the current total.
            //
            HashValue = ((SBIT32) (((HashValue | 1) + (*(Source ++))) * PrimeSBit32));
            HashValue = ((HashValue >> 16) + (HashValue << 16));
            HashValue &= HashMaskSBit32;
            }

        return HashValue;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Slow Checksum.                                                                            */
            /*                                                                                            */
            /*  A slow checksum is a time consuming but gives a good evenly spread result.                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SBIT32 MEMORY_FUNCTIONS<TYPE>::SlowChecksum( VOID *Source,SNATIVE Size )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) Source) | Size) & 0x7);

        //
        //  Automatically select a checksum size based on the aligment of the source data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                return (MemoryFunctions.PrivateSlowChecksum( ((SBIT64*) Source),Size ));
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                return (MemoryFunctions.PrivateSlowChecksum( ((SBIT32*) Source),Size ));
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                return (MemoryFunctions.PrivateSlowChecksum( ((SBIT16*) Source),Size ));
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                return (MemoryFunctions.PrivateSlowChecksum( ((SBIT8*) Source),Size ));
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Verify.                                                                           */
            /*                                                                                            */
            /*  A verify is a straight check that a series of values are zero (or any supplied value).    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN MEMORY_FUNCTIONS<TYPE>::PrivateVerify( TYPE *Source,SNATIVE Size,TYPE Value )
        {
        //
        //  Verify a series of values.
        //
        for ( /* void */;Size > 0;Size -= sizeof(TYPE) )
            {
            if ( (*(Source ++)) != Value )
                { return False; }
            }

        return True;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Verify.                                                                                   */
            /*                                                                                            */
            /*  A verify is a straight check that a series of values are zero (or any supplied value).    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN MEMORY_FUNCTIONS<TYPE>::Verify( VOID *Source,SNATIVE Size,SNATIVE Value )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) Source) | Size) & 0x7);

        //
        //  Automatically select a verify size based on the aligment of the copy data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                return (MemoryFunctions.PrivateVerify( ((SBIT64*) Source),Size,((SBIT64) Value) ));
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                return (MemoryFunctions.PrivateVerify( ((SBIT32*) Source),Size,((SBIT32) Value) ));
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                return (MemoryFunctions.PrivateVerify( ((SBIT16*) Source),Size,((SBIT16) Value) ));
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                return (MemoryFunctions.PrivateVerify( ((SBIT8*) Source),Size,((SBIT8) Value) ));
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Private Zero.                                                                             */
            /*                                                                                            */
            /*  A straight overwrite with zeros (or any supplied value).                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> VOID MEMORY_FUNCTIONS<TYPE>::PrivateZero( TYPE *Destination,SNATIVE Size,TYPE Value )
        {
        //
        //  Zero a series of values.
        //
        for ( /* void */;Size > 0;Size -= sizeof(TYPE) )
            { (*(Destination ++)) = Value; }
        }


            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero.                                                                                     */
            /*                                                                                            */
            /*  A straight overwrite with zeros (or any supplied value).                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> VOID MEMORY_FUNCTIONS<TYPE>::Zero( VOID *Destination,SNATIVE Size,SNATIVE Value )
        {
        REGISTER SNATIVE Alignment = ((((SNATIVE) Destination) | Size) & 0x7);

        //
        //  Automatically select a zero size based on the aligment of the copy data and the size.
        //
        switch ( Alignment )
            {
            case 0:
#ifdef PLATFORM_64BIT
                {
                AUTO MEMORY_FUNCTIONS<SBIT64> MemoryFunctions;

                MemoryFunctions.PrivateZero( ((SBIT64*) Destination),Size,((SBIT64) Value) );

                return;
                }

#endif
            case 4:
                {
                AUTO MEMORY_FUNCTIONS<SBIT32> MemoryFunctions;

                MemoryFunctions.PrivateZero( ((SBIT32*) Destination),Size,((SBIT32) Value) );

                return;
                }

            case 6:
            case 2:
                {
                AUTO MEMORY_FUNCTIONS<SBIT16> MemoryFunctions;

                MemoryFunctions.PrivateZero( ((SBIT16*) Destination),Size,((SBIT16) Value) );

                return;
                }

            default:
                {
                AUTO MEMORY_FUNCTIONS<SBIT8> MemoryFunctions;

                MemoryFunctions.PrivateZero( ((SBIT8*) Destination),Size,((SBIT8) Value) );

                return;
                }
            }
        }
    }
#endif