#ifndef _STRING_FUNCTIONS_HPP_
#define _STRING_FUNCTIONS_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "CompilerFlags.hpp"
#include "PublicConstants.hpp"
#include "MemoryFunctions.hpp"
#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  String Functions.                                                                                 */
        /*                                                                                                    */
        /*  A significant amount of effort has been undertaken to remove all external dependancies to         */
        /*  make the code highly portable.  This has lead to the reinvention of of basic compare and          */
        /*  copy related functionality within this module.                                                    */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  String Functions.                                                                         */
            /*                                                                                            */
            /*  There is no real need for a class for the generic block compare and copy related          */
            /*  functionality.  Nonetheless, we include one just for consistency as there are no          */
            /*  significant performance costs.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> class STRING_FUNCTIONS
        {
            //
            //  Self referencing friend class.
            //
            friend class STRING_FUNCTIONS;

            //
            //  Private constants.
            //
            STATIC CONST SNATIVE MaxCompareSize = (1 << 30);
            STATIC CONST SBIT32 MaxDigits       = 64;

        public:
            //
            //  Public functions.
            //
            STRING_FUNCTIONS( VOID ) { /* void */ }

            STATIC SNATIVE Compare( TYPE *First,TYPE *Second,SNATIVE Size = MaxCompareSize );

            STATIC BOOLEAN ConvertIntegerToText( SNATIVE Value,TYPE **Text );

            STATIC SNATIVE Copy( TYPE *Source,TYPE *Destination,SNATIVE MaxSize );

            STATIC SNATIVE Copy( TYPE *Strings[],TYPE *String,SNATIVE MaxSize );

            STATIC BOOLEAN FindFirst
                (
                TYPE                            *Find,
                TYPE                            *String,
                TYPE                            **Found,
                SNATIVE                         Size = MaxCompareSize,
                BOOLEAN                         Skip = True
                );

            STATIC BOOLEAN FindLast
                (
                TYPE                            *Find,
                TYPE                            *String,
                TYPE                            **Found,
                SNATIVE                         Size = MaxCompareSize,
                BOOLEAN                         Skip = True
                );

            STATIC SNATIVE Length( TYPE *String );

            STATIC SNATIVE Length( TYPE *Strings[] );

            ~STRING_FUNCTIONS( VOID ) { /* void */ }

        private:
            //
            //  Disabled functionality.
            //
            STRING_FUNCTIONS( CONST STRING_FUNCTIONS & NoCopies );

            VOID operator=( CONST STRING_FUNCTIONS & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Compare.                                                                                  */
            /*                                                                                            */
            /*  A compare does a straight compare of a series of values based on the supplied size        */
            /*  while watching for zero terminators.                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE STRING_FUNCTIONS<TYPE>::Compare( TYPE *First,TYPE *Second,SNATIVE Size )
        {
        //
        //  A simple fast compare.  Most compilers manage to optimise the following into a loop with 5 or 6
        //  instructions.  In some cases this proves to be faster than hardware instructions.
        //

        for ( /* void */;Size > 0;Size -= sizeof(TYPE) )
            {
            REGISTER TYPE Current1 = (*(First ++));
            REGISTER TYPE Current2 = (*(Second ++));

            //
            //  We compare each value and exit if there is no match or we find a zero terminator.
            //
            if ( (Current1 == NULL) || (Current1 != Current2) )
                { return (Current2 - Current1); }
            }

        return 0;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Convert an Integer to Text.                                                               */
            /*                                                                                            */
            /*  An integer is converted to a array of characters with an optional leading sign.           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN STRING_FUNCTIONS<TYPE>::ConvertIntegerToText( SNATIVE Value,TYPE **Text )
        {
        REGISTER BOOLEAN Negative = (Value < 0);
        REGISTER SBIT32 Count = (MaxDigits-3);
        STATIC TYPE Digits[ MaxDigits ];

        //
        //  We initially write '0' to the digit buffer.
        //
        Digits[ (MaxDigits-1) ] = NULL;
        Digits[ (MaxDigits-2) ] = ((TYPE) L'0');

        //
        //  If the value is not zero we convert the digits into a string.
        //
        if ( Value != 0 )
            {
            //
            //  We convert each digit and place it in the digit buffer as long as there
            //  is enough space.
            //
            for
                    (
                    Count = (MaxDigits-2), Value = ((Negative) ? -Value : Value);
                    (Count > 0) && (Value > 0);
                    Count --, Value /= 10
                    )
                { Digits[ Count ] = ((TYPE) (L'0' + (Value % 10))); }
            }

        //
        //  When a value is negative we add leading sign to the digit buffer before all
        //  the digits.
        //
        if ( Negative )
            {
            //
            //  We add a negative sign if there is space or exit.
            //
            if ( Count > 0 )
                { Digits[ (Count --) ] = ((TYPE) L'-'); }
            else
                { return False; }
            }

        //
        //  We now do a final check just to make sure that everything worked and there were
        //  no serious problems.
        //
        if ( Value == 0 )
            {
            //
            //  All seems well so return the digit buffer.
            //
            (*Text) = & Digits[ (Count + 1) ];

            return True;
            }
        else
            { return False; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A copy does a straight copy of a series of values based on the supplied size while        */
            /*  watching for zero terminators.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE STRING_FUNCTIONS<TYPE>::Copy( TYPE *Source,TYPE *Destination,SNATIVE MaxSize )
        {
        REGISTER SNATIVE Size = (MaxSize - sizeof(TYPE));
        REGISTER SNATIVE Count;

        //
        //  We copy each value while watching for a zero terminator.
        //
        for ( Count=0;(Count < Size) && (*(Source) != NULL);Count += sizeof(TYPE) )
            { (*(Destination ++)) = (*(Source ++)); }

        //
        //  Regardless add a zero terminator.
        //
        (*Destination) = NULL;

        return ((Count + 1) * sizeof(TYPE));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Copy.                                                                                     */
            /*                                                                                            */
            /*  A copy does a straight copy of a series of values based on the supplied size while        */
            /*  watching for zero terminators.                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE STRING_FUNCTIONS<TYPE>::Copy( TYPE *Strings[],TYPE *String,SNATIVE MaxSize )
        {
        REGISTER SNATIVE Count;
        REGISTER SNATIVE Total;

        //
        //  We copy each of the series of values in the array into the target until we find a zero
        //  terminator.  If the target becomes full we cease copying any further values and exit.
        //
        for ( Count=0, Total=0;(Strings[ Count ] != NULL) && (Total < MaxSize);Count ++ )
            {
            //
            //  Copy each series of values in turn and add the length the the current total.
            //
            Total += (Copy( Strings[ Count ],& String[ Total ],(MaxSize - Total) ));

            //
            //  We remove one for each zero terminator copied and add it again at the end.
            //
            Total -= sizeof(TYPE);
            }

        //
        //  We add one for the final zero terminator that has been added to the values copied.
        //
        return ((Total + 1) * sizeof(TYPE));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find First.                                                                               */
            /*                                                                                            */
            /*  A find first finds the first occurance of a sub-string in a larger string.                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN STRING_FUNCTIONS<TYPE>::FindFirst
            (
            TYPE                                *Find,
            TYPE                                *String,
            TYPE                                **Found,
            SNATIVE                             Size,
            BOOLEAN                             Skip
            )
        {
        REGISTER SNATIVE Length1 = (Length( Find ));
        REGISTER SNATIVE Length2 = ((MaxCompareSize != Size) ? (Size / sizeof(TYPE)) : Length( String ));

        //
        //  Let's just make sure that it is logically possible to find the sub-string.
        //
        if ( (Length1 > 0) && (Length1 <= Length2) )
            {
            AUTO MEMORY_FUNCTIONS<TYPE> MemoryFunctions;
            REGISTER SNATIVE Length3 = (Length2 - Length1);
            REGISTER SNATIVE Count;

            //
            //  We try to find a match a character at a time.
            //
            for ( Count=0;Count <= Length3;Count ++ )
                {
                REGISTER TYPE *Current = & String[ Count ];

                //
                //  If we find a match we exit and return the match point.
                //
                if ( MemoryFunctions.Compare( Current,Find,Length1 ) )
                    {
                    REGISTER SNATIVE SkipSize = ((Skip) ? Length1 : 0);

                    //
                    //  The string we return either includes or skips the search string.
                    //
                    (*Found) = & String[ (Count + SkipSize) ];

                    return True;
                    }
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Last.                                                                                */
            /*                                                                                            */
            /*  A find first finds the last occurance of a sub-string in a larger string.                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> BOOLEAN STRING_FUNCTIONS<TYPE>::FindLast
            (
            TYPE                                *Find,
            TYPE                                *String,
            TYPE                                **Found,
            SNATIVE                             Size,
            BOOLEAN                             Skip
            )
        {
        REGISTER SNATIVE Length1 = (Length( Find ));
        REGISTER SNATIVE Length2 = ((MaxCompareSize != Size) ? (Size / sizeof(TYPE)) : Length( String ));

        //
        //  Let's just make sure that it is logically possible to find the sub-string.
        //
        if ( (Length1 > 0) && (Length1 <= Length2) )
            {
            AUTO MEMORY_FUNCTIONS<TYPE> MemoryFunctions;
            REGISTER SNATIVE Length3 = (Length2 - Length1);
            REGISTER SNATIVE Count;

            //
            //  We try to find a match a character at a time.
            //
            for ( Count=Length3;Count >= 0;Count -- )
                {
                REGISTER TYPE *Current = & String[ Count ];

                //
                //  If we find a match we exit and return the match point.
                //
                if ( MemoryFunctions.Compare( Current,Find,Length1 ) )
                    {
                    REGISTER SNATIVE SkipSize = ((Skip) ? Length1 : 0);

                    //
                    //  The string we return either includes or skips the search string.
                    //
                    (*Found) = & String[ (Count + SkipSize) ];

                    return True;
                    }
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Length.                                                                                   */
            /*                                                                                            */
            /*  Find the length of a string by finding the zero terminator.                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE STRING_FUNCTIONS<TYPE>::Length( TYPE *String )
        {
        REGISTER SNATIVE Count = 0;

        //
        //  We skip null valued strings.
        //
        if ( String != NULL )
            {
            //
            //  Find the zero terminator.
            //
            for ( /* void */;(*(String ++)) != NULL;Count ++ )
                { /* void */ }
            }

        return Count;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Length.                                                                                   */
            /*                                                                                            */
            /*  Find the length of an array of strings by finding the zero terminators and combining      */
            /*  the lengths.                                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE> SNATIVE STRING_FUNCTIONS<TYPE>::Length( TYPE *Strings[] )
        {
        REGISTER SNATIVE Total = 0;

        //
        //  We skip null valued strings.
        //
        if ( Strings != NULL )
            {
            REGISTER SNATIVE Count;

            //
            //  Calculate the total length of all the series of values.
            //
            for ( Count=0;Strings[ Count ] != NULL;Count ++ )
                { Total += Length( Strings[ Count ] ); }
            }

        return Total;
        }
    }
#endif