#ifndef _AUTOMATIC_TRANSACTION_SCOPE_HPP_
#define _AUTOMATIC_TRANSACTION_SCOPE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "BaseTypes.hpp"
#include "ExportMemoryLock.hpp"
#include "Failure.hpp"
#include "ReservedWords.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Automatic Transaction Scope.                                                                      */
        /*                                                                                                    */
        /*  An automatic transaction scope ensures that a transaction is correctly terminated at the          */
        /*  end of a block.                                                                               */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic Transaction Scope.                                                              */
            /*                                                                                            */
            /*  An automatic transaction scope ensures that a transaction is correctly terminated         */
            /*  at the end of a block.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> class AUTOMATIC_TRANSACTION_SCOPE
        {
            //
            //  Private data.
            //
            BOOLEAN                             Active;
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope;
            HEAP                                *Heap;

        public:
            //
            //  Public functions.
            //
            AUTOMATIC_TRANSACTION_SCOPE( AUTOMATIC_HEAP_SCOPE<HEAP> *NewAutomaticHeapScope );
            AUTOMATIC_TRANSACTION_SCOPE( HEAP *NewHeap );

            BOOLEAN BeginTransaction( FILE_ADDRESS *NewTransactionID = NULL,EXPORT_MEMORY_LOCK *UserLock = NULL );

            BOOLEAN EndTransaction( BOOLEAN Abort = False,BOOLEAN Commit = False );

            ~AUTOMATIC_TRANSACTION_SCOPE( VOID );

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_TRANSACTION_SCOPE( CONST AUTOMATIC_TRANSACTION_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_TRANSACTION_SCOPE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_TRANSACTION_SCOPE'.                                            */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_TRANSACTION_SCOPE' prepares the class for use.             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_TRANSACTION_SCOPE<HEAP>::AUTOMATIC_TRANSACTION_SCOPE
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *NewAutomaticHeapScope
            )
        {
        AUTO HEAP *NewHeap = (NewAutomaticHeapScope -> GetHeap());

        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( (Heap = NewHeap) -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  Setup the class members.
            //
            Active = False;
            AutomaticHeapScope = NewAutomaticHeapScope;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_TRANSACTION_SCOPE'.                                            */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_TRANSACTION_SCOPE' prepares the class for active use.      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_TRANSACTION_SCOPE<HEAP>::AUTOMATIC_TRANSACTION_SCOPE
            (
            HEAP                                *NewHeap
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( (Heap = NewHeap) -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  Setup the class members.
            //
            Active = False;
            AutomaticHeapScope = NULL;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Begin Transaction.                                                                        */
            /*                                                                                            */
            /*  Begin a transaction.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_TRANSACTION_SCOPE<HEAP>::BeginTransaction
            (
            FILE_ADDRESS                        *NewTransactionID,
            EXPORT_MEMORY_LOCK                  *UserLock
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will 'BeginTransaction()' if we have not done so already.
            //
            if ( ! Active )
                {
                //
                //  If we don't have an 'AutomaticHeapScope' we will use the 'Heap'.
                //
                if ( AutomaticHeapScope != NULL )
                    { return (Active = (AutomaticHeapScope -> BeginTransaction( NewTransactionID,UserLock ))); }
                else
                    { return (Active = (Heap -> BeginTransaction( NewTransactionID,UserLock ))); }
                }
            else
                { return (Heap -> SetError( ErrorExistingTransaction )); }
            }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  End Transaction.                                                                          */
            /*                                                                                            */
            /*  End a transaction.                                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_TRANSACTION_SCOPE<HEAP>::EndTransaction
            (
            BOOLEAN                             Abort,
            BOOLEAN                             Commit
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We will 'EndTransaction()' if we have not done so already.
            //
            if ( Active )
                {
                //
                //  If we don't have an 'AutomaticHeapScope' we will use the 'Heap'.
                //
                if ( AutomaticHeapScope != NULL )
                    {
                    if ( AutomaticHeapScope -> EndTransaction( Abort,Commit ) )
                        { return ((Active = False), True); }
                    }
                else
                    {
                    if ( Heap -> EndTransaction( Abort,Commit ) )
                        { return ((Active = False), True); }
                    }

                return False;
                }
            else
                { return (Heap -> SetError( ErrorNoTransaction )); }
            }
        else
            { return True; }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'AUTOMATIC_TRANSACTION_SCOPE'.                                             */
            /*                                                                                            */
            /*  The destructor for 'AUTOMATIC_TRANSACTION_SCOPE' cleans up the class.                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_TRANSACTION_SCOPE<HEAP>::~AUTOMATIC_TRANSACTION_SCOPE( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We end any outstanding transaction.
            //
            if ( (Active) && (! EndTransaction()) )
                { FAILURE( "Unable to end a transaction" ); }

            //
            //  Zero the class members.
            //
            Heap = NULL;
            AutomaticHeapScope = NULL;
            Active = False;
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define AUTOMATIC_TRANSACTION_SCOPE ROCKALL::AUTOMATIC_TRANSACTION_SCOPE
#endif
#endif