#ifndef _AUTOMATIC_VIEW_SCOPE_HPP_
#define _AUTOMATIC_VIEW_SCOPE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "BaseTypes.hpp"
#include "Failure.hpp"
#include "ReservedWords.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Automatic View Scope.                                                                             */
        /*                                                                                                    */
        /*  An automatic view scope ensures that a view on a heap is correctly terminated at the end          */
        /*  of a block.                                                                                       */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Automatic View Scope.                                                                     */
            /*                                                                                            */
            /*  An automatic view scope ensures that a view on a heap is correctly terminated at          */
            /*  the end of a block.                                                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> class AUTOMATIC_VIEW_SCOPE
        {
            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        EndAddress;
            BOOLEAN                             Exclusive;
            HEAP                                *Heap;

        public:
            //
            //  Public functions.
            //
            AUTOMATIC_VIEW_SCOPE( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );
            AUTOMATIC_VIEW_SCOPE( HEAP *NewHeap );

            BOOLEAN ExclusiveView( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL );

            VOID Release( VOID );

            BOOLEAN View( FILE_ADDRESS Address,VOID **Data,FILE_SIZE *Space = NULL );

            ~AUTOMATIC_VIEW_SCOPE( VOID );

        private:
            //
            //  Disabled functionality.
            //
            AUTOMATIC_VIEW_SCOPE( CONST AUTOMATIC_VIEW_SCOPE & NoCopies );

            VOID operator=( CONST AUTOMATIC_VIEW_SCOPE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_VIEW_SCOPE'.                                                   */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_VIEW_SCOPE' prepares the class for use.                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_VIEW_SCOPE<HEAP>::AUTOMATIC_VIEW_SCOPE
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO HEAP *NewHeap = (AutomaticHeapScope -> GetHeap());

        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( (Heap = NewHeap) -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  Setup the class members.
            //
            Active = False;
            EndAddress = NULL;
            Exclusive = False;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'AUTOMATIC_VIEW_SCOPE'.                                                   */
            /*                                                                                            */
            /*  The constructor for 'AUTOMATIC_VIEW_SCOPE' prepares the class for use.                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_VIEW_SCOPE<HEAP>::AUTOMATIC_VIEW_SCOPE
            (
            HEAP                                *NewHeap
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( (Heap = NewHeap) -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  Setup the class members.
            //
            Active = False;
            EndAddress = NULL;
            Exclusive = False;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Exclusive View.                                                                           */
            /*                                                                                            */
            /*  An exclusive view of an area.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_VIEW_SCOPE<HEAP>::ExclusiveView
            (
            FILE_ADDRESS                        Address,
            VOID                                **Data,
            FILE_SIZE                           *Space
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We now claim an 'ExclusiveView()'.
            //
            if ( (! Active) && (Heap -> ExclusiveView( Address,Data,Space )) )
                {
                Active = True;
                EndAddress = Address;
                Exclusive = True;

                return True;
                }
            }
        else
            {
            //
            //  When the 'Heap' does not support transactions we may need to call 'ActuualSize()'
            //  if we need 'Space'.  Regardless, we need to cast 'Address' to 'Data' as they are
            //  known to be the same in this case.
            //
            if ( (Space == NULL) || (Heap -> Details( Address,NULL,Space )) )
                {
                if ( Data != NULL )
                    { (*Data) = ((VOID*) Address); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Release.                                                                                  */
            /*                                                                                            */
            /*  Release a view.                                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> VOID AUTOMATIC_VIEW_SCOPE<HEAP>::Release( VOID )
        {
        //
        //  We do not need to do anything unless we are 'Active' and have the related features
        //  in the heap.
        //
        if ( (Active) && (Heap -> Feature( RockallTransactionalSupport )) )
            {
            //
            //  We release the appropriate type of view.
            //
            if ( Exclusive )
                {
                //
                //  We end an 'ExclusiveView()'.
                //
                if ( Heap -> EndExclusiveView( EndAddress,False ) )
                    {
                    Active = False;
                    EndAddress = NULL;
                    Exclusive = False;
                    }
                else
                    { FAILURE( "Unable to end exclusive view" ); }
                }
            else
                {
                //
                //  We end a 'View()'.
                //
                if ( Heap -> EndView( EndAddress,False ) )
                    {
                    Active = False;
                    EndAddress = NULL;
                    Exclusive = False;
                    }
                else
                    { FAILURE( "Unable to end view" ); }
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  View.                                                                                     */
            /*                                                                                            */
            /*  View an area.                                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> BOOLEAN AUTOMATIC_VIEW_SCOPE<HEAP>::View
            (
            FILE_ADDRESS                        Address,
            VOID                                **Data,
            FILE_SIZE                           *Space
            )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We now claim a 'View()'.
            //
            if ( (! Active) && (Heap -> View( Address,Data,Space )) )
                {
                Active = True;
                EndAddress = Address;
                Exclusive = False;

                return True;
                }
            }
        else
            {
            //
            //  When the 'Heap' does not support transactions we may need to call 'ActuualSize()'
            //  if we need 'Space'.  Regardless, we need to cast 'Address' to 'Data' as they are
            //  known to be the same in this case.
            //
            if ( (Space == NULL) || (Heap -> Details( Address,NULL,Space )) )
                {
                if ( Data != NULL )
                    { (*Data) = ((VOID*) Address); }

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'AUTOMATIC_VIEW_SCOPE'.                                                    */
            /*                                                                                            */
            /*  The destructor for 'AUTOMATIC_VIEW_SCOPE' cleans up the class.                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP> AUTOMATIC_VIEW_SCOPE<HEAP>::~AUTOMATIC_VIEW_SCOPE( VOID )
        {
        //
        //  We do not need to do anything unless we have the related features in the heap.
        //
        if ( Heap -> Feature( RockallTransactionalSupport ) )
            {
            //
            //  We 'Release()' any outstanding 'ExclusiveView()' or 'View()'.
            //
            Release();

            //
            //  Zero the class members.
            //
            Heap = NULL;
            Exclusive = False;
            EndAddress = NULL;
            Active = False;
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define AUTOMATIC_VIEW_SCOPE ROCKALL::AUTOMATIC_VIEW_SCOPE
#endif
#endif