#ifndef _LINKED_LIST_HPP_
#define _LINKED_LIST_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "BaseTypes.hpp"
#include "CompilerFlags.hpp"
#include "Failure.hpp"
#include "NoClass.hpp"
#include "PublicConstants.hpp"
#include "ReservedWords.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Linked List.                                                                                      */
        /*                                                                                                    */
        /*  A linked list is a templated generic double linked list which should be suitable for a wide       */
        /*  range of applications.                                                                            */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Linked List.                                                                              */
            /*                                                                                            */
            /*  A linked list supports many of the features of other list classes and can be easily       */
            /*  substituted in many cases.                                                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE=VOID*,class BASE_CLASS=NO_CLASS,SBIT32 DISPLACEMENT=0> class LINKED_LIST : public BASE_CLASS
        {
            //
            //  Private types.
            //
            typedef LINKED_LIST<TYPE,BASE_CLASS,DISPLACEMENT> LIST;

            //
            //  Private data.
            //
#ifdef DEBUGGING
            LIST                                *GuardWord;
#endif
            TYPE                                *Forward;
            TYPE                                *Backward;

        public:
            //
            //  Public functions.
            //
            LINKED_LIST( VOID );

            VOID InsertAfter( LIST *Header,TYPE *NewType );

            VOID InsertBefore( LIST *Header,TYPE *NewType );

            VOID InsertFirst( LIST *Header );

            VOID InsertLast( LIST *Header );

            VOID Unlink( LIST *Header );

            ~LINKED_LIST( VOID );

            //
            //  Public inline functions.
            //
            INLINE TYPE *Backwards( VOID )
                { return Backward; }

            INLINE BOOLEAN End( VOID )
                { return (this == NULL); }

            INLINE TYPE *First( VOID )
                { return Forward; }

            INLINE TYPE *Forwards( VOID )
                { return Forward; }

            INLINE TYPE *Last( VOID )
                { return Backward; }

        private:
            //
            //  Private functions.
            //
            VOID EmptyList( LIST *Header );

            VOID VerifyList( LIST *Header );

            //
            //  Private inline functions.
            //
            INLINE LIST *ComputeList( TYPE *Type )
                { return ((LIST*) (((CHAR*) Type) + DISPLACEMENT)); }

            //
            //  Disabled functionality.
            //
            LINKED_LIST( CONST LINKED_LIST & NoCopies );

            VOID operator=( CONST LINKED_LIST & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'LINKED LIST'.                                                            */
            /*                                                                                            */
            /*  We make sure that we zero all pointers to make debugging somewhat more tractable.         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::LINKED_LIST( VOID )
        {
        //
        //  We always zero each pointer so we can perform various validation tests later.
        //
#ifdef DEBUGGING
        GuardWord = NULL;
#endif
        Backward = NULL;
        Forward = NULL;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Insert After.                                                                             */
            /*                                                                                            */
            /*  We insert a new element after the current element and update all of the references.       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::InsertAfter
            (
            LIST                                *Header,
            TYPE                                *Type
            )
        {
        //
        //  When the list is completely empty we need to do special processing.
        //
        if ( Header -> Backward != NULL )
            {
            AUTO LIST *Current = ComputeList( Type );
            AUTO LIST *New = ComputeList( ((TYPE*) this) );
#ifdef DEBUGGING

            //
            //  We ensure that we insert the new element into the correct list.
            //
            if ( (Current -> GuardWord == Header) && (New -> GuardWord == NULL) )
                {
#endif
                //
                //  We setup the new element links reday for insertion into the list.
                //
                New -> Backward = Type;
                New -> Forward = Current -> Forward;
#ifdef DEBUGGING
                New -> GuardWord = Header;
#endif

                //
                //  Let's make sure that this is not the last element in the list.
                //
                if ( Current -> Forward != NULL )
                    {
                    AUTO LIST *Next = ComputeList( Current -> Forward );

                    //
                    //  We insert the new element between the current element and
                    //  the next element.
                    //
                    Current -> Forward = ((TYPE*) this);
                    Next -> Backward = ((TYPE*) this);
                    }
                else
                    {
                    //
                    //  We insert the new element at the end of the list.
                    //
                    Current -> Forward = ((TYPE*) this);
                    Header -> Backward = ((TYPE*) this);
                    }
#ifdef DEBUGGING
                }
            else
                { FAILURE( "Linked list header mismatch" ); }
#endif
            }
        else
            {
            //
            //  We deal with the special processing for empty lists in a seperate function.
            //
            EmptyList( Header );
            }
#ifdef SPECIAL_LIST_DEBUGGING

        //
        //  Verify the linked list.
        //
        VerifyList( Header );
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Insert Before.                                                                            */
            /*                                                                                            */
            /*  We insert a new element before the current element and update all of the references.      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::InsertBefore
            (
            LIST                                *Header,
            TYPE                                *Type
            )
        {
        //
        //  When the list is completely empty we need to do special processing.
        //
        if ( Header -> Backward != NULL )
            {
            AUTO LIST *Current = ComputeList( Type );
            AUTO LIST *New = ComputeList( ((TYPE*) this) );
#ifdef DEBUGGING

            //
            //  We ensure that we insert the new element into the correct list.
            //
            if ( (Current -> GuardWord == Header) && (New -> GuardWord == NULL) )
                {
#endif
                //
                //  We setup the new element links reday for insertion into the list.
                //
                New -> Backward = Current -> Backward;
                New -> Forward = Type;
#ifdef DEBUGGING
                New -> GuardWord = Header;
#endif

                //
                //  Let's make sure that this is not the first element in the list.
                //
                if ( Current -> Backward != NULL )
                    {
                    AUTO LIST *Previous = ComputeList( Current -> Backward );

                    //
                    //  We insert the new element between thecurrent element and the
                    //  previous element.
                    //
                    Current -> Backward = ((TYPE*) this);
                    Previous -> Forward = ((TYPE*) this);
                    }
                else
                    {
                    //
                    //  We insert the new element at the head of the list.
                    //
                    Current -> Backward = ((TYPE*) this);
                    Header -> Forward = ((TYPE*) this);
                    }
#ifdef DEBUGGING
                }
            else
                { FAILURE( "Linked list header mismatch" ); }
#endif
            }
        else
            {
            //
            //  We deal with the special processing for empty lists in a seperate function.
            //
            EmptyList( Header );
            }
#ifdef SPECIAL_LIST_DEBUGGING

        //
        //  Verify the linked list.
        //
        VerifyList( Header );
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Empty List.                                                                               */
            /*                                                                                            */
            /*  We insert a new element into an empty list and then update all of the references.         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::EmptyList
            (
            LIST                                *Header
            )
        {
        AUTO LIST *New = ComputeList( ((TYPE*) this) );
#ifdef DEBUGGING

        //
        //  We ensure that we insert the new element into the correct list.
        //
        if
                (
                ((Header -> Backward == NULL) && (Header -> Forward == NULL))
                    &&
                ((Header -> GuardWord == NULL) && (New -> GuardWord == NULL))
                    &&
                ((New -> Backward == NULL) && (New -> Forward == NULL))
                )
            {
#endif
            //
            //  As the list is empty the  head and the tail of the list point to
            //  the new element.
            //
            Header -> Forward = ((TYPE*) this);
            Header -> Backward = ((TYPE*) this);

            //
            //  There are no forward and backward links for the fist elemnet in a list.
            //
            New -> Forward = NULL;
            New -> Backward = NULL;
#ifdef DEBUGGING
            New -> GuardWord = Header;
            }
        else
            { FAILURE( "Linked list header mismatch" ); }
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Insert First.                                                                             */
            /*                                                                                            */
            /*  We insert a new element at the start of the list and update all of the references.        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::InsertFirst
            (
            LIST                                *Header
            )
        {
        //
        //  When the list is completely empty we need to do special processing.
        //
        if ( Header -> Backward != NULL )
            {
            AUTO LIST *Current = ComputeList( Header -> Forward );
            AUTO LIST *New = ComputeList( ((TYPE*) this) );
#ifdef DEBUGGING

            //
            //  We ensure that we insert the new element into the correct list.
            //
            if ( (Current -> GuardWord == Header) && (New -> GuardWord == NULL) )
                {
#endif

                //
                //  We setup the new element links reday for insertion into the list.
                //
                New -> Forward = Header -> Forward;
                New -> Backward = NULL;
#ifdef DEBUGGING
                New -> GuardWord = Header;
#endif

                //
                //  Finally, we insert the new element at the start of the list.
                //
                Current -> Backward = ((TYPE*) this);
                Header -> Forward = ((TYPE*) this);
#ifdef DEBUGGING
                }
            else
                { FAILURE( "Linked list header mismatch" ); }
#endif
            }
        else
            {
            //
            //  We deal with the special processing for empty lists in a seperate function.
            //
            EmptyList( Header );
            }
#ifdef SPECIAL_LIST_DEBUGGING

        //
        //  Verify the linked list.
        //
        VerifyList( Header );
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Insert Last.                                                                              */
            /*                                                                                            */
            /*  We insert a new element at the end of the list and update all of the references.          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::InsertLast
            (
            LIST                                *Header
            )
        {
        //
        //  When the list is completely empty we need to do special processing.
        //
        if ( Header -> Backward != NULL )
            {
            AUTO LIST *Current = ComputeList( Header -> Backward );
            AUTO LIST *New = ComputeList( ((TYPE*) this) );
#ifdef DEBUGGING

            //
            //  We ensure that we insert the new element into the correct list.
            //
            if ( (Current -> GuardWord == Header) && (New -> GuardWord == NULL) )
                {
#endif

                //
                //  We setup the new element links reday for insertion into the list.
                //
                New -> Forward = NULL;
                New -> Backward = Header -> Backward;
#ifdef DEBUGGING
                New -> GuardWord = Header;
#endif

                //
                //  Finally, we insert the new element at the end of the list.
                //
                Current -> Forward = ((TYPE*) this);
                Header -> Backward = ((TYPE*) this);
#ifdef DEBUGGING
                }
            else
                { FAILURE( "Linked list header mismatch" ); }
#endif
            }
        else
            {
            //
            //  We deal with the special processing for empty lists in a seperate function.
            //
            EmptyList( Header );
            }
#ifdef SPECIAL_LIST_DEBUGGING

        //
        //  Verify the linked list.
        //
        VerifyList( Header );
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Unlink.                                                                                   */
            /*                                                                                            */
            /*  We unlink an element from the list and update all of of the references.                   */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::Unlink
            (
            LIST                                *Header
            )
        {
        AUTO LIST *Current = ComputeList( ((TYPE*) this) );

        //
        //  Let's make sure that this is not the first element in the list.
        //
        if ( Current -> Backward != NULL )
            {
            AUTO LIST *Previous = ComputeList( Current -> Backward );

            //
            //  We delete ourselves from the forward chain.
            //
            Previous -> Forward = Current -> Forward;
            }
        else
            {
            //
            //  We delete ourselves from the forward chain and update the header.
            //
            Header -> Forward = Current -> Forward;
            }

        //
        //  Let's make sure that this is not the last element in the list.
        //
        if ( Current -> Forward != NULL )
            {
            AUTO LIST *Next = ComputeList( Current -> Forward );

            //
            //  We delete ourselves from the backward chain.
            //
            Next -> Backward = Current -> Backward;
            }
        else
            {
            //
            //  We delete ourselves from the backward chain and update the header.
            //
            Header -> Backward = Current -> Backward;
            }

        //
        //  We should have removed ourselves from all chains so now we can delete
        //  the original links.
        //
        Current -> Backward = NULL;
        Current -> Forward = NULL;
#ifdef DEBUGGING

        //
        //  We ensure that we deleted the element from the correct list.
        //
        if ( Current -> GuardWord == Header )
            { Current -> GuardWord = NULL; }
        else
            { FAILURE( "Linked list header mismatch" ); }
#endif
#ifdef SPECIAL_LIST_DEBUGGING

        //
        //  Verify the linked list.
        //
        VerifyList( Header );
#endif
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Verify List.                                                                              */
            /*                                                                                            */
            /*  We verify that the linked list is correctly formed.                                       */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        VOID LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::VerifyList
            (
            LIST                                *Header
            )
        {
        AUTO BOOLEAN Sorted = True;
        AUTO SNATIVE Length = 0;

        //
        //  Verify the linked list header.
        //
        if
                (
                ((Header -> Backward == NULL) && (Header -> Forward == NULL))
                    ||
                ((Header -> Backward != NULL) && (Header -> Forward != NULL))
                )
            {
            AUTO LIST *Current = Header -> Forward;
            AUTO LIST *Previous = NULL;

            //
            //  We walk along the list checking each cell.
            //
            for ( /* void */;(! Current -> End());Current=(Current -> Forwards()), Length ++ )
                {
                //
                //  We check the backward link.
                //
                if ( Current -> Backward == Previous )
                    {
                    //
                    //  We update the sorted flag.
                    //
                    if ( (Sorted) && (Previous != NULL) )
                        { Sorted = (Current < Previous); }

                    //
                    //  We update update 'Previous' to keep track of the last cell.
                    //
                    Previous = Current;
                    }
                else
                    { FAILURE( "List backward pointer corrupt" ); }
                }

            //
            //  Let's just check the header backward pointer.
            //
            if ( Header -> Backward != Previous )
                { FAILURE( "List header corrupt" ); }
            }
        else
            { FAILURE( "List header corrupt" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor of 'LINKED LIST'.                                                              */
            /*                                                                                            */
            /*  There is no real need for a destructor but just to be tidy we verify the list is empty    */
            /*  and zero all the fields putting the class fields back to their original values.           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class TYPE,class BASE_TYPE,SBIT32 DISPLACEMENT>
        LINKED_LIST<TYPE,BASE_TYPE,DISPLACEMENT>::~LINKED_LIST( VOID )
        {
        //
        //  We verify all the links have been deleted before we exit.
        //
        if
                (
#ifdef DEBUGGING
                (GuardWord != NULL)
                    ||
#endif
                (Backward != NULL)
                    ||
                (Forward != NULL)
                )
            { FAILURE( "Linked List cell not empty" ); }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define LINKED_LIST ROCKALL::LINKED_LIST
#endif
#endif