#ifndef _TREE_HPP_
#define _TREE_HPP_ 1
//                                              Standard Code Ruler
//34*678*012*45678901234567890123456789012345678*0123456789012345678901234567890123456789012345678901234567*901*345*7890

    /**************************************************************************************************************/
    /*                                                                                                            */
    /*  Standard copyright and disclaimer information.                                                            */
    /*                                                                                                            */
    /*  Copyright  Michael Parkes 2015.  All rights reserved.                                                    */
    /*                                                                                                            */
    /*  Reproduction or transmission of all or part of this work, whether by photocopying or storing in           */
    /*  any medium by electronic means or otherwise, without the written permission of the owner, is              */
    /*  prohibited. The commission of any unauthorised act in relation to this work may result in civil           */
    /*  or criminal actions. Licences issued by the Copyright Licensing Agency Limited do not extend to           */
    /*  this work.                                                                                                */
    /*                                                                                                            */
    /*  The contents of this file may to contain errors, mistakes, omissions and other types of issues.           */
    /*  Consequently, the contents of this file are provided 'as is' and it is the users responsibility to        */
    /*  verify that the contents of the file are fit for purpose for which it is used.  Additionally,             */
    /*  it might be shown at some point that portions of this file are subject to patent claims or other          */
    /*  types of legal claims.  Therefore, any use of the contents of this file is undertaken entirely at         */
    /*  the users own risk.  No responsibility can be accepted for any consequences or any losses unless          */
    /*  clearly and explicitly contractually agreed in writing prior to any use of the file.  Additionally,       */
    /*  any user of this file agrees to take reasonable steps to keep the contents of this file private and       */
    /*  to indemnify the owners for any claims or losses the owners may find themselves liable for as a           */
    /*  result of the use of this file.                                                                           */
    /*                                                                                                            */
    /*  Please see the related license for additional terms and conditions.  Any use of this file indicates       */
    /*  acceptance of all the associated terms and conditions.                                                    */
    /*                                                                                                            */
    /**************************************************************************************************************/

#include "AutomaticHeapScope.hpp"
#include "AutomaticLockScope.hpp"
#include "AutomaticTransactionScope.hpp"
#include "AutomaticViewScope.hpp"
#include "Error.hpp"
#include "Failure.hpp"
#include "FileLock.hpp"
#include "MemoryFunctions.hpp"
#include "MemoryLock.hpp"
#include "NoLock.hpp"
#include "PlacementNew.hpp"
#include "RockallDelete.hpp"
#include "RockallNew.hpp"
#include "TransactionLock.hpp"
#include "VirtualHeap.hpp"

        /******************************************************************************************************/
        /*                                                                                                    */
        /*  Tree.                                                                                             */
        /*                                                                                                    */
        /*  A tree is a templated multi-way tree that is stored in sorted key order and translates            */
        /*  keys to their related values.                                                                     */
        /*                                                                                                    */
        /******************************************************************************************************/

NAMESPACE ROCKALL
    {
            /**********************************************************************************************/
            /*                                                                                            */
            /*  Public Constants.                                                                         */
            /*                                                                                            */
            /*  The public constants specify values that are of use to external code.                     */
            /*                                                                                            */
            /**********************************************************************************************/

    STATIC CONST SBIT32 DefaultLeaves           = 1024;
    STATIC CONST SBIT32 DefaultStems            = 256;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Tree.                                                                                     */
            /*                                                                                            */
            /*  A tree is a templated multi-way tree consisting of key and value pairs that can be        */
            /*  stored either in memory or on backing storage.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES=DefaultLeaves,INT STEMS=DefaultStems> class TREE
        {
            //
            //  Private constants.
            //
            STATIC CONST SBIT32 FullPage        = 100;
            STATIC CONST SBIT32 MaxLeafDensity  = 80;
            STATIC CONST SBIT32 MaxStemDensity  = 80;
            STATIC CONST SBIT32 MinLeafDensity  = 20;
            STATIC CONST SBIT32 MinStemDensity  = 60;
            STATIC CONST SBIT32 ViewSize        = 8192;

            STATIC CONST SBIT32 MaxLeaves       = 16384;
            STATIC CONST SBIT32 MaxStems        = 16384;
            STATIC CONST SBIT32 MinLeaves       = 16;
            STATIC CONST SBIT32 MinStems        = 16;

            //
            //  The root signature.
            //
            STATIC CONST SBIT64 Signature1      = ((sizeof(KEY) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature2      = (((Signature1 + LEAVES) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature3      = (((Signature2 + sizeof(LOCK)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature4      = (((Signature3 + STEMS) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature5      = (((Signature4 + sizeof(VALUE)) * PrimeSBit32) & HashMaskSBit32);
            STATIC CONST SBIT64 Signature       = (0x0000500000000000 + Signature5);

            //
            //  Private types.
            //
            typedef TRANSACTION_LOCK<HEAP> TRANSACTIONAL_LOCK;

            typedef enum
                {
                TreeRoot                        = 0x0000,
                TreeBranch                      = 0x0001,
                TreeTwig                        = 0x0002,
                EndOfTreeType
                }
            TREE_TYPE;

            //
            //  A 'LEAF' holds the user key and the associated value.
            //
            typedef struct
                {
                KEY                             Key;
                VALUE                           Value;
                }
            LEAF;

            //
            //  The sorted order of the keys is held in 'ORDER'.
            //
            typedef struct
                {
                SBIT16                          IndexToValue;
                SBIT16                          ValueToOrder;
                }
            ORDER;

            //
            //  A 'STEM' is part of a 'BRANCH' and points to a 'TWIG'.
            //
            typedef struct
                {
                KEY                             Key;
                FILE_ADDRESS                    Address;
                }
            STEM;

            //
            //  A 'TWIG' contains a number of 'LEAF' structures packed together into a
            //  single allocation (i.e. a page).  The construction of the tree is that
            //  the 'ROOT' points to a 'BRANCH' which contains one or more 'STEM' structures
            //  that point to either a 'BRANCH' or 'TWIG'.  A 'TWIG' contains one or more
            //  'LEAF' structures.
            //
            typedef struct
                {
                //
                //  Fixed common header shared by every 'Type'.
                //
                FILE_ADDRESS                    Address;
                LOCK                            TwigLock;
                TREE_TYPE                       Type;

                //
                //  Flexible body specific to each 'Type'.
                //
                SBIT32                          MaxSize;
                SBIT32                          MaxSpace;
                SBIT32                          Used;

                SBIT16                          MaxDensity;
                SBIT16                          MinDensity;

                LEAF                            Leaf[ LEAVES ];
                //
                //  There is a hidden array following 'Leaf' of the same size but
                //  of type 'ORDER'.  This array keeps track of the logical order of
                //  the 'Leaf' values which typically differs from the physical order.
                //  The 'DoNotTouch' array provides space for this hidden array when the
                //  structure is its defualt size but as the structure grows the array
                //  will move further down the structure.
                //
                ORDER                           DoNotTouch[ LEAVES ];
                }
            TWIG;

            //
            //  A 'BRANCH' contains a number of 'STEM' structures packed together into a
            //  single allocation (i.e. a page).  The construction of the tree is that
            //  the 'ROOT' points to a 'BRANCH' which contains one or more 'STEM' structures
            //  that point to either a 'BRANCH' or 'TWIG'.  A 'TWIG' contains one or more
            //  'LEAF' structures.
            //
            typedef struct
                {
                //
                //  Fixed common header shared by every 'Type'.
                //
                FILE_ADDRESS                    Address;
                LOCK                            BranchLock;
                TREE_TYPE                       Type;

                //
                //  Flexible body specific to each 'Type'.
                //
                SBIT32                          MaxSize;
                SBIT32                          MaxSpace;
                SBIT32                          Depth;
                SBIT32                          Used;

                SBIT16                          MaxDensity;
                SBIT16                          MinDensity;

                STEM                            Stem[ STEMS ];
                //
                //  There is a hidden array following 'Stem' of the same size but
                //  of type 'ORDER'.  This array keeps track of the logical order of
                //  the 'Stem' values which typically differs from the physical order.
                //  The 'DoNotTouch' array provides space for this hidden array when the
                //  structure is its default size but as the structure grows the array
                //  will move further down the structure.
                //
                ORDER                           DoNotTouch[ STEMS ];
                }
            BRANCH;

            //
            //  The 'ROOT' is at the base of the tree.
            //
            typedef struct
                {
                //
                //  Fixed common header shared by every 'Type'.
                //
                FILE_ADDRESS                    Address;
                LOCK                            ClassLock;
                TREE_TYPE                       Type;

                //
                //  Flexible body specific to each 'Type'.
                //
                SBIT32                          MaxDepth;
                FILE_SIZE                       Signature;
                FILE_ADDRESS                    Trunk;
                }
            ROOT;

            //
            //  A 'FIND' is a collection of useful data passed around the tree and used to
            //  help find a range of keys and values.
            //
            typedef struct
                {
                KEY                             *Highest;
                KEY                             *Keys;
                KEY                             *Lowest;
                FILE_SIZE                       MaxSize;
                FILE_SIZE                       *Size;
                VALUE                           *Values;
                }
            FIND_RANGE;

            //
            //  A 'NAVIGATION' is a collection of useful data passed around the tree and used
            //  to help insert or delete values.
            //
            typedef struct
                {
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope;
                AUTOMATIC_LOCK_SCOPE<LOCK>      *AutomaticLockScope;
                AUTOMATIC_VIEW_SCOPE<HEAP>      *AutomaticViewScope;
                BRANCH                          *Branch;
                SBIT32                          Calls;
                SBIT32                          Depth;
                KEY                             *Key;
                BOOLEAN                         LastStem;
                BOOLEAN                         Overwrite;
                FILE_ADDRESS                    ParentAddress;
                FILE_ADDRESS                    *Reference;
                BOOLEAN                         ReleaseParent;
                ROOT                            *Root;
                BOOLEAN                         *SharedNavigation;
                TREE_TYPE                       Type;
                VALUE                           *Value;
                }
            NAVIGATION;

            //
            //  Private data.
            //
            BOOLEAN                             Active;
            FILE_ADDRESS                        Address;

            //
            //  Private static data.
            //
            STATIC MEMORY_FUNCTIONS<SNATIVE>    MemoryFunctions;

        public:
            //
            //  Public types.
            //
            typedef BOOLEAN (*TREE_WALK_FUNCTION)
                (
                KEY                             & Key,
                VALUE                           & Value,
                VOID                            *UserValue
                );

            //
            //  Public functions.
            //
            TREE( VOID );
            TREE( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteAll( VOID );
            BOOLEAN DeleteAll( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope );

            BOOLEAN DeleteKey( CONST KEY & Key );
            BOOLEAN DeleteKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key );

            BOOLEAN FindKey( CONST KEY & Key,VALUE *Value );
            BOOLEAN FindKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key,VALUE *Value );

            BOOLEAN FindKeys
                (
                CONST KEY                       & Lowest,
                CONST KEY                       & Highest,
                FILE_SIZE                       MaxSize,
                FILE_SIZE                       *Size = NULL,
                KEY                             *Keys = NULL,
                VALUE                           *Values = NULL
                );

            BOOLEAN FindKeys
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                CONST KEY                       & Lowest,
                CONST KEY                       & Highest,
                FILE_SIZE                       MaxSize,
                FILE_SIZE                       *Size = NULL,
                KEY                             *Keys = NULL,
                VALUE                           *Values = NULL
                );

            BOOLEAN NewKey( CONST KEY & Key,CONST VALUE & Value,BOOLEAN Overwrite = False );

            BOOLEAN NewKey
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                CONST KEY                       & Key,
                CONST VALUE                     & Value,
                BOOLEAN                         Overwrite = False
                );

            BOOLEAN Size( FILE_SIZE *Value );
            BOOLEAN Size( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,FILE_SIZE *Value );

            BOOLEAN UpdateKey( CONST KEY & Key,CONST VALUE & Value );
            BOOLEAN UpdateKey( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,CONST KEY & Key,CONST VALUE & Value );

            BOOLEAN Walk( TREE_WALK_FUNCTION WalkFunction,VOID *UserValue = NULL );

            BOOLEAN Walk
                (
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                TREE_WALK_FUNCTION              WalkFunction,
                VOID                            *UserValue = NULL
                );

            ~TREE( VOID );

        private:
            //
            //  Private functions.
            //
#ifdef DEBUGGING
            VOID DefragmentBranch( BRANCH *Branch );

            VOID DefragmentTwig( TWIG *Twig );

#endif
            VOID DeleteBranch( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,BRANCH *Branch );

            BOOLEAN DeleteAllNavigation( NAVIGATION *Navigation );

            VOID DeleteLeaf( SBIT16 Offset,TWIG *Twig );

            VOID DeleteStem( BRANCH *Branch,SBIT16 Offset );

            VOID DeleteTwig( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TWIG *Twig );

            BOOLEAN DeleteKeyExclusiveNavigation( BOOLEAN *Empty,NAVIGATION *Navigation );

            BOOLEAN DeleteKeySharedNavigation( NAVIGATION *Navigation );

            BOOLEAN FindKeyNavigation( NAVIGATION *Navigation );

            BOOLEAN FindKeysNavigation( FIND_RANGE *FindRange,NAVIGATION *Navigation );

            BOOLEAN FindLeaf( KEY *Key,SBIT16 *Offset,BOOLEAN Nearest,TWIG *Twig );

            BOOLEAN FindRangeValues( FIND_RANGE *FindRange,SBIT32 *Found,NAVIGATION *Navigation,BOOLEAN *Result );

            BOOLEAN FindStem( BRANCH *Branch,KEY *Key,BOOLEAN Nearest,SBIT16 *Offset );

            BOOLEAN NewBranch
                (
                FILE_ADDRESS                    *Address,
                AUTOMATIC_HEAP_SCOPE<HEAP>      *AutomaticHeapScope,
                BRANCH                          **Branch,
                SBIT32                          Depth
                );

            BOOLEAN NewLeaf( KEY *Key,BOOLEAN Overwrite,TWIG *Twig,VALUE *Value );

            BOOLEAN NewStem( FILE_ADDRESS Address,BRANCH *Branch,KEY *Key );

            BOOLEAN NewTwig( FILE_ADDRESS *Address,AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,TWIG **Twig );

            BOOLEAN NewKeyExclusiveNavigation( NAVIGATION *Navigation );

            BOOLEAN NewKeySharedNavigation( NAVIGATION *Navigation );

            BOOLEAN Rebalance( BRANCH *Branch,NAVIGATION *Navigation  );

            BOOLEAN RebalanceStems( BRANCH *Branch,NAVIGATION *Navigation,ROOT *Root );

            BOOLEAN RebalanceTwigs( BRANCH *Branch,NAVIGATION *Navigation,ROOT *Root );

            BOOLEAN ReleaseParent( NAVIGATION *Navigation,BOOLEAN Result );

            BOOLEAN ReleaseSelf
                (
                AUTOMATIC_LOCK_SCOPE<LOCK>      *AutomaticLockScope,
                AUTOMATIC_VIEW_SCOPE<HEAP>      *AutomaticViewScope,
                BOOLEAN                         Result
                );

            BOOLEAN SplitTwig( AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope,BRANCH *Branch,TWIG *Twig );

            BOOLEAN SizeNavigation( NAVIGATION *Navigation,FILE_SIZE *Value );

            BOOLEAN UpdateKeyNavigation( NAVIGATION *Navigation );

            BOOLEAN WalkNavigation( NAVIGATION *Navigation,TREE_WALK_FUNCTION WalkFunction,VOID *UserValue );

            VOID ZeroLeaves( LEAF *Leaves,SBIT32 Size );

            VOID ZeroStems( STEM *Stems,SBIT32 Size );

            //
            //  Private inline functions.
            //
            INLINE ORDER *ComputeOrderAddress( BRANCH *Branch )
                { return ((ORDER*) & Branch -> Stem[ Branch -> MaxSize ]); }

            INLINE ORDER *ComputeOrderAddress( TWIG *Twig )
                { return ((ORDER*) & Twig -> Leaf[ Twig -> MaxSize ]); }

            INLINE BOOLEAN GreaterOrEqual( KEY *Key1,KEY *Key2 )
                { return ((*Key1) >= (*Key2)); }

            INLINE BOOLEAN GreaterThan( KEY *Key1,KEY *Key2 )
                { return ((*Key1) > (*Key2)); }

            INLINE BOOLEAN MatchingKeys( KEY *Key1,KEY *Key2 )
                { return ((*Key1) == (*Key2)); }

            INLINE BOOLEAN LessOrEqual( KEY *Key1,KEY *Key2 )
                { return ((*Key1) <= (*Key2)); }

            INLINE BOOLEAN LessThan( KEY *Key1,KEY *Key2 )
                { return ((*Key1) < (*Key2)); }

            //
            //  Disabled functionality.
            //
            TREE( CONST TREE & NoCopies );

            VOID operator=( CONST TREE & NoCopies );
        };

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Class Static Data.                                                                        */
            /*                                                                                            */
            /*  The class static data needs to be initialised before it's ready for use.                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        MEMORY_FUNCTIONS<SNATIVE> TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::MemoryFunctions;

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'TREE'.                                                                   */
            /*                                                                                            */
            /*  The constructor for 'TREE' prepares the class for use.                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
            TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::TREE( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        //
        //  When we create a tree we ensure that the size of the 'LEAVES' and 'STEMS' are reasonable.
        //
        if ( ((LEAVES >= MinLeaves) && (LEAVES <= MaxLeaves)) && ((STEMS >= MinStems) && (STEMS <= MaxStems))  )
            {
            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
            //  we fail to highlight the issue.
            //
            if ( AutomaticHeapScope.Feature( (LOCK::Feature()) ) )
                {
                //
                //  Where necessary we automatically wrap all the changes within a new transaction.
                //
                if ( AutomaticHeapScope.BeginTransaction() )
                    {
                    AUTO ROOT *Root;

                    //
                    //  We create a new 'Root' on the heap and store the 'Address'.
                    //
                    if ( ROCKALL_NEW<ROOT>::New( & AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                        {
                        //
                        //  We now setup the 'Root' of the 'Tree'.
                        //
                        Root -> Address = Address;
                        Root -> Type = TreeRoot;
                        Root -> MaxDepth = 0;
                        Root -> Signature = Signature;
                        Root -> Trunk = NULL;

                        Active = True;
                        }
                    }
                }
            else
                { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
            }
        else
            { FAILURE( "The number of leaves or stems is outside the permitted range" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Constructor for 'TREE'.                                                                   */
            /*                                                                                            */
            /*  The constructor for 'TREE' prepares the class for use.                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::TREE
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        //
        //  When we create a tree we ensure that the size of the 'LEAVES' and 'STEMS' are reasonable.
        //
        if ( ((LEAVES >= MinLeaves) && (LEAVES <= MaxLeaves)) && ((STEMS >= MinStems) && (STEMS <= MaxStems))  )
            {
            //
            //  Setup class members.
            //
            Active = False;
            Address = NULL;

            //
            //  We verify that the 'Lock' requirements match with the type of 'Heap'.  If not,
            //  we fail to highlight the issue.
            //
            if ( AutomaticHeapScope -> Feature( (LOCK::Feature()) ) )
                {
                AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );

                //
                //  Where necessary we automatically wrap all the changes within a new transaction.
                //
                if ( AutomaticTransactionScope.BeginTransaction() )
                    {
                    AUTO ROOT *Root;

                    //
                    //  We create a new 'Root' on the heap and store the 'Address'.
                    //
                    if ( ROCKALL_NEW<ROOT>::New( AutomaticHeapScope,& Address,& Root,1,NULL,False,True ) )
                        {
                        //
                        //  We now setup the 'Root' of the 'Tree'.
                        //
                        Root -> Address = Address;
                        Root -> Type = TreeRoot;
                        Root -> MaxDepth = 0;
                        Root -> Signature = Signature;
                        Root -> Trunk = NULL;

                        Active = True;
                        }
                    }
                }
            else
                { FAILURE( "The requirements of the lock conflict with the type of heap" ); }
            }
        else
            { FAILURE( "The number of leaves or stems is outside the permitted range" ); }
        }
#ifdef DEBUGGING

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Defragment Branch.                                                                        */
            /*                                                                                            */
            /*  All the stems in the branch are sorted into physical order.                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DefragmentBranch
            (
            BRANCH                              *Branch
            )
        {
        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
        AUTO SBIT16 Count;

        //
        //  We walk through the stems checking and sorting them into physical order.
        //
        for ( Count=0;Count < Branch -> Used;Count ++ )
            {
            AUTO ORDER *Current1 = & Order[ Count ];

            //
            //  We do not need to do anything if the logical 'Order' matches the
            //  physical order.
            //
            if ( Count != Current1 -> IndexToValue )
                {
                AUTO ORDER *Current2 = & Order[ Current1 -> IndexToValue ];
                AUTO ORDER *Current3 = & Order[ Current1 -> ValueToOrder ];

                //
                //  We validate the 'Order' values before sorting to find any problems.
                //
                if
                        (
                        (Current1 -> IndexToValue == Current2 -> ValueToOrder)
                            &&
                        (Current1 -> ValueToOrder == Current3 -> IndexToValue)
                        )
                    {
                    AUTO STEM Temporary = Branch -> Stem[ Count ];

                    //
                    //  Move the logical 'Stem' into the correct physical slot.
                    //
                    Branch -> Stem[ Count ] = Branch -> Stem[ Current1 -> IndexToValue ];
                    Branch -> Stem[ Current1 -> IndexToValue ] = Temporary;

                    Current3 -> IndexToValue = Current1 -> IndexToValue;
                    Current2 -> ValueToOrder = Current1 -> ValueToOrder;

                    Current1 -> IndexToValue = Count;
                    Current1 -> ValueToOrder = Count;
                    }
                else
                    { FAILURE( "Mismatching order values" ); }
                }
            else
                {
                //
                //  We validate the 'Order' values even if we don't need to do any sorting just
                //  to find any related problems.
                //
                if ( Count != Current1 -> IndexToValue )
                    { FAILURE( "Mismatching order values" ); }
                }
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Defragment Twig.                                                                          */
            /*                                                                                            */
            /*  All the leaves in the twig are sorted into physical order.                                */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DefragmentTwig
            (
            TWIG                                *Twig
            )
        {
        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
        AUTO SBIT16 Count;

        //
        //  We walk through the leaves checking and sorting them into physical order.
        //
        for ( Count=0;Count < Twig -> Used;Count ++ )
            {
            AUTO ORDER *Current1 = & Order[ Count ];

            //
            //  We do not need to do anything if the logical 'Order' matches the
            //  physical order.
            //
            if ( Count != Current1 -> IndexToValue )
                {
                AUTO ORDER *Current2 = & Order[ Current1 -> IndexToValue ];
                AUTO ORDER *Current3 = & Order[ Current1 -> ValueToOrder ];

                //
                //  We validate the 'Order' values before sorting to find any problems.
                //
                if ( (Count == Current2 -> ValueToOrder) && (Count == Current3 -> IndexToValue) )
                    {
                    AUTO LEAF Temporary = Twig -> Leaf[ Count ];

                    //
                    //  Move the logical 'Leaf' into the correct physical slot.
                    //
                    Twig -> Leaf[ Count ] = Twig -> Leaf[ Current1 -> IndexToValue ];
                    Twig -> Leaf[ Current1 -> IndexToValue ] = Temporary;

                    Current3 -> IndexToValue = Current1 -> IndexToValue;
                    Current2 -> ValueToOrder = Current1 -> ValueToOrder;

                    Current1 -> IndexToValue = Count;
                    Current1 -> ValueToOrder = Count;
                    }
                else
                    { FAILURE( "Mismatching order values" ); }
                }
            else
                {
                //
                //  We validate the 'Order' values even if we don't need to do any sorting just
                //  to find any related problems.
                //
                if ( Count != Current1 -> IndexToValue )
                    { FAILURE( "Mismatching order values" ); }
                }
            }
        }
#endif

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the key and value pairs in the tree.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
            BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteAll( VOID )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteAll( & AutomaticHeapScope ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All.                                                                               */
            /*                                                                                            */
            /*  Delete all the key and value pairs in the tree.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteAll
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
            AUTO ROOT *Root;

            //
            //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the tree.
            //
            if
                    (
                    (AutomaticHeapScope -> Update( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticExclusiveLock2,
                    NULL,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    NULL,
                    True,
                    False,
                    Address,
                    & Root -> Trunk,
                    False,
                    Root,
                    NULL,
                    Root -> Type,
                    NULL
                    };

                //
                //  When we have multiple threads in a single transaction we need to claim a
                //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                //  and then claim an exclusive lock to prevent them from returning.  Once we
                //  have claimed the exclusive lock we can release the 'TransactionLock' as
                //  we no longer need it.
                //
                AutomaticExclusiveLock1.Release();

                //
                //  Delete all of the tree.
                //
                if ( DeleteAllNavigation( & NextLevel ) )
                    {
                    Root -> MaxDepth = 0;
                    Root -> Trunk = NULL;

                    return True;
                    }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete All Navigation.                                                                    */
            /*                                                                                            */
            /*  Delete all the key and value pairs in the tree.                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteAllNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticHeapScope -> Update( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO SBIT32 Count;

                        //
                        //  We delete every 'Stem' in the 'Branch'.  We delete the stems in
                        //  reverse order just in case we have a problem, so that we are always
                        //  left with a valid tree.
                        //
                        for ( Count=(Branch -> Used-1);Count >= 0;Count -- )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticExclusiveLock,
                                NULL,
                                Branch,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                NULL,
                                True,
                                False,
                                Address,
                                & Branch -> Stem[ Count ].Address,
                                False,
                                NULL,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                NULL
                                };

                            //
                            //  Delete a 'Stem' from the 'Branch'.
                            //
                            if ( DeleteAllNavigation( & NextLevel ) )
                                { DeleteStem( Branch,(Branch -> Used-1) ); }
                            }

                        //
                        //  If we were able to delete all the stems we can now delete the 'Branch'.
                        //  If not, we will still have a valid tree and so we can safely exit.
                        //
                        if ( Branch -> Used <= 0 )
                            {
                            //
                            //  We release the lock and delete the 'Branch'.
                            //
                            ReleaseSelf( & AutomaticExclusiveLock,NULL,True );

                            DeleteBranch( AutomaticHeapScope,Branch );

                            return True;
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        //
                        //  We release the lock and delete the 'Twig'.
                        //
                        ReleaseSelf( & AutomaticExclusiveLock,NULL,True );

                        DeleteTwig( AutomaticHeapScope,Twig );

                        return True;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return True; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Branch.                                                                            */
            /*                                                                                            */
            /*  Delete a branch.                                                                          */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteBranch
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            BRANCH                              *Branch
            )
        {
        AUTO SBIT32 Count;

        //
        //  Call any extra destructors.
        //
        for ( Count=STEMS;Count < Branch -> MaxSize;Count ++ )
            { PLACEMENT_DELETE( & Branch -> Stem[ Count ],STEM ); }

        //
        //  Delete the 'Branch'.
        //
        if ( ! ROCKALL_DELETE<BRANCH>::Delete( AutomaticHeapScope,Branch -> Address ) )
            { FAILURE( "Unable to delete a branch" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key.                                                                               */
            /*                                                                                            */
            /*  Delete a key and value pair in the tree.                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteKey
            (
            CONST KEY                           & Key
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (DeleteKey( & AutomaticHeapScope,Key ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key.                                                                               */
            /*                                                                                            */
            /*  Delete a key and value pair in the tree.                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()', 'ExclusiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO BOOLEAN SharedNavigation = True;

            //
            //  Clearly, it is much better to use 'SharedNavigation' where we can to avoid
            //  exclusive locks.
            //
            if ( SharedNavigation )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'View()' the 'Root' as we need to examine it.  We claimed the
                //  'TransactionLock' above to prevent threads in the same transaction
                //  from changing the tree.
                //
                if
                        (
                        (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                    AUTO NAVIGATION NextLevel =
                        {
                        AutomaticHeapScope,
                        & AutomaticShareableLock2,
                        & AutomaticViewScope,
                        NULL,
                        0,
                        Root -> MaxDepth,
                        ((KEY*) & Key),
                        True,
                        False,
                        Address,
                        & Root -> Trunk,
                        True,
                        Root,
                        & SharedNavigation,
                        Root -> Type,
                        NULL
                        };

                    //
                    //  Delete the 'Key' in the tree.
                    //
                    if ( DeleteKeySharedNavigation( & NextLevel ) )
                        { return True; }
                    else
                        {
                        //
                        //  We switch to exclusive locks if required in this case
                        //  otherwise we exit.
                        //
                        if ( SharedNavigation )
                            { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                        }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }

            //
            //  We need to use exclusive locks if we need to alter any branches or the
            //  structure of the tree.
            //
            if ( ! SharedNavigation )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'ExclusiveView()' the 'Root' as we need to examine it.  We claimed
                //  the 'TransactionLock' above to prevent threads in the same transaction
                //  from accessing the tree.
                //
                if
                        (
                        (AutomaticViewScope.ExclusiveView( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );
                    AUTO BOOLEAN Empty = False;
                    AUTO NAVIGATION NextLevel =
                        {
                        AutomaticHeapScope,
                        & AutomaticExclusiveLock2,
                        & AutomaticViewScope,
                        NULL,
                        0,
                        Root -> MaxDepth,
                        ((KEY*) & Key),
                        True,
                        False,
                        Address,
                        & Root -> Trunk,
                        True,
                        Root,
                        NULL,
                        Root -> Type,
                        NULL
                        };

                    //
                    //  When we have multiple threads in a single transaction we need to claim a
                    //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                    //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                    //  and then claim an exclusive lock to prevent them from returning.  Once we
                    //  have claimed the exclusive lock we can release the 'TransactionLock' as
                    //  we no longer need it.
                    //
                    AutomaticExclusiveLock1.Release();

                    //
                    //  Delete the 'Key' in the tree.
                    //
                    if ( DeleteKeyExclusiveNavigation( & Empty,& NextLevel ) )
                        {
                        //
                        //  If the tree is 'Empty' we zero the 'Trunk' so it no longer points
                        //  to any branches or twigs.
                        //
                        if ( Empty )
                            {
                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Root) ) )
                                {
                                Root -> MaxDepth = 0;
                                Root -> Trunk = NULL;
                                }
                            else
                                { FAILURE( "Unable to delete root" ); }
                            }

                        return True;
                        }
                    else
                        { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }


        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key Exclusive Navigation.                                                          */
            /*                                                                                            */
            /*  Navigate the branches and twigs to delete a key and value pair.                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteKeyExclusiveNavigation
            (
            BOOLEAN                             *Empty,
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  We set 'Empty' to 'False' by default.
        //
        (*Empty) = False;

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        //
                        //  Let's see if the 'Branch' needs rebalancing.  We do not rebalance when
                        //  a 'Twig' has low density as this is quite frequent and rebalancing is
                        //  very expensive.  However, if the 'Branch' density gets low this means
                        //  means that a lot of 'Twigs' must have been deleted and so it's probably
                        //  worth rebalancing the tree.
                        //
                        if
                                (
                                ((Branch -> Used >= Branch -> MinDensity) && (Branch -> Used <= Branch -> MaxDensity))
                                    ||
                                (Navigation -> LastStem)
                                )
                            {
                            AUTO BOOLEAN EmptyChild = False;
                            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                            AUTO BOOLEAN Result = False;
                            AUTO SBIT16 Offset;

                            //
                            //  We can navigate towards the correct 'Twig' as long we can find a
                            //  suitable 'Stem'.  When we return we need to see if the next level
                            //  down has become 'Empty'.  If so, we will need to delete the 'Stem'
                            //  and maybe ourseleves as well.
                            //
                            if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                                {
                                AUTO NAVIGATION NextLevel =
                                    {
                                    Navigation -> AutomaticHeapScope,
                                    & AutomaticExclusiveLock,
                                    & AutomaticViewScope,
                                    Branch,
                                    (Navigation -> Calls+1),
                                    (Navigation -> Depth-1),
                                    Navigation -> Key,
                                    (Order[ Offset ].ValueToOrder == (Branch -> Used-1)),
                                    False,
                                    Address,
                                    & Branch -> Stem[ Offset ].Address,
                                    True,
                                    NULL,
                                    Navigation -> SharedNavigation,
                                    Branch -> Type,
                                    Navigation -> Value
                                    };

                                //
                                //  We will release 'ExclusiveView()' and unlock on our 'Parent'
                                //  if we have multiple stems, as it should not be possible
                                //  for us to become 'Empty' and so require our 'Parent' to make
                                //  any changes or use 'Update()'.  This will allow other
                                //  threads to use it.
                                //
                                Result =
                                    (
                                    ((Branch -> Used == 1) || (ReleaseParent( Navigation,True )))
                                        &&
                                    (DeleteKeyExclusiveNavigation( & EmptyChild,& NextLevel ))
                                    );

                                //
                                //  When we have an 'EmptyChild' we need to delete any reference to
                                //  it.  If this is the last 'Stem' in the 'Branch' then we will need
                                //  to delete ourselves as well.
                                //
                                if ( (Result) && (EmptyChild) )
                                    {
                                    //
                                    //  We need to make changes to the 'Branch'.
                                    //
                                    if ( AutomaticHeapScope -> Update( Address,((VOID**) & Branch) ) )
                                        {
                                        //
                                        //  Delete the related 'Stem'.
                                        //
                                        DeleteStem( Branch,Offset );

                                        //
                                        //  Delete the 'Branch' if it is empty. We should be holding
                                        //  an 'ExclusiveView()' and an 'Update()' on the 'Branch'
                                        //  so release the 'ExclusiveView()' before deleting the
                                        //  'Branch'.
                                        //
                                        if ( ((*Empty) = (Branch -> Used == 0)) )
                                            {
                                            ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True );

                                            DeleteBranch( AutomaticHeapScope,Branch );

                                            return True;
                                            }
                                        }
                                    else
                                        { FAILURE( "Unable to delete stem" ); }
                                    }
                                }

                            return Result;
                            }
                        else
                            {
                            //
                            //  We will choose to 'Rebalance()' a 'Branch' if the density is getting
                            //  low. We try to avoid rebalancing as it is very costly so we leave it
                            //  until we find a branch that has a low density (but not not the last
                            //  branch as this often has a low density).
                            //
                            if ( Rebalance( Branch,Navigation ) )
                                {
                                //
                                //  When we 'Rebalance()' everthing can move.  Consequently, we can't
                                //  depend on the work we did previously.  Here we redo the work that
                                //  as been undone by 'Rebalance()' and then we simulate the original
                                //  call from our parent again.
                                //
                                switch ( Navigation -> Type )
                                    {
                                    case TreeBranch:
                                        {
                                        AUTO BRANCH *Parent = (Navigation -> Branch);
                                        AUTO ORDER *Order = (ComputeOrderAddress( Parent ));
                                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Parent -> Stem);
                                        AUTO SBIT16 Offset;

                                        //
                                        //  When we did the 'Rebalance()' all the 'Stems' may have moved
                                        //  around in our parent 'Branch' so let's find the correct
                                        //  'Stem' again.
                                        //
                                        if ( FindStem( Parent,Navigation -> Key,True,& Offset ) )
                                            {
                                            //
                                            //  We need to update 'Navigation' with the correct information
                                            //  as the 'Parent' may have changed during 'Rebalance()'.
                                            //
                                            Navigation -> Calls ++;
                                            Navigation -> LastStem = (Order[ Offset ].ValueToOrder == (Parent -> Used-1));
                                            Navigation -> Reference = & Parent -> Stem[ Offset ].Address;

                                            //
                                            //  We now simulate our parent calling us again.
                                            //
                                            return
                                                (
                                                (ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True ))
                                                    &&
                                                (DeleteKeyExclusiveNavigation( Empty,Navigation ))
                                                );
                                            }
                                        else
                                            { return (FAILURE( "Key present before rebalance but now missing" )); }
                                        }

                                    case TreeRoot:
                                        {
                                        //
                                        //  We now simulate our parent calling us again.
                                        //
                                        return
                                            (
                                            (ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True ))
                                                &&
                                            (DeleteKeyExclusiveNavigation( Empty,Navigation ))
                                            );
                                        }

                                    default:
                                        { return (FAILURE( "Unknown node type" )); }
                                    }
                                }
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                        AUTO SBIT16 Offset;

                        //
                        //  We need to find the 'Key' in the 'Twig' so we can delete it.
                        //
                        if ( FindLeaf( Navigation -> Key,& Offset,False,Twig ) )
                            {
                            //
                            //  We now need to 'Update()' the 'Twig'.
                            //
                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                {
                                //
                                //  Delete the related 'Leaf'.
                                //
                                DeleteLeaf( Offset,Twig );

                                //
                                //  Delete the 'Twig' if it is empty. We should be holding
                                //  a 'TwigLock' and an 'Update()' on the 'Twig' so release
                                //  the 'TwigLock' before deleting the 'Twig'.
                                //
                                if ( ((*Empty) = (Twig -> Used == 0)) )
                                    {
                                    ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True );

                                    DeleteTwig( AutomaticHeapScope,Twig );
                                    }

                                return True;
                                }
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return False; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Key Shared Navigation.                                                             */
            /*                                                                                            */
            /*  Navigate the branches and twigs to delete a key and value pair.                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteKeySharedNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        //
                        //  We can only use 'SharedNavigation' if the tree does not need rebalancing.
                        //
                        if
                                (
                                ((Branch -> Used >= Branch -> MinDensity) && (Branch -> Used <= Branch -> MaxDensity))
                                    ||
                                (Navigation -> LastStem)
                                )
                            {
                            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                            AUTO SBIT16 Offset;

                            //
                            //  We can navigate towards the correct 'Twig' as long we can find a
                            //  suitable 'Stem'.  When we return we need to see if the next level
                            //  down has become 'Empty'.  If so, we will need to delete the 'Stem'
                            //  and maybe ourseleves as well.
                            //
                            if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                                {
                                AUTO NAVIGATION NextLevel =
                                    {
                                    Navigation -> AutomaticHeapScope,
                                    & AutomaticShareableLock,
                                    & AutomaticViewScope,
                                    Branch,
                                    (Navigation -> Calls+1),
                                    (Navigation -> Depth-1),
                                    Navigation -> Key,
                                    (Order[ Offset ].ValueToOrder == (Branch -> Used-1)),
                                    False,
                                    Address,
                                    & Branch -> Stem[ Offset ].Address,
                                    True,
                                    NULL,
                                    Navigation -> SharedNavigation,
                                    Branch -> Type,
                                    Navigation -> Value
                                    };

                                //
                                //  We will release the 'View()' and unlock on our 'Parent'
                                //  as we are not allowed to delete the last 'Leaf' in a
                                //  'Twig'.
                                //
                                return
                                    (
                                    (ReleaseParent( Navigation,True ))
                                        &&
                                    (DeleteKeySharedNavigation( & NextLevel ))
                                    );
                                }
                            }
                        else
                            { return ((*Navigation -> SharedNavigation) = False); }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        //
                        //  We are not allowed to delete the last 'Leaf' in the 'Twig'.
                        //
                        if ( Twig -> Used > 1 )
                            {
                            AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                            AUTO SBIT16 Offset;

                            //
                            //  We need to find the 'Key' in the 'Twig' so we can delete it.
                            //
                            if ( FindLeaf( Navigation -> Key,& Offset,False,Twig ) )
                                {
                                //
                                //  We now need to 'Update()' the 'Twig'.
                                //
                                if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                    {
                                    //
                                    //  Delete the related 'Leaf'.
                                    //
                                    DeleteLeaf( Offset,Twig );

                                    return True;
                                    }
                                }
                            }
                        else
                            { return ((*Navigation -> SharedNavigation) = False); }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return False; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Leaf.                                                                              */
            /*                                                                                            */
            /*  Delete a leaf in a twig.                                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteLeaf
            (
            SBIT16                              Offset,
            TWIG                                *Twig
            )
        {
        //
        //  Let's ensure that the 'Offset' falls with the 'Leaf' table.
        //
        if ( (Offset >= 0) && (Offset < Twig -> Used) )
            {
            AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
            AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
            AUTO SBIT16 Current1 = Order[ (-- Twig -> Used) ].ValueToOrder;
            AUTO SBIT16 Current2 = Order[ Offset ].ValueToOrder;
            AUTO SBIT16 Count;

            //
            //  When we delete a 'Leaf' we simply overwrite it with the last value in the
            //  'Leaf' in table and then update the 'Order' table with its new location.
            //
            Twig -> Leaf[ Offset ] = Twig -> Leaf[ Twig -> Used ];

            //
            //  We now need to update 'Order' as the last 'Leaf' has moved to to 'Offset'
            //  in the table.  The value of 'Current1' is the offset in the sorted 'Order'
            //  table of the old last value in 'Leaf' table (i.e. the one we moved).
            //
            Order[ Current1 ].IndexToValue = Offset;
            Order[ Offset ].ValueToOrder = Current1;

            //
            //  We now need to slide the enties in the 'Order' table down, as the old 'Leaf'
            //  needs to be removed from the sorted 'Order' table.  The value of 'Current2'
            //  is the sort position of the 'Leaf' we are deleting in the 'Order' table.
            //
            for ( Count=Current2;Count < Twig -> Used;Count ++ )
                {
                AUTO ORDER *Current3 = & Order[ Count ];
                AUTO ORDER *Current4 = & Order[ (Count+1) ];
                AUTO ORDER *Current5 = & Order[ Current4 -> IndexToValue ];

                //
                //  We simply move the 'IndexToValue' down one element in the 'Order'
                //  table fixing the associated 'ValueToOrder' for every entry we move.
                //
                Current3 -> IndexToValue = Current4 -> IndexToValue;
                Current5 -> ValueToOrder = Count;
                }

            //
            //  Zero the empty 'Order' and 'Leaf' just to be tidy.
            //
            MemoryFunctions.Zero( & Order[ Twig -> Used ],sizeof(ORDER) );
            ZeroLeaves( & Twig -> Leaf[ Twig -> Used ],1 );
            }
        else
            { FAILURE( "No leaf in twig" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Stem.                                                                              */
            /*                                                                                            */
            /*  Delete a stem.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteStem
            (
            BRANCH                              *Branch,
            SBIT16                              Offset
            )
        {
        //
        //  Let's ensure that the 'Offset' falls with the 'Stem' table.
        //
        if ( (Offset >= 0) && (Offset < Branch -> Used) )
            {
            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
            AUTO SBIT16 Current1 = Order[ (-- Branch -> Used) ].ValueToOrder;
            AUTO SBIT16 Current2 = Order[ Offset ].ValueToOrder;
            AUTO SBIT16 Count;

            //
            //  When we delete the first element in the 'Order' table we need to deal with
            //  the 'Key', as this may be refered to by a 'Branch' higher in the tree.  The
            //  'Branch' higher in the tree will think that we have all the 'Key' values in
            //  this range.  We cheat by simply storing the 'Key' over the second 'Key' in
            //  the 'Order' table and extending the range of the values supported by the
            //  associated 'Stem'.
            //
            if ( (Current1 == 0) && (Branch -> Used > 0) )
                { Branch -> Stem[ Order[1].IndexToValue ].Key = Branch -> Stem[ Order[1].IndexToValue ].Key; }

            //
            //  When we delete a 'Stem' we simply overwrite it with the last value in the
            //  'Stem' in table and then update the 'Order' table with its new location.
            //
            Branch -> Stem[ Offset ] = Branch -> Stem[ Branch -> Used ];

            //
            //  We now need to update 'Order' as the last 'Stem' has moved to to 'Offset'
            //  in the table.  The value of 'Current1' is the offset in the sorted 'Order'
            //  table of the old last value in 'Stem' table (i.e. the one we moved).
            //
            Order[ Current1 ].IndexToValue = Offset;
            Order[ Offset ].ValueToOrder = Current1;

            //
            //  We now need to slide the enties in the 'Order' table down, as the old 'Stem'
            //  needs to be removed from the sorted 'Order' table.  The value of 'Current2'
            //  is the sort position of the 'Stem' we are deleting in the 'Order' table.
            //
            for ( Count=Current2;Count < Branch -> Used;Count ++ )
                {
                AUTO ORDER *Current3 = & Order[ Count ];
                AUTO ORDER *Current4 = & Order[ (Count+1) ];
                AUTO ORDER *Current5 = & Order[ Current4 -> IndexToValue ];

                //
                //  We simply move the 'IndexToValue' down one element in the 'Order'
                //  table fixing the associated 'ValueToOrder' for every entry we move.
                //
                Current3 -> IndexToValue = Current4 -> IndexToValue;
                Current5 -> ValueToOrder = Count;
                }

            //
            //  Zero the empty 'Order' and 'Stem' just to be tidy.
            //
            MemoryFunctions.Zero( & Order[ Branch -> Used ],sizeof(ORDER) );
            ZeroStems( & Branch -> Stem[ Branch -> Used ],1 );
            }
        else
            { FAILURE( "No stem in branch" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Delete Twig.                                                                              */
            /*                                                                                            */
            /*  Delete a twig.                                                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::DeleteTwig
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TWIG                                *Twig
            )
        {
        AUTO SBIT32 Count;

        //
        //  Call any extra destructors.
        //
        for ( Count=LEAVES;Count < Twig -> MaxSize;Count ++ )
            { PLACEMENT_DELETE( & Twig -> Leaf[ Count ],LEAF ); }

        //
        //  Delete the 'Twig'.
        //
        if ( ! ROCKALL_DELETE<TWIG>::Delete( AutomaticHeapScope,Twig -> Address ) )
            { FAILURE( "Unable to delete a twig" ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Key.                                                                                 */
            /*                                                                                            */
            /*  Find a key and its related value in the tree.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKey
            (
            CONST KEY                           & Key,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (FindKey( & AutomaticHeapScope,Key,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Key.                                                                                 */
            /*                                                                                            */
            /*  Find a key and its related value in the tree.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            VALUE                               *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the tree.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticShareableLock2,
                    & AutomaticViewScope,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    ((KEY*) & Key),
                    True,
                    False,
                    Address,
                    & Root -> Trunk,
                    True,
                    Root,
                    NULL,
                    Root -> Type,
                    Value
                    };

                //
                //  Find a 'Key' and the associated 'Value' in the tree.
                //
                if  ( FindKeyNavigation( & NextLevel ) )
                    { return True; }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Key Navigation.                                                                      */
            /*                                                                                            */
            /*  Find a key and its related value in the tree.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKeyNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Offset;

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  When we have a 'Branch' we find closest 'Stem' which is less than
                        //  the 'Key'.
                        //
                        if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticShareableLock,
                                & AutomaticViewScope,
                                Branch,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                Navigation -> Key,
                                False,
                                False,
                                Address,
                                & Branch -> Stem[ Offset ].Address,
                                True,
                                NULL,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                Navigation -> Value
                                };

                            //
                            //  Search the next level down in the tree.
                            //
                            return (FindKeyNavigation( & NextLevel ));
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                        AUTO SBIT16 Offset;

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  When we finally reach a 'Twig' we find the 'Leaf' which contains
                        //  the supplied 'Key' and return the associated 'Value'.
                        //
                        if ( FindLeaf( Navigation -> Key,& Offset,False,Twig ) )
                            {
                            (*(Navigation -> Value)) = Twig -> Leaf[ Offset ].Value;

                            return True;
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return False; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Keys.                                                                                */
            /*                                                                                            */
            /*  Find a a range of keys and values in the tree.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKeys
            (
            CONST KEY                           & Lowest,
            CONST KEY                           & Highest,
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            KEY                                 *Keys,
            VALUE                               *Values
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (FindKeys( & AutomaticHeapScope,Lowest,Highest,MaxSize,Size,Keys,Values ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Keys.                                                                            */
            /*                                                                                            */
            /*  Find a a range of keys and values in the tree.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKeys
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Lowest,
            CONST KEY                           & Highest,
            FILE_SIZE                           MaxSize,
            FILE_SIZE                           *Size,
            KEY                                 *Keys,
            VALUE                               *Values
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the tree.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO FIND_RANGE FindRange = { ((KEY*) & Highest),Keys,((KEY*) & Lowest),MaxSize,Size,Values };
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticShareableLock2,
                    & AutomaticViewScope,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    NULL,
                    True,
                    False,
                    Address,
                    & Root -> Trunk,
                    True,
                    Root,
                    NULL,
                    Root -> Type,
                    NULL
                    };

                //
                //  We 'Zero' the 'Size' just to ensure we return a value.
                //
                if ( Size != NULL )
                    { (*Size) = 0; }

                //
                //  Find a key range and the associated values in the tree.
                //
                return (FindKeysNavigation( & FindRange,& NextLevel ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Keys Navigation.                                                                     */
            /*                                                                                            */
            /*  Find a a range of keys and values in the tree.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindKeysNavigation
            (
            FIND_RANGE                          *FindRange,
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            AUTO BOOLEAN Result = True;
            AUTO SBIT32 Found = 0;

            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Highest = 0;
                        AUTO SBIT16 Lowest = 0;

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  When we have a 'Branch' we find the 'Highest' and 'Lowest' values in
                        //  sorted 'Order'.
                        //
                        if ( FindStem( Branch,FindRange -> Highest,True,& Highest ) )
                            { Highest = Order[ Highest ].ValueToOrder; }
                        else
                            { Highest = (Branch -> Used-1); }

                        if ( FindStem( Branch,FindRange -> Lowest,True,& Lowest ) )
                            { Lowest = Order[ Lowest ].ValueToOrder; }
                        else
                            { Lowest = 0; }

                        //
                        //  When 'Highest' and 'Lowest' match we know that the whole range is
                        //  stored within a child 'Branch'.  If so, we go down a level so that we
                        //  can release all the locks and views on our parent.
                        //
                        if ( Lowest == Highest )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticShareableLock,
                                & AutomaticViewScope,
                                Branch,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                NULL,
                                False,
                                False,
                                Address,
                                & Branch -> Stem[ Order[ Lowest ].IndexToValue ].Address,
                                True,
                                NULL,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                NULL
                                };

                            //
                            //  Find all the values in the range.
                            //
                            return (FindKeysNavigation( FindRange,& NextLevel ));
                            }
                        else
                            {
                            AUTO SBIT32 Count;

                            //
                            //  We have found the lowest 'Branch' that contains the whole range
                            //  within it's children. So now process all the 'Stem' structures
                            //  collecting the necessary values.
                            //
                            for ( Count=Lowest;(Count <= Highest) && (Result);Count ++ )
                                {
                                AUTO NAVIGATION NextLevel =
                                    {
                                    Navigation -> AutomaticHeapScope,
                                    & AutomaticShareableLock,
                                    & AutomaticViewScope,
                                    Branch,
                                    (Navigation -> Calls+1),
                                    (Navigation -> Depth-1),
                                    NULL,
                                    False,
                                    False,
                                    Address,
                                    & Branch -> Stem[ Order[ Count ].IndexToValue ].Address,
                                    True,
                                    NULL,
                                    Navigation -> SharedNavigation,
                                    Branch -> Type,
                                    NULL
                                    };

                                //
                                //  Find all the values in the range.
                                //
                                if ( ! FindRangeValues( FindRange,& Found,& NextLevel,& Result ) )
                                    { break; }
                                }

                            //
                            //  We should have 'Found' all the 'Values' now so update the
                            //  return values.
                            //
                            if ( FindRange -> Size != NULL )
                                { (*FindRange -> Size) = Found; }

                            return Result;
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  Find all the values in the range.
                        //
                        FindRangeValues( FindRange,& Found,Navigation,& Result );

                        //
                        //  We should have 'Found' all the 'Values' now so update the
                        //  return values.
                        //
                        if ( FindRange -> Size != NULL )
                            { (*FindRange -> Size) = Found; }

                        return Result;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return True; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Leaf.                                                                                */
            /*                                                                                            */
            /*  Find a leaf in a twig.                                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindLeaf
            (
            KEY                                 *Key,
            SBIT16                              *Offset,
            BOOLEAN                             Nearest,
            TWIG                                *Twig
            )
        {
        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
        AUTO SBIT16 Maximum = (Twig -> Used-1);
        AUTO SBIT16 Minimum = 0;

        //
        //  We can find a 'Leaf' using a binary search as the 'Order' stores all the leaves
        //  in sorted order.
        //
        while ( Maximum >= Minimum )
            {
            AUTO SBIT16 MidPoint = ((Maximum + Minimum) / 2);

            //
            //  Let's see if the 'MidPoint' value is greater than the target 'Key' value.
            //
            if ( GreaterThan( Key,& Twig -> Leaf[ Order[ MidPoint ].IndexToValue ].Key ) )
                { Minimum = (MidPoint + 1); }
            else
                {
                //
                //  Let's see if the 'MidPoint' value is less than the target 'Key' value.
                //
                if ( LessThan( Key,& Twig -> Leaf[ Order[ MidPoint ].IndexToValue ].Key ) )
                    { Maximum = (MidPoint - 1); }
                else
                    {
                    //
                    //  We have found the target 'Key' value.
                    //
                    (*Offset) = Order[ MidPoint ].IndexToValue;

                    return True;
                    }
                }
            }

        //
        //  When we are trying to find the 'Nearest' then the closest 'Key' less than the supplied
        //  'Key' is perfectly acceptable.  However, the 'Offset' must fall within the 'Leaf' array
        //  to be considered a match.
        //
        if ( Nearest )
            {
            //
            //  When 'Nearest' is 'True' the target 'Key' will fall just before 'Minimum' and just
            //  after 'Maximum'. Consequently, we can set 'Offset' to '(Minimum-1)' as long as this
            //  falls within the 'Leaf' array.
            //
            if ( (Minimum > 0) && (Minimum <= Twig -> Used) )
                {
                (*Offset) = Order[ (Minimum-1) ].IndexToValue;

                return True;
                }
            else
                { return False; }
            }
        else
            {
            //
            //  Even when we are not trying to find the 'Nearest' Key the 'Offset' must still fall
            //  within the 'Leaf' array (i.e. zero to 'Twig -> Used') as we depend on the 'Offset'
            //  for placing any new keys.  The 'Offset' will point to the 'Key' just *after* the
            //  supplied 'Key' and therefore the supplied 'Key' will need to be inserted at the
            //  'Offset' in the array requiring anything after it to be moved down the array.
            //
            //
            (*Offset) = Minimum;

            return False;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Range Values.                                                                        */
            /*                                                                                            */
            /*  Find a range of values in a sub-tree.                                                     */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindRangeValues
            (
            FIND_RANGE                          *FindRange,
            SBIT32                              *Found,
            NAVIGATION                          *Navigation,
            BOOLEAN                             *Result
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));
        AUTO BOOLEAN Finished = False;

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Count;

                        //
                        //  We walk every 'Stem' in the 'Branch' collecting the required values.
                        //
                        for ( Count=0;(Count < Branch -> Used) && (*Result);Count ++ )
                            {
                            AUTO STEM *Stem = & Branch -> Stem[ Order[ Count ].IndexToValue ];

                            //
                            //  We only need to consider a 'Stem' if it's 'Key' falls into
                            //  the required range.
                            //
                            if
                                    (
                                    (GreaterOrEqual( & Stem -> Key,FindRange -> Lowest ))
                                        &&
                                    (LessOrEqual( & Stem -> Key,FindRange -> Highest ))
                                    )
                                {
                                AUTO NAVIGATION NextLevel =
                                    {
                                    Navigation -> AutomaticHeapScope,
                                    & AutomaticShareableLock,
                                    & AutomaticViewScope,
                                    Branch,
                                    (Navigation -> Calls+1),
                                    (Navigation -> Depth-1),
                                    NULL,
                                    False,
                                    False,
                                    Address,
                                    & Stem -> Address,
                                    True,
                                    NULL,
                                    Navigation -> SharedNavigation,
                                    Branch -> Type,
                                    NULL
                                    };

                                //
                                //  Find all the values in the range.
                                //
                                if (! FindRangeValues( FindRange,Found,& NextLevel,Result ) )
                                    { break; }
                                }
                            }

                        return (*Result);
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                        AUTO SBIT16 Highest = 0;
                        AUTO SBIT16 Lowest = 0;
                        AUTO SBIT16 Count;

                        //
                        //  We have found that the part of the range falls within this 'Twig'
                        //  So let's return all the approriate values.
                        //
                        if ( FindLeaf( FindRange -> Highest,& Highest,True,Twig ) )
                            { Highest = Order[ Highest ].ValueToOrder; }
                        else
                            { Highest = Order[ (Twig -> Used-1) ].ValueToOrder; }

                        if ( FindLeaf( FindRange -> Lowest,& Lowest,True,Twig ) )
                            {
                            if ( MatchingKeys( FindRange -> Lowest,& Twig -> Leaf[ Lowest ].Key ) )
                                { Lowest = Order[ Lowest ].ValueToOrder; }
                            else
                                { Lowest = Order[ (Lowest+1) ].ValueToOrder; }
                            }
                        else
                            { Lowest = 0; }

                        //
                        //  We either copy the values or just count them.
                        //
                        if ( (FindRange -> Keys != NULL) || (FindRange -> Values != NULL) )
                            {
                            //
                            //  We now copy all the relevant 'Keys' and 'Values'.
                            //
                            for
                                    (
                                    Count=Lowest;
                                    (Count <= Highest) && ((*Found) < FindRange -> MaxSize);
                                    Count ++,(*Found) ++
                                    )
                                {
                                AUTO LEAF *Leaf = & Twig -> Leaf[ Order[ Count ].IndexToValue ];

                                //
                                //  If 'Keys' or 'Values' is 'NULL' we skip them.
                                //
                                if ( FindRange -> Keys != NULL )
                                    { FindRange -> Keys[ (*Found) ] = Leaf -> Key; }

                                if ( FindRange -> Values != NULL )
                                    { FindRange -> Values[ (*Found) ] = Leaf -> Value; }
                                }
                            }
                        else
                            {
                            //
                            //  We just count the 'Keys' and 'Values'.
                            //
                            for ( Count=Lowest;(Count <= Highest);Count ++,(*Found) ++ )
                                { /* void */ }
                            }

                        return (*Result);
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return True; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return ((*Result) = False);
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Find Stem.                                                                                */
            /*                                                                                            */
            /*  Find a stem in a branch.                                                                  */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::FindStem
            (
            BRANCH                              *Branch,
            KEY                                 *Key,
            BOOLEAN                             Nearest,
            SBIT16                              *Offset
            )
        {
        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
        AUTO SBIT16 Maximum = (Branch -> Used-1);
        AUTO SBIT16 Minimum = 0;

        //
        //  We can find a 'Stem' using a binary search as the 'Order' stores all the stems
        //  in sorted order.
        //
        while ( Maximum >= Minimum )
            {
            AUTO SBIT16 MidPoint = ((Maximum + Minimum) / 2);

            //
            //  Let's see if the 'MidPoint' value is greater than the target 'Key' value.
            //
            if ( GreaterThan( Key,& Branch -> Stem[ Order[ MidPoint ].IndexToValue ].Key ) )
                { Minimum = (MidPoint + 1); }
            else
                {
                //
                //  Let's see if the 'MidPoint' value is less than the target 'Key' value.
                //
                if ( LessThan( Key,& Branch -> Stem[ Order[ MidPoint ].IndexToValue ].Key ) )
                    { Maximum = (MidPoint - 1); }
                else
                    {
                    //
                    //  We have found the target 'Key' value.
                    //
                    (*Offset) = Order[ MidPoint ].IndexToValue;

                    return True;
                    }
                }
            }

        //
        //  When we are trying to find the 'Nearest' then the closest 'Key' less than the supplied
        //  'Key' is perfectly acceptable.  However, the 'Offset' must fall within the 'Stem' array
        //  to be considered a match.
        //
        if ( Nearest )
            {
            //
            //  When 'Nearest' is 'True' the target 'Key' will fall just before 'Minimum' and just
            //  after 'Maximum'. Consequently, we can set 'Offset' to '(Minimum-1)' as long as this
            //  falls within the 'Stem' array.
            //
            if ( (Minimum > 0) && (Minimum <= Branch -> Used) )
                {
                (*Offset) = Order[ (Minimum-1) ].IndexToValue;

                return True;
                }
            else
                { return False; }
            }
        else
            {
            //
            //  Even when we are not trying to find the 'Nearest' Key the 'Offset' must still fall
            //  within the 'Stem' array (i.e. zero to 'Branch -> Used') as we depend on the 'Offset'
            //  for placing any new keys.  The 'Offset' will point to the 'Key' just *after* the
            //  supplied 'Key' and therefore the supplied 'Key' will need to be inserted at the
            //  'Offset' in the array requiring anything after it to be moved down the array.
            //
            //
            (*Offset) = Minimum;

            return False;
            }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Branch.                                                                               */
            /*                                                                                            */
            /*  Create a new branch.                                                                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewBranch
            (
            FILE_ADDRESS                        *Address,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            BRANCH                              **Branch,
            SBIT32                              Depth
            )
        {
        AUTO FILE_SIZE Space;

        //
        //  We allocate a new 'Branch' on the heap.
        //
        if ( ROCKALL_NEW<BRANCH>::New( AutomaticHeapScope,Address,Branch,1,& Space,False,True ) )
            {
            AUTO SBIT32 ExtraBytes = ((SBIT32) (Space - sizeof(BRANCH)));
            AUTO SBIT32 StemSize = (sizeof(STEM) + sizeof(ORDER));
            AUTO SBIT32 Count;

            //
            //  Setup the new 'Branch'.
            //
            (*Branch) -> Type = TreeBranch;

            (*Branch) -> Address = (*Address);
            (*Branch) -> MaxSize = (STEMS + (ExtraBytes / StemSize));
            (*Branch) -> MaxSpace = ((SBIT32) Space);
            (*Branch) -> Depth = Depth;
            (*Branch) -> Used = 0;

            //
            //  We are limited to the size of 'SBIT16' by 'ORDER'.
            //
            if ( (*Branch) -> MaxSize > MaxSBit16 )
                { (*Branch) -> MaxSize = MaxSBit16; }

            //
            //  We now compute the ideal density range.
            //
            (*Branch) -> MaxDensity = ((SBIT16) (((*Branch) -> MaxSize * MaxStemDensity) / FullPage));
            (*Branch) -> MinDensity = ((SBIT16) (((*Branch) -> MaxSize * MinStemDensity) / FullPage));

            //
            //  Call any extra constructors.
            //
            for ( Count=STEMS;Count < (*Branch) -> MaxSize;Count ++ )
                { PLACEMENT_NEW( & (*Branch) -> Stem[ Count ],STEM ); }

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key.                                                                                  */
            /*                                                                                            */
            /*  Create a new key and value pair in the tree.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewKey
            (
            CONST KEY                           & Key,
            CONST VALUE                         & Value,
            BOOLEAN                             Overwrite
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (NewKey( & AutomaticHeapScope,Key,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key.                                                                                  */
            /*                                                                                            */
            /*  Create a new key and value pair in the tree.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            CONST VALUE                         & Value,
            BOOLEAN                             Overwrite
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()', 'ExclusiveView()',
        //  'Update()' and 'View()' should be automatically optimized away for heaps that do not
        //  support this functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO BOOLEAN SharedNavigation = True;

            //
            //  Clearly, it is much better to use 'SharedNavigation' where we can to avoid
            //  exclusive locks.
            //
            if ( SharedNavigation )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'View()' the 'Root' as we need to examine it.  We claimed the
                //  'TransactionLock' above to prevent threads in the same transaction
                //  from changing the tree.
                //
                if
                        (
                        (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                    AUTO NAVIGATION NextLevel =
                        {
                        AutomaticHeapScope,
                        & AutomaticShareableLock2,
                        & AutomaticViewScope,
                        NULL,
                        0,
                        Root -> MaxDepth,
                        ((KEY*) & Key),
                        True,
                        Overwrite,
                        Address,
                        & Root -> Trunk,
                        True,
                        Root,
                        & SharedNavigation,
                        Root -> Type,
                        ((VALUE*) & Value)
                        };

                    //
                    //  Create a new key and value pair in the tree.
                    //
                    if ( NewKeySharedNavigation( & NextLevel ) )
                        { return True; }
                    else
                        {
                        //
                        //  We switch to exclusive locks if required in this case
                        //  otherwise we exit.
                        //
                        if ( SharedNavigation )
                            { return (AutomaticHeapScope -> SetError( ErrorDuplicateKey )); }
                        }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }

            //
            //  We need to use exclusive locks if we need to alter any branches or the
            //  structure of the tree.
            //
            if ( ! SharedNavigation )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO ROOT *Root;

                //
                //  We 'View()' the 'Root' as we need to examine it.  We claimed the
                //  'TransactionLock' above to prevent threads in the same transaction
                //  from changing the tree.
                //
                if
                        (
                        (AutomaticViewScope.ExclusiveView( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );
                    AUTO NAVIGATION NextLevel =
                        {
                        AutomaticHeapScope,
                        & AutomaticExclusiveLock2,
                        & AutomaticViewScope,
                        NULL,
                        0,
                        Root -> MaxDepth,
                        ((KEY*) & Key),
                        True,
                        Overwrite,
                        Address,
                        & Root -> Trunk,
                        True,
                        Root,
                        NULL,
                        Root -> Type,
                        ((VALUE*) & Value)
                        };

                    //
                    //  When we have multiple threads in a single transaction we need to claim a
                    //  shareable 'TransactionLock' when we 'View()' areas of the data structure.
                    //  We claim an exclusive 'TransactionLock' to flush out all of these threads
                    //  and then claim an exclusive lock to prevent them from returning.  Once we
                    //  have claimed the exclusive lock we can release the 'TransactionLock' as
                    //  we no longer need it.
                    //
                    AutomaticExclusiveLock1.Release();

                    //
                    //  Create a new key and value pair in the tree.
                    //
                    if ( NewKeyExclusiveNavigation( & NextLevel ) )
                        { return True; }
                    else
                        { return (AutomaticHeapScope -> SetError( ErrorDuplicateKey )); }
                    }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
                }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key Exclusive Navigation.                                                             */
            /*                                                                                            */
            /*  Create a new key and value pair in the tree.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewKeyExclusiveNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        //
                        //  We can navigate towards the correct 'Twig' as long as the current
                        //  'Branch' is not full.  If not, we will need to rebalance the 'Branch'
                        //  before continuing.
                        //
                        if ( Branch -> Used < Branch -> MaxSize )
                            {
                            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                            AUTO SBIT16 Offset;

                            //
                            //  We only need to hold an 'ExclusiveView()' and a lock on our 'Parent'
                            //  until we have decided that we do no need to rebalance the tree. We
                            //  can now safely release the 'ExclusiveView()' and lock on our 'Parent'
                            //  as we now hold an 'ExclusiveView()' and a lock on the next level down.
                            //
                            ReleaseParent( Navigation,True );

                            //
                            //  When we are adding a new 'Key' it might be the lowest 'Key' in the
                            //  'Branch'.  If so, we will need to update the first 'Stem' of the
                            //  'Branch' with the new 'Key' value otherwise we will not be able to
                            //  find the 'Key' later.
                            //
                            if ( LessThan( Navigation -> Key,& Branch -> Stem[ Order[0].IndexToValue ].Key ) )
                                {
                                //
                                //  We need to 'Update()' the 'Branch'.
                                //
                                if ( AutomaticHeapScope -> Update( Address,((VOID**) & Branch) ) )
                                    { Branch -> Stem[ Order[0].IndexToValue ].Key = (*Navigation -> Key); }
                                }

                            //
                            //  We find the correct 'Stem' in the 'Branch'.
                            //
                            if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                                {
                                AUTO NAVIGATION NextLevel =
                                    {
                                    Navigation -> AutomaticHeapScope,
                                    & AutomaticExclusiveLock,
                                    & AutomaticViewScope,
                                    Branch,
                                    (Navigation -> Calls+1),
                                    (Navigation -> Depth-1),
                                    Navigation -> Key,
                                    (Order[ Offset ].ValueToOrder == (Branch -> Used-1)),
                                    Navigation -> Overwrite,
                                    Address,
                                    & Branch -> Stem[ Offset ].Address,
                                    True,
                                    Navigation -> Root,
                                    Navigation -> SharedNavigation,
                                    Branch -> Type,
                                    Navigation -> Value
                                    };

                                //
                                //  Navigate down to the next level of the tree.
                                //
                                return ( NewKeyExclusiveNavigation( & NextLevel ) );
                                }
                            }
                        else
                            {
                            //
                            //  We need to rebalance the 'Branch' to ensure there is enough space
                            //  for us to insert the new 'Key' and 'Value' pair. We then call ourselves
                            //  recursively as everything may have moved during the rebalancing so we
                            //  will need to find the right 'Branch' again.
                            //
                            if ( Rebalance( Branch,Navigation ) )
                                {
                                //
                                //  When we rebalance everthing can move.  Consequently, we can't
                                //  even depend on the work our 'Parent' did for us previously.
                                //  Here we redo the work that as been undone by rebalancing and then
                                //  simulate our parent calling us again.
                                //
                                switch ( Navigation -> Type )
                                    {
                                    case TreeBranch:
                                        {
                                        AUTO FILE_ADDRESS ParentAddress = Navigation -> ParentAddress;
                                        AUTO BRANCH *Parent = Navigation -> Branch;
                                        AUTO ORDER *Order = (ComputeOrderAddress( Parent ));
                                        AUTO KEY *FirstKey = & Parent -> Stem[ Order[0].IndexToValue ].Key;
                                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Parent -> Stem);
                                        AUTO SBIT16 Offset;

                                        //
                                        //  When we are adding a new 'Key' it might be the lowest
                                        //  'Key' in the 'Parent'.  If so, we will need to update
                                        //  the first 'Stem' of the 'Parent' with the new 'Key'
                                        //  value otherwise we will not be able to find the 'Key'
                                        //  later.
                                        //
                                        if ( LessThan( Navigation -> Key,FirstKey ) )
                                            {
                                            //
                                            //  We need to 'Update()' the 'Branch'.
                                            //
                                            if ( AutomaticHeapScope -> Update( ParentAddress,((VOID**) & Parent) ) )
                                                { (*FirstKey) = (*Navigation -> Key); }
                                            else
                                                { return False; }
                                            }

                                        //
                                        //  When we did the 'Rebalance()' all the 'Stems' may have
                                        //  moved around in our 'Parent' so we need to find the
                                        //  correct 'Stem' again.
                                        //
                                        if ( FindStem( Parent,Navigation -> Key,True,& Offset ) )
                                            {
                                            //
                                            //  We need to update 'Navigation' with the correct information
                                            //  as the 'Parent' may have changed during 'Rebalance()'.
                                            //
                                            Navigation -> Calls ++;
                                            Navigation -> Depth = (Parent -> Depth-1);
                                            Navigation -> LastStem = (Order[ Offset ].ValueToOrder == (Parent -> Used-1));
                                            Navigation -> Reference = & Parent -> Stem[ Offset ].Address;

                                            //
                                            //  We now simulate our parent calling us again.
                                            //
                                            return
                                                (
                                                (ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True ))
                                                    &&
                                                (NewKeyExclusiveNavigation( Navigation ))
                                                );
                                            }
                                        else
                                            { FAILURE( "Key present before rebalance but now missing" ); }

                                        break;
                                        }

                                    case TreeRoot:
                                        {
                                        //
                                        //  We need to update 'Navigation' with the correct information
                                        //  as the 'Parent' may have changed during 'Rebalance()'.
                                        //
                                        Navigation -> Calls ++;
                                        Navigation -> Depth = Navigation -> Root -> MaxDepth;

                                        //
                                        //  We now simulate our parent calling us again.
                                        //
                                        return
                                            (
                                            (ReleaseSelf( & AutomaticExclusiveLock,& AutomaticViewScope,True ))
                                                &&
                                            (NewKeyExclusiveNavigation( Navigation ))
                                            );
                                        }

                                    default:
                                        { return (FAILURE( "Unknown node type" )); }
                                    }
                                }
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);

                        //
                        //  We have reached a 'Twig' in the tree so let's see if there is enough space
                        //  to insert the new 'Key' and 'Value' pair.
                        //
                        if ( Twig -> Used < Twig -> MaxSize )
                            {
                            //
                            //  We now need to 'Update()' the 'Twig'.
                            //
                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                {
                                //
                                //  We can now insert the new 'Key' and 'Value' pair into the
                                //  existing 'Twig'.
                                //
                                if ( NewLeaf( Navigation -> Key,Navigation -> Overwrite,Twig,Navigation -> Value ) )
                                    { return True; }
                                }
                            }
                        else
                            {
                            AUTO FILE_ADDRESS ParentAddress = (Navigation -> ParentAddress);
                            AUTO BRANCH *Branch;

                            //
                            //  If the 'Twig' is full then we need to create a new 'Branch' so we can
                            //  use it. However, if the 'Parent' is already a 'Branch' we can split
                            //  the 'Twig' and use the new space.
                            //
                            switch ( Navigation -> Type )
                                {
                                case TreeBranch:
                                    {
                                    //
                                    //  If the 'Parent' is a 'Branch' we can split the 'Twig' as long
                                    //  as the 'Branch' has enough space.  The 'Branch should never run
                                    //  out of space because when it becomes full as the 'Branch' should
                                    //  get rebalanced before this occurs.
                                    //
                                    if
                                            (
                                            (AutomaticHeapScope -> Update( ParentAddress,((VOID**) & Branch) ))
                                                &&
                                            (AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ))
                                            )
                                        {
                                        //
                                        //  We now split the twig into two parts.
                                        //
                                        if ( SplitTwig( Navigation -> AutomaticHeapScope,Branch,Twig ) )
                                            {
                                            AUTO BRANCH *Parent = (Navigation -> Branch);
                                            AUTO SBIT16 Offset;

                                            //
                                            //  When we did the 'SplitTwig()' all the 'Stems' may
                                            //  have moved around in our 'Parent' so we need to
                                            //  find the correct 'Stem' again.
                                            //
                                            if ( FindStem( Parent,Navigation -> Key,True,& Offset ) )
                                                {
                                                //
                                                //  We need to update 'Navigation' with the
                                                //  correct information as the 'Parent' may have
                                                //  changed during 'SplitTwig()'.
                                                //
                                                Navigation -> Calls ++;
                                                Navigation -> Reference = & Parent -> Stem[ Offset ].Address;

                                                //
                                                //  We now simulate our parent calling us again.
                                                //
                                                return
                                                    (
                                                    (ReleaseSelf( & AutomaticExclusiveLock,NULL,True ))
                                                        &&
                                                    (NewKeyExclusiveNavigation( Navigation ))
                                                    );
                                                }
                                            }
                                        }

                                    break;
                                    }

                                case TreeRoot:
                                    {
                                    //
                                    //  We need to create the first 'Branch' in the tree and link it
                                    //  to the the 'Root'.  Next, we need to add a 'Stem' pointing to
                                    //  the existing 'Twig'. Next, we need to rebalance the twigs to
                                    //  create the extra space we need.  Finally, we call ourselves
                                    //  recursively as everything may have moved around during the
                                    //  rebalancing, so we will need to find the right 'Twig' again.
                                    //
                                    if ( AutomaticHeapScope -> Update( ParentAddress,((VOID**) & Navigation -> Root) ) )
                                        {
                                        AUTO SBIT32 NewDepth = (Navigation -> Root -> MaxDepth+1);

                                        //
                                        //  Create the first 'Branch' in the tree.
                                        //
                                        if ( NewBranch( Navigation -> Reference,AutomaticHeapScope,& Branch,NewDepth ) )
                                            {
                                            //
                                            //  We now need to 'Update()' the 'Twig'.
                                            //
                                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                                {
                                                AUTO LEAF *Leaf = & Twig -> Leaf[ Order[0].IndexToValue ];

                                                //
                                                //  Create the first 'Stem' in the tree.
                                                //
                                                if ( NewStem( Address,Branch,& Leaf -> Key ) )
                                                    {
                                                    //
                                                    //  We need to update 'Navigation' with the
                                                    //  correct information as the 'Parent' may have
                                                    //  changed during processing.
                                                    //
                                                    Navigation -> Calls ++;
                                                    Navigation -> Depth = NewDepth;
                                                    Navigation -> Root -> MaxDepth = NewDepth;

                                                    //
                                                    //  Now we can 'Rebalance()' the tree to ensure that
                                                    //  there is enough space.
                                                    //
                                                    return
                                                        (
                                                        (RebalanceTwigs( Branch,Navigation,Navigation -> Root ))
                                                            &&
                                                        (ReleaseSelf( & AutomaticExclusiveLock,NULL,True ))
                                                            &&
                                                        (NewKeyExclusiveNavigation( Navigation ))
                                                        );
                                                    }
                                                else
                                                    { FAILURE( "Unable to create new stem" ); }
                                                }
                                            }
                                        }

                                    break;
                                    }

                                default:
                                    { return (FAILURE( "Unknown node type" )); }
                                }
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  When a tree is empty we simply create a new 'Twig' and insert the new 'Key'
            //  and 'Value' pair into it.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                {
                AUTO FILE_ADDRESS ParentAddress = (Navigation -> ParentAddress);

                //
                //  We need to make changes to the 'Root'.
                //
                if ( AutomaticHeapScope -> Update( ParentAddress,((VOID**) & Navigation -> Root) ) )
                    {
                    AUTO TWIG *Twig;

                    //
                    //  Create the first 'Twig' in the tree.
                    //
                    if ( NewTwig( Navigation -> Reference,Navigation -> AutomaticHeapScope,& Twig ) )
                        {
                        //
                        //  Create the first 'Leaf' in the tree.
                        //
                        if ( NewLeaf( Navigation -> Key,Navigation -> Overwrite,Twig,Navigation -> Value ) )
                            {
                            //
                            //  Update the 'MaxDepth'.
                            //
                            Navigation -> Root -> MaxDepth ++;

                            return True;
                            }
                        }
                    }
                }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Key Shared Navigation.                                                                */
            /*                                                                                            */
            /*  Create a new key and value pair in the tree.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewKeySharedNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        //
                        //  We can navigate towards the correct 'Twig' as long as the current
                        //  'Branch' is not full.  If not, we will need to rebalance the 'Branch'
                        //  before continuing.
                        //
                        if ( Branch -> Used < Branch -> MaxSize )
                            {
                            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                            AUTO SBIT16 Offset;

                            //
                            //  We only need to hold an 'View()' and a lock on our 'Parent' until
                            //  we have decided that we do no need to rebalance the tree. We can
                            //  now safely release the 'View()' and lock on our 'Parent' as we now
                            //  hold a 'View()' and a lock on the next level down.
                            //
                            ReleaseParent( Navigation,True );

                            //
                            //  When we are adding a new 'Key' it might be the lowest 'Key' in the
                            //  'Branch'.  If so, we will need to update the first 'Stem' of the
                            //  'Branch' with the new 'Key' value otherwise we will not be able to
                            //  find the 'Key' later.  However, can't do this when we are using
                            //  'SharedNavigation' so we exit.
                            //
                            if ( GreaterOrEqual( Navigation -> Key,& Branch -> Stem[ Order[0].IndexToValue ].Key ) )
                                {
                                //
                                //  We find the correct 'Stem' in the 'Branch'.
                                //
                                if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                                    {
                                    AUTO NAVIGATION NextLevel =
                                        {
                                        Navigation -> AutomaticHeapScope,
                                        & AutomaticShareableLock,
                                        & AutomaticViewScope,
                                        Branch,
                                        (Navigation -> Calls+1),
                                        (Navigation -> Depth-1),
                                        Navigation -> Key,
                                        (Order[ Offset ].ValueToOrder == (Branch -> Used-1)),
                                        Navigation -> Overwrite,
                                        Address,
                                        & Branch -> Stem[ Offset ].Address,
                                        True,
                                        Navigation -> Root,
                                        Navigation -> SharedNavigation,
                                        Branch -> Type,
                                        Navigation -> Value
                                        };

                                    //
                                    //  Navigate down to the next level of the tree.
                                    //
                                    return ( NewKeySharedNavigation( & NextLevel ) );
                                    }
                                }
                            else
                                { return ((*Navigation -> SharedNavigation) = False); }
                            }
                        else
                            { return ((*Navigation -> SharedNavigation) = False); }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);

                        //
                        //  We have reached a 'Twig' in the tree so let's see if there is enough space
                        //  to insert the new 'Key' and 'Value' pair.
                        //
                        if ( Twig -> Used < Twig -> MaxSize )
                            {
                            //
                            //  We now need to 'Update()' the 'Twig'.
                            //
                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                {
                                //
                                //  We can now insert the new 'Key' and 'Value' pair into the
                                //  existing 'Twig'.
                                //
                                if ( NewLeaf( Navigation -> Key,Navigation -> Overwrite,Twig,Navigation -> Value ) )
                                    { return True; }
                                }
                            }
                        else
                            { return ((*Navigation -> SharedNavigation) = False); }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            { return ((*Navigation -> SharedNavigation) = False); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Leaf.                                                                                 */
            /*                                                                                            */
            /*  A new leaf is added into the node.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewLeaf
            (
            KEY                                 *Key,
            BOOLEAN                             Overwrite,
            TWIG                                *Twig,
            VALUE                               *Value
            )
        {
        AUTO SBIT16 Offset;

        //
        //  We verify that there is space in the 'Twig' and that there is no matching 'Key'
        //  in the 'Twig'.
        //
        if ( (! FindLeaf( Key,& Offset,False,Twig )) && (Twig -> Used < Twig -> MaxSize) )
            {
            AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
            AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
            AUTO ORDER *Current1 = & Order[ Offset ];
            AUTO ORDER *Current2 = & Order[ Twig -> Used ];
            AUTO LEAF *Current3 = & Twig -> Leaf[ Twig -> Used ];
            AUTO SBIT16 Count;

            //
            //  We slide the values in the sorted 'Order' table along so we can place the
            //  reference to the 'Key' and 'Value' pair in the correct place in the table.
            //
            for ( Count=Twig -> Used;Count > Offset;Count -- )
                {
                AUTO ORDER *Current4 = & Order[ Count ];
                AUTO ORDER *Current5 = & Order[ (Count-1) ];
                AUTO ORDER *Current6 = & Order[ Current5 -> IndexToValue ];

                //
                //  We simply move the 'IndexToValue' up one element in the 'Order'
                //  table fixing the associated 'ValueToOrder' for every entry we move.
                //
                Current4 -> IndexToValue = Current5 -> IndexToValue;
                Current6 -> ValueToOrder = Count;
                }

            //
            //  We insert the new 'Key' and 'Value' pair value in the correct place
            //  in the 'Order' values (i.e. in sorted order).  The actual 'Key' and
            //  'Value' will be stored at the end of the 'Leaf' table and the 'Order'
            //  table allows us to navigate from 'IndexToValue' and 'ValueToOrder'.
            //
            Current1 -> IndexToValue = Twig -> Used ++;
            Current2 -> ValueToOrder = Offset;

            //
            //  We update the end of the 'Leaf' table with the new 'Key' and 'Value'.
            //
            Current3 -> Key = (*Key);
            Current3 -> Value = (*Value);

            return True;
            }
        else
            {
            //
            //  We will 'Overwrite' any existing value if needed.
            //
            if ( Overwrite )
                {
                Twig -> Leaf[ Offset ].Value = (*Value);

                return True;
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Stem.                                                                                 */
            /*                                                                                            */
            /*  A new stem is added into a branch.                                                        */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewStem
            (
            FILE_ADDRESS                        Address,
            BRANCH                              *Branch,
            KEY                                 *Key
            )
        {
        AUTO SBIT16 Offset;

        //
        //  We verify that there is space in the 'Branch' and then find the closest existing
        //  'Key' that is less than the new 'Key'.
        //
        if ( (Branch -> Used < Branch -> MaxSize) && (! FindStem( Branch,Key,False,& Offset )) )
            {
            AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
            AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
            AUTO ORDER *Current1 = & Order[ Offset ];
            AUTO ORDER *Current2 = & Order[ Branch -> Used ];
            AUTO STEM *Current3 = & Branch -> Stem[ Branch -> Used ];
            AUTO SBIT16 Count;

            //
            //  We slide the values in the sorted 'Order' table along so we can place the
            //  reference to the 'Address' and 'Key' pair in the correct place in the table.
            //
            for ( Count=Branch -> Used;Count > Offset;Count -- )
                {
                AUTO ORDER *Current4 = & Order[ Count ];
                AUTO ORDER *Current5 = & Order[ (Count-1) ];
                AUTO ORDER *Current6 = & Order[ Current5 -> IndexToValue ];

                //
                //  We simply move the 'IndexToValue' up one element in the 'Order'
                //  table fixing the associated 'ValueToOrder' for every entry we move.
                //
                Current4 -> IndexToValue = Current5 -> IndexToValue;
                Current6 -> ValueToOrder = Count;
                }

            //
            //  We insert the new 'Address' and 'Key' pair value in the correct place
            //  in the 'Order' values (i.e. in sorted order).  The actual 'Address' and
            //  'Key' will be stored at the end of the 'Stem' table and the 'Order' table
            //  allows us to navigate from 'IndexToValue' and 'ValueToOrder'.
            //
            Current1 -> IndexToValue = (Branch -> Used ++);
            Current2 -> ValueToOrder = Offset;

            //
            //  We update the end of the 'Stem' table with the new 'Address' and 'Key'.
            //
            Current3 -> Address = Address;
            Current3 -> Key = (*Key);

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  New Twig.                                                                                 */
            /*                                                                                            */
            /*  A new collection of empty leaves is made.                                                 */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::NewTwig
            (
            FILE_ADDRESS                        *Address,
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TWIG                                **Twig
            )
        {
        AUTO FILE_SIZE Space;

        //
        //  We allocate a new 'Twig' on the heap.
        //
        if ( ROCKALL_NEW<TWIG>::New( AutomaticHeapScope,Address,Twig,1,& Space,False,True ) )
            {
            AUTO SBIT32 ExtraBytes = ((SBIT32) (Space - sizeof(TWIG)));
            AUTO SBIT32 LeafSize = (sizeof(LEAF) + sizeof(ORDER));
            AUTO SBIT32 Count;

            //
            //  Setup the new 'Branch'.
            //
            (*Twig) -> Type = TreeTwig;

            (*Twig) -> Address = (*Address);
            (*Twig) -> MaxSize = (LEAVES + (ExtraBytes / LeafSize));
            (*Twig) -> MaxSpace = ((SBIT32) Space);
            (*Twig) -> Used = 0;

            //
            //  We are limited to the size of 'SBIT16' by 'ORDER'.
            //
            if ( (*Twig) -> MaxSize > MaxSBit16 )
                { (*Twig) -> MaxSize = MaxSBit16; }

            //
            //  We now compute the ideal density range.
            //
            (*Twig) -> MaxDensity = ((SBIT16) (((*Twig) -> MaxSize * MaxLeafDensity) / FullPage));
            (*Twig) -> MinDensity = ((SBIT16) (((*Twig) -> MaxSize * MinLeafDensity) / FullPage));

            //
            //  Call any extra constructors.
            //
            for ( Count=LEAVES;Count < (*Twig) -> MaxSize;Count ++ )
                { PLACEMENT_NEW( & (*Twig) -> Leaf[ Count ],LEAF ); }

            return True;
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rebalance.                                                                                */
            /*                                                                                            */
            /*  The current level of the tree is rebalanced.                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::Rebalance
            (
            BRANCH                              *Branch,
            NAVIGATION                          *Navigation
            )
        {
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));
        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
            AUTO FILE_ADDRESS ParentAddress = (Navigation -> ParentAddress);

            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and
            //  'Type' of node in the tree.
            //
            if
                    (
                    (Branch -> Address == Address)
                        &&
                    (Branch -> Depth == Navigation -> Depth)
                        &&
                    (Branch -> Type == TreeBranch)
                        &&
                    (Branch -> Used > 0)
                    )
                {
                //
                //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
                //  helpful and allows us to claim the correct type of access and locks.  If the
                //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
                //  a 'Twig' and zero the tree must be empty.
                //
                if ( Navigation -> Depth > 2 )
                    {
                    //
                    //  We may need to 'Update()' the 'Parent' but the method for this is
                    //  different if the 'Parent' is the 'Root', so care must be taken.
                    //
                    switch ( Navigation -> Type )
                        {
                        case TreeBranch:
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                            AUTO BRANCH *Above;
                            AUTO BRANCH *Below;

                            //
                            //  The 'Parent' is a 'Branch' and below us are more branches.
                            //  We are full so we will rebalance our children and then our
                            //  'Parent' which should give us a lot more space.
                            //
                            return
                                (
                                (AutomaticViewScope1.ExclusiveView( ParentAddress,((VOID**) & Above) ))
                                    &&
                                (AutomaticViewScope2.ExclusiveView( Address,((VOID**) & Below) ))
                                    &&
                                (RebalanceStems( Below,Navigation,NULL ))
                                    &&
                                (RebalanceStems( Above,Navigation,NULL ))
                                );
                            }

                        case TreeRoot:
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                            AUTO BRANCH *Below;

                            //
                            //  The 'Parent' is the 'Root' and below are more branches.
                            //  We are full so we will lock the 'Root' and ourseleves
                            //  into memory and rebalance.  This may create a new level
                            //  in the tree.
                            //
                            return
                                (
                                (AutomaticViewScope1.ExclusiveView( ParentAddress,((VOID**) & Navigation -> Root) ))
                                    &&
                                (AutomaticViewScope2.ExclusiveView( Address,((VOID**) & Below) ))
                                    &&
                                (RebalanceStems( Below,Navigation,Navigation -> Root ))
                                );
                            }

                        default:
                            { return (FAILURE( "Unknown node type" )); }
                        }
                    }
                else
                    {
                    //
                    //  We need to lock the 'Parent' for 'Update()' but the method for this
                    //  is different if the 'Parent' is the 'Root', so care must be taken.
                    //
                    switch ( Navigation -> Type )
                        {
                        case TreeBranch:
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                            AUTO BRANCH *Above;
                            AUTO BRANCH *Below;

                            //
                            //  The 'Parent' is a 'Branch' and below us are twigs.  We
                            //  are full so we will rebalance our children and then our
                            //  'Parent' which should give us a lot more space.
                            //
                            return
                                (
                                (AutomaticViewScope1.ExclusiveView( ParentAddress,((VOID**) & Above) ))
                                    &&
                                (AutomaticViewScope2.ExclusiveView( Address,((VOID**) & Below) ))
                                    &&
                                (RebalanceTwigs( Below,Navigation,NULL ))
                                    &&
                                (RebalanceStems( Above,Navigation,NULL ))
                                );
                            }

                        case TreeRoot:
                            {
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope1( AutomaticHeapScope );
                            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope2( AutomaticHeapScope );
                            AUTO BRANCH *Below;

                            //
                            //  The 'Parent' is the 'Root' and below us are twigs.
                            //  We are full so we will lock the 'Root' and ourseleves
                            //  into memory and rebalance.  This may create a new level
                            //  in the tree.
                            //
                            return
                                (
                                (AutomaticViewScope1.ExclusiveView( ParentAddress,((VOID**) & Navigation -> Root) ))
                                    &&
                                (AutomaticViewScope2.ExclusiveView( Address,((VOID**) & Below) ))
                                    &&
                                (RebalanceTwigs( Below,Navigation,Navigation -> Root ))
                                );
                            }

                        default:
                            { return (FAILURE( "Unknown node type" )); }
                        }
                    }
                }
            else
                { FAILURE( "Malformed tree" ); }
            }
        else
            { FAILURE( "Malformed tree" ); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rebalance Stems.                                                                          */
            /*                                                                                            */
            /*  All the stems within the sub-branches of a branch are rebalanced.                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::RebalanceStems
            (
            BRANCH                              *Branch,
            NAVIGATION                          *Navigation,
            ROOT                                *Root
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO ORDER *Order1 = (ComputeOrderAddress( Branch ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order1);
        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
        AUTO SBIT32 ErrorCode = NoError;
        AUTO BOOLEAN Result = True;
        AUTO BRANCH **SubBranches;

        //
        //  We need an array to keep track of all of the related 'SubBranches'.
        //
        if ( ROCKALL_NEW<BRANCH*>::New( AutomaticHeapScope,NULL,& SubBranches,Branch -> MaxSize,NULL,False,True ) )
            {
            AUTO SBIT32 Count1;
            AUTO SBIT32 IdealStems;
            AUTO SBIT32 TotalSize;
            AUTO SBIT32 TotalUsed;
            AUTO STEM *Stems;

            //
            //  We now 'Touch()' all the 'SubBranches' to ensure they are all in main memory.
            //
            for ( Count1=0;Count1 < Branch -> Used;Count1 ++ )
                { AutomaticHeapScope -> Touch( Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address ); }

            //
            //  We need to count the total number of 'SubBranches' in the 'Branch'.
            //
            for ( Count1=0,TotalSize=0,TotalUsed=0;Count1 < Branch -> Used;Count1 ++ )
                {
                AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;
                AUTO BRANCH **Current2 = & SubBranches[ Count1 ];

                //
                //  We may need to update every 'Stem' as part of balancing the 'Branch'.
                //
                if ( AutomaticHeapScope -> ExclusiveView( Current1,((VOID**) Current2) ) )
                    {
                    //
                    //  Count the total number of 'SubBranches' available and used.
                    //
                    if ( (*Current2) -> Type == TreeBranch )
                        {
                        TotalSize += (*Current2) -> MaxSize;
                        TotalUsed += (*Current2) -> Used;
                        }
                    else
                        { FAILURE( "Tree branch expected" ); }
                    }
                else
                    {
                    //
                    //  We call 'EndExclusiveView()' for all the 'SubBranches' if we can't
                    //  lock the whole 'Branch'.
                    //
                    for ( Count1 --;Count1 >= 0;Count1 -- )
                        {
                        AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;

                        if ( ! AutomaticHeapScope -> EndExclusiveView( Current1,False ) )
                            { FAILURE( "Unable to end exclusive view" ); }
                        }

                    Result = False;

                    break;
                    }
                }

            //
            //  Now that we have an 'ExclusiveView()' on all the whole 'Branch' we can
            //  call 'Update()'.
            //
            if ( Result )
                {
                //
                //  Let's 'Update()' the 'Root' if needed.
                //
                if ( Root != NULL )
                    {
                    if ( ! AutomaticHeapScope -> Update( Root -> Address,((VOID**) & Root) ) )
                        { FAILURE( "Unable to update branch" ); }
                    }

                //
                //  Let's 'Update()' the 'Branch'.
                //
                if ( AutomaticHeapScope -> Update( Branch -> Address,((VOID**) & Branch) ) )
                    {
                    for ( Count1=0;Count1 < Branch -> Used;Count1 ++ )
                        {
                        AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;
                        AUTO BRANCH **Current2 = & SubBranches[ Count1 ];

                        //
                        //  Let's 'Update()' all the 'SubBranches'.
                        //
                        if ( ! AutomaticHeapScope -> Update( Current1,((VOID**) Current2) ) )
                            { FAILURE( "Unable to update sub-branch" ); }
                        }
                    }
                else
                    { FAILURE( "Unable to update branch" ); }
                }

            //
            //  Let's ensure that all the 'SubBranches' are ready for updating.
            //
            if ( Result )
                {
                AUTO SBIT16 MaxDensity = Branch -> MaxDensity;
                AUTO SBIT32 IdealBranches = ((TotalUsed + (MaxDensity-1)) / MaxDensity);
                AUTO SBIT32 NewDepth = (Branch -> Depth-1);

                //
                //  We will add extra 'SubBranches' to the 'Branch' if the density exceeds the
                //  desired threashold.
                //
                for ( Count1=0;(Branch -> Used < IdealBranches) && (Branch -> Used < Branch -> MaxSize);Count1 ++ )
                    {
                    AUTO STEM *Current1 = & Branch -> Stem[ Branch -> Used ];
                    AUTO BRANCH **Current2 = & SubBranches[ Branch -> Used ];

                    //
                    //  Create a new 'Branch' and attach it to the tree.
                    //
                    if ( NewBranch( & Current1 -> Address,AutomaticHeapScope,Current2,NewDepth ) )
                        {
                        //
                        //  Update the 'TotalSize' and the number of twigs on the branch.
                        //
                        TotalSize += (*Current2) -> MaxSize;

                        Branch -> Used ++;
                        }
                    else
                        {
                        //
                        //  We will carry on balancing the 'Branch' if we were able to add
                        //  some extra 'SubBranches'.  If not, we exit as we might run out
                        //  of space later.
                        //
                        Result = (Count1 > 0);

                        break;
                        }
                    }

                //
                //  We now compute the 'IdealStems' for each of the 'SubBranches'.
                //
                if ( Branch -> Used < IdealBranches )
                    { IdealStems = ((TotalUsed + (Branch -> Used-1)) / Branch -> Used); }
                else
                    { IdealStems = MaxDensity; }
                }

            //
            //  We need an array to store all of the 'Stems' in the 'Branch'.  We copy all of the
            //  'Stems' into this array in sorted order to simplify copying them back later.
            //
            if
                    (
                    (Result)
                        &&
                    (ROCKALL_NEW<STEM>::New( AutomaticHeapScope,NULL,& Stems,TotalUsed,NULL,False,True ))
                    )
                {
                AUTO SBIT32 Count2;
                AUTO SBIT32 Count3;
                AUTO SBIT32 Count4;

                //
                //  We walk through all the 'SubBranches' copying each 'Stem' into our new array.
                //
                for ( Count1=0,Count3=0;Count1 < Branch -> Used;Count1 ++ )
                    {
                    AUTO BRANCH *Current1 = SubBranches[ Count1 ];
                    AUTO ORDER *Order2 = (ComputeOrderAddress( Current1 ));
                    AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order2);
                    AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Current1 -> Stem);

                    //
                    //  We copy every 'Stem' from a sub-branch into the array in sorted order.
                    //
                    for ( Count2=0;Count2 < Current1 -> Used;Count2 ++ )
                        { Stems[ (Count3 ++) ] = Current1 -> Stem[ Order2[ Count2 ].IndexToValue ]; }
                    }

                //
                //  We have a great opportunity to verify that 'Key' values are in sorted order.
                //
                for ( Count1=1;Count1 < Count3;Count1 ++ )
                    {
                    if ( GreaterOrEqual( & Stems[ (Count1-1) ].Key,& Stems[ Count1 ].Key ) )
                        { FAILURE( "Stems not in sorted order" ); }
                    }

                //
                //  We now copy each 'Stem' in our array back into a suitable sub-branch so as to
                //  balance and redistribute all the 'Stems' amoung the 'SubBranches'.
                //
                for ( Count1=0,Count4=0;Count1 < Branch -> Used;Count1 ++ )
                    {
                    AUTO BRANCH *Current1 = SubBranches[ Count1 ];
                    AUTO ORDER *Order2 = (ComputeOrderAddress( Current1 ));
                    AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order2);
                    AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Current1 -> Stem);

                    //
                    //  We copy the 'IdealStems' amount of 'Stems' into each sub-branch.  If we
                    //  run out of 'Stems' we will leave the final 'SubBranches' empty (or at
                    //  least less full).
                    //
                    for ( Count2=0;(Count2 < IdealStems) && (Count3 > Count4);Count2 ++ )
                        {
                        AUTO ORDER *Current2 = & Order2[ Count2 ];

                        Current1 -> Stem[ Count2 ] = Stems[ (Count4 ++) ];

                        Current2 -> IndexToValue = Count2;
                        Current2 -> ValueToOrder = Count2;
                        }

                    //
                    //  We zero any unused 'Stem' just to be tidy.
                    //
                    if ( Count2 < Current1 -> Used )
                        {
                        AUTO SBIT16 Spare = (Current1 -> Used - Count2);

                        MemoryFunctions.Zero( & Order2[ Count2 ],(Spare * sizeof(ORDER)) );
                        ZeroStems( & Current1 -> Stem[ Count2 ],Spare );
                        }

                    //
                    //  Finally, we update the number of 'Stems' in the sub-branch.
                    //
                    Current1 -> Used = Count2;
                    }

                //
                //  When we move the 'Stems' between 'SubBranches' we invalidate the 'Key' in the
                //  related 'Stem'.  We repair this damage by updating the 'Key' in the 'Stem'
                //  with the first 'Key' (i.e. the lowest) in the sub-branch.  If the sub-branch
                //  is empty we delete it to recover any associated space.
                //
                for ( Count1=(Branch -> Used-1);Count1 >= 0;Count1 -- )
                    {
                    AUTO STEM *Current1 = & Branch -> Stem[ Count1 ];
                    AUTO BRANCH *Current2 = SubBranches[ Count1 ];
                    AUTO ORDER *Current3 = & Order1[ Count1 ];

                    //
                    //  We update the 'Stem' if it is not empty.  If not, otherwise we delete it.
                    //
                    if ( Current2 -> Used > 0 )
                        {
                        //
                        //  We update the 'Stem' as the 'SubBranches' have been adjusted.
                        //
                        Current1 -> Key = Current2 -> Stem[0].Key;
                        Current1 -> Address = Current2 -> Address;

                        Current3 -> IndexToValue = Count1;
                        Current3 -> ValueToOrder = Count1;
                        }
                    else
                        {
                        //
                        //  We no longer need the 'Stem' so delete it.
                        //
                        DeleteBranch( AutomaticHeapScope,Current2 );

                        MemoryFunctions.Zero( Current3,sizeof(ORDER) );
                        ZeroStems( Current1,1 );

                        Branch -> Used --;
                        }
                    }

                //
                //  We will add a whole extra level into the tree if the 'Density' of the 'Branch'
                //  rises above the associated threashold.  We do not let the 'Branch' become full
                //  because this will lead to repeated rebalancing as we near its capacity because
                //  we will keep trying to squeeze more and more into the 'SubBranches'.  Instead,
                //  we give up early so as to leave plenty of space.  We will only do this at the
                //  top level, so as to ensure that the tree always remains balanced.  We know that
                //  each level of the tree will balance itself when it becomes full, so it can never
                //  be necessary to add a new level anywhere else in the tree.
                //
                if ( (Root != NULL) && (Branch -> Used > Branch -> MaxDensity) )
                    {
                    AUTO BRANCH *NewLevel;

                    //
                    //  Create a new branch above the current branch.
                    //
                    if ( NewBranch( & Root -> Trunk,AutomaticHeapScope,& NewLevel,(Root -> MaxDepth+1) ) )
                        {
                        //
                        //  Create a new stem in the new branch refering to the current branch.
                        //
                        if ( NewStem( Branch -> Address,NewLevel,& Branch -> Stem[0].Key ) )
                            {
                            //
                            //  Update the 'MaxDepth' of the tree.
                            //
                            Root -> MaxDepth ++;

                            //
                            //  We rebalance the new branch to make sure we don't immediately
                            //  delete it (as it only has one stem).
                            //
                            Result = (RebalanceStems( NewLevel,Navigation,NULL ));
                            }
                        else
                            { FAILURE( "Unable to add a stem to new branch" ); }
                        }
                    else
                        { Result = False; }
                    }

                //
                //  We will delete the current 'Branch' if only has one 'Stem' and we are currently
                //  at the top level of the tree.  Clearly, any 'Branch' that does not have more than
                //  one 'Stem' is pointless and so there is no reason to keep it.
                //
                if ( (Root != NULL) && (Branch -> Used == 1) )
                    {
                    //
                    //  Update the 'Root' and delete the 'Branch'.
                    //
                    Root -> MaxDepth --;

                    Root -> Trunk = Branch -> Stem[0].Address;

                    DeleteBranch( AutomaticHeapScope,Branch );
                    }

                //
                //  We can now delete the array be created earlier but we must be sure to
                //  preserve any 'ErrorCode'.
                //
                if
                        (
                        (AutomaticHeapScope -> GetError( & ErrorCode ), False)
                            ||
                        (! ROCKALL_DELETE<STEM>::Delete( AutomaticHeapScope,Stems,TotalUsed ))
                            ||
                        (AutomaticHeapScope -> SetError( ErrorCode ), False)
                        )
                    { FAILURE( "Unable to delete stems" ); }
                }
            else
                { Result = False; }

            //
            //  We can now delete the array be created earlier but we must be sure to
            //  preserve any 'ErrorCode'.
            //
            if
                    (
                    (AutomaticHeapScope -> GetError( & ErrorCode ), False)
                        ||
                    (! ROCKALL_DELETE<BRANCH*>::Delete( AutomaticHeapScope,SubBranches,Branch -> MaxSize ))
                        ||
                    (AutomaticHeapScope -> SetError( ErrorCode ), False)
                    )
                { FAILURE( "Unable to delete sub-branches" ); }
            }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rebalance Twigs.                                                                          */
            /*                                                                                            */
            /*  All the twigs within a branch are rebalanced.                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::RebalanceTwigs
            (
            BRANCH                              *Branch,
            NAVIGATION                          *Navigation,
            ROOT                                *Root
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO ORDER *Order1 = (ComputeOrderAddress( Branch ));
        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order1);
        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
        AUTO SBIT32 ErrorCode = NoError;
        AUTO BOOLEAN Result = True;
        AUTO TWIG **Twigs;

        //
        //  We need an array to keep track of all of the related twigs.
        //
        if ( ROCKALL_NEW<TWIG*>::New( AutomaticHeapScope,NULL,& Twigs,Branch -> MaxSize,NULL,False,True ) )
            {
            AUTO SBIT32 Count1;
            AUTO SBIT32 IdealLeaves;
            AUTO SBIT32 TotalSize;
            AUTO SBIT32 TotalUsed;
            AUTO LEAF *Leaves;

            //
            //  We now 'Touch()' all the 'Leaves' to ensure they are all in main memory.
            //
            for ( Count1=0;Count1 < Branch -> Used;Count1 ++ )
                { AutomaticHeapScope -> Touch( Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address ); }

            //
            //  We need to count the total number of leaves on the branch.
            //
            for ( Count1=0,TotalSize=0,TotalUsed=0;Count1 < Branch -> Used;Count1 ++ )
                {
                AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;
                AUTO TWIG **Current2 = & Twigs[ Count1 ];

                //
                //  We may need to update every 'Twig' as part of balancing the branch.
                //
                if ( AutomaticHeapScope -> ExclusiveView( Current1,((VOID**) Current2) ) )
                    {
                    //
                    //  Count the total number of leaves available and used.
                    //
                    if ( (*Current2) -> Type == TreeTwig )
                        {
                        TotalSize += (*Current2) -> MaxSize;
                        TotalUsed += (*Current2) -> Used;
                        }
                    else
                        { FAILURE( "Tree twig expected" ); }
                    }
                else
                    {
                    //
                    //  We call 'EndExclusiveView()' for all the 'Twigs' if we can't lock
                    //  the whole 'Branch'.
                    //
                    for ( Count1 --;Count1 >= 0;Count1 -- )
                        {
                        AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;

                        if ( ! AutomaticHeapScope -> EndExclusiveView( Current1,False ) )
                            { FAILURE( "Unable to end exclusive view" ); }
                        }

                    Result = False;

                    break;
                    }
                }

            //
            //  Now that we have an 'ExclusiveView()' on all the whole 'Branch' we can
            //  call 'Update()'.
            //
            if ( Result )
                {
                //
                //  Let's 'Update()' the 'Root' if needed.
                //
                if ( Root != NULL )
                    {
                    if ( ! AutomaticHeapScope -> Update( Root -> Address,((VOID**) & Root) ) )
                        { FAILURE( "Unable to update branch" ); }
                    }

                //
                //  Let's 'Update()' the 'Branch'.
                //
                if ( AutomaticHeapScope -> Update( Branch -> Address,((VOID**) & Branch) ) )
                    {
                    for ( Count1=0;Count1 < Branch -> Used;Count1 ++ )
                        {
                        AUTO FILE_ADDRESS Current1 = Branch -> Stem[ Order1[ Count1 ].IndexToValue ].Address;
                        AUTO TWIG **Current2 = & Twigs[ Count1 ];

                        //
                        //  Let's 'Update()' all the 'Twogs'.
                        //
                        if ( ! AutomaticHeapScope -> Update( Current1,((VOID**) Current2) ) )
                            { FAILURE( "Unable to update twig" ); }
                        }
                    }
                else
                    { FAILURE( "Unable to update branch" ); }
                }

            //
            //  Let's ensure that all the 'Twigs' are ready for updating.
            //
            if ( Result )
                {
                AUTO SBIT16 MaxDensity = Twigs[0] -> MaxDensity;
                AUTO SBIT32 IdealTwigs = ((TotalUsed + (MaxDensity-1)) / MaxDensity);

                //
                //  We will add extra 'Twigs' to the 'Branch' if the density exceeds the
                //  desired threashold.
                //
                for ( Count1=0;(Branch -> Used < IdealTwigs) && (Branch -> Used < Branch -> MaxSize);Count1 ++ )
                    {
                    AUTO STEM *Current1 = & Branch -> Stem[ Branch -> Used ];
                    AUTO TWIG **Current2 = & Twigs[ Branch -> Used ];

                    //
                    //  Create a new 'Twig' and attach it to the 'Branch'.
                    //
                    if ( NewTwig( & Current1 -> Address,AutomaticHeapScope,Current2 ) )
                        {
                        //
                        //  Update the 'TotalSize' and the number of twigs on the branch.
                        //
                        TotalSize += (*Current2) -> MaxSize;

                        Branch -> Used ++;
                        }
                    else
                        {
                        //
                        //  We will carry on balancing the branch if we were able to add at least
                        //  one new twig.  If not, we exit as we might run out of space later.
                        //
                        Result = (Count1 > 0);

                        break;
                        }
                    }

                //
                //  We now compute the 'IdealLeaves' for each 'Twig'.
                //
                if ( Branch -> Used < IdealTwigs )
                    { IdealLeaves = ((TotalUsed + (Branch -> Used-1)) / Branch -> Used); }
                else
                    { IdealLeaves = MaxDensity; }
                }

            //
            //  We need an array to store all of the 'Leaves' in the 'Branch'.  We copy all of the
            //  'Leaves' into this array in sorted order to simplify copying them back later.
            //
            if
                    (
                    (Result)
                        &&
                    (ROCKALL_NEW<LEAF>::New( AutomaticHeapScope,NULL,& Leaves,TotalUsed,NULL,False,True ))
                    )
                {
                AUTO SBIT32 Count2;
                AUTO SBIT32 Count3;
                AUTO SBIT32 Count4;

                //
                //  We walk through all the 'Twigs' copying each 'Leaf' into our new array.
                //
                for ( Count1=0,Count3=0;Count1 < Branch -> Used;Count1 ++ )
                    {
                    AUTO TWIG *Current1 = Twigs[ Count1 ];
                    AUTO ORDER *Order2 = (ComputeOrderAddress( Current1 ));
                    AUTO ORDER (*View3)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order2);
                    AUTO LEAF (*View4)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Current1 -> Leaf);

                    //
                    //  We copy every 'Leaf' from a twig into the array in sorted order.
                    //
                    for ( Count2=0;Count2 < Current1 -> Used;Count2 ++ )
                        { Leaves[ (Count3 ++) ] = Current1 -> Leaf[ Order2[ Count2 ].IndexToValue ]; }
                    }

                //
                //  We have a great opportunity to verify that 'Key' values are in sorted order.
                //
                for ( Count1=1;Count1 < Count3;Count1 ++ )
                    {
                    if ( GreaterOrEqual( & Leaves[ (Count1-1) ].Key,& Leaves[ Count1 ].Key ) )
                        { FAILURE( "Leaves not in sorted order" ); }
                    }

                //
                //  We now copy each 'Leaf' in our array back into a suitable twig so as to balance
                //  and redistribute all the 'Leaves' amoung the 'Twigs'.
                //
                for ( Count1=0,Count4=0;Count1 < Branch -> Used;Count1 ++ )
                    {
                    AUTO TWIG *Current1 = Twigs[ Count1 ];
                    AUTO ORDER *Order2 = (ComputeOrderAddress( Current1 ));
                    AUTO ORDER (*View3)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order2);
                    AUTO LEAF (*View4)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Current1 -> Leaf);

                    //
                    //  We copy the 'IdealLeaves' into each twig.  If we run out of 'Leaves'
                    //  we will leave the final 'Twigs' empty (or at least less full).
                    //
                    for ( Count2=0;(Count2 < IdealLeaves) && (Count3 > Count4);Count2 ++ )
                        {
                        AUTO ORDER *Current2 = & Order2[ Count2 ];

                        Current1 -> Leaf[ Count2 ] = Leaves[ (Count4 ++) ];

                        Current2 -> IndexToValue = Count2;
                        Current2 -> ValueToOrder = Count2;
                        }

                    //
                    //  We zero any unused 'Leaf' just to be tidy.
                    //
                    if ( Count2 < Current1 -> Used )
                        {
                        AUTO SBIT16 Spare = (Current1 -> Used - Count2);

                        MemoryFunctions.Zero( & Order2[ Count2 ],(Spare * sizeof(ORDER)) );
                        ZeroLeaves( & Current1 -> Leaf[ Count2 ],Spare );
                        }

                    //
                    //  Finally, we update the number of 'Leaves' in the twig.
                    //
                    Current1 -> Used = Count2;
                    }

                //
                //  When we move the 'Leaves' between 'Twigs' we invalidate the 'Key' in the
                //  related 'Stem'.  We repair this damage by updating the 'Key' in the 'Stem'
                //  with the first 'Key' (i.e. the lowest) in the twig.  If the twig is empty
                //  we delete it to recover any associated space.
                //
                for ( Count1=(Branch -> Used-1);Count1 >= 0;Count1 -- )
                    {
                    AUTO STEM *Current1 = & Branch -> Stem[ Count1 ];
                    AUTO TWIG *Current2 = Twigs[ Count1 ];
                    AUTO ORDER *Current3 = & Order1[ Count1 ];

                    //
                    //  We update the 'Stem' if it is not empty.  If not, we delete it.
                    //
                    if ( Current2 -> Used > 0 )
                        {
                        //
                        //  We update the 'Stem' as the 'Twigs' have been adjusted.
                        //
                        Current1 -> Key = Current2 -> Leaf[0].Key;
                        Current1 -> Address = Current2 -> Address;

                        Current3 -> IndexToValue = Count1;
                        Current3 -> ValueToOrder = Count1;
                        }
                    else
                        {
                        //
                        //  We no longer need the 'Stem' so delete it.
                        //
                        DeleteTwig( AutomaticHeapScope,Current2 );

                        MemoryFunctions.Zero( Current3,sizeof(ORDER) );
                        ZeroStems( Current1,1 );

                        Branch -> Used --;
                        }
                    }

                //
                //  We will add a whole extra level into the tree if the 'Density' of the 'Branch'
                //  rises about the associated threashold.  We do not let the 'Branch' become full
                //  because this will lead to repeated rebalancing as we near its capacity because
                //  we will keep trying to squeeze more and more into the leaves.  Instead, we give
                //  up early so as to leave plenty of space.  We will only do this at the top level,
                //  so as to ensure that the tree always remains balanced.  We know that each level
                //  of the tree will balance itself when it becomes full, so it can never be necessary
                //  to add a new level anywhere else in the tree.
                //
                if ( (Root != NULL) && (Branch -> Used > Branch -> MaxDensity) )
                    {
                    AUTO BRANCH *NewLevel;

                    //
                    //  Create a new branch above the current branch.
                    //
                    if ( NewBranch( & Root -> Trunk,AutomaticHeapScope,& NewLevel,(Root -> MaxDepth+1) ) )
                        {
                        //
                        //  Create a new stem in the new branch refering to the current branch.
                        //
                        if ( NewStem( Branch -> Address,NewLevel,& Branch -> Stem[0].Key ) )
                            {
                            //
                            //  Update the 'MaxDepth' of the tree.
                            //
                            Root -> MaxDepth ++;

                            //
                            //  We rebalance the new branch to make sure we don't immediately
                            //  delete it (as it only has one stem).
                            //
                            Result = (RebalanceStems( NewLevel,Navigation,NULL ));
                            }
                        else
                            { FAILURE( "Unable to add a stem to new branch" ); }
                        }
                    else
                        { Result = False; }
                    }

                //
                //  We will delete the current 'Branch' if only has one 'Stem' and we are currently
                //  at the top level of the tree.  Clearly, any 'Branch' that does not have more than
                //  one 'Stem' is pointless and so there is no reason to keep it.
                //
                if ( (Root != NULL) && (Branch -> Used == 1) )
                    {
                    //
                    //  Update the 'Root' and delete the 'Branch'.
                    //
                    Root -> MaxDepth --;

                    Root -> Trunk = Branch -> Stem[0].Address;

                    DeleteBranch( AutomaticHeapScope,Branch );
                    }

                //
                //  We can now delete the array be created earlier but we must be sure to
                //  preserve any 'ErrorCode'.
                //
                if
                        (
                        (AutomaticHeapScope -> GetError( & ErrorCode ), False)
                            ||
                        (! ROCKALL_DELETE<LEAF>::Delete( AutomaticHeapScope,Leaves,TotalUsed ))
                            ||
                        (AutomaticHeapScope -> SetError( ErrorCode ), False)
                        )
                    { FAILURE( "Unable to delete leaves" ); }
                }
            else
                { Result = False; }

            //
            //  We can now delete the array be created earlier but we must be sure to
            //  preserve any 'ErrorCode'.
            //
            if
                    (
                    (AutomaticHeapScope -> GetError( & ErrorCode ), False)
                        ||
                    (! ROCKALL_DELETE<TWIG*>::Delete( AutomaticHeapScope,Twigs,Branch -> MaxSize ))
                        ||
                    (AutomaticHeapScope -> SetError( ErrorCode ), False)
                    )
                { FAILURE( "Unable to delete twigs" ); }
            }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Release Parent.                                                                           */
            /*                                                                                            */
            /*  Release any locks or views we have on our parent.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::ReleaseParent
            (
            NAVIGATION                          *Navigation,
            BOOLEAN                             Result
            )
        {
        //
        //  We will release any locks or views we currently hold on the parent.
        //
        if ( Navigation -> ReleaseParent )
            {
            //
            //  Reset 'ReleaseParent' to show that we no longer have any parent
            //  locks or views.
            //
            Navigation -> ReleaseParent = False;

            //
            //  Release the current parent lock.
            //
            Navigation -> AutomaticLockScope -> Release();

            //
            //  Release the current parent 'View()'.
            //
            Navigation -> AutomaticViewScope -> Release();
            }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Release Self.                                                                             */
            /*                                                                                            */
            /*  Release any locks or views we have on the current place in the tree.                      */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::ReleaseSelf
            (
            AUTOMATIC_LOCK_SCOPE<LOCK>          *AutomaticLockScope,
            AUTOMATIC_VIEW_SCOPE<HEAP>          *AutomaticViewScope,
            BOOLEAN                             Result
            )
        {
        //
        //  We will release any lock passed to the call.
        //
        if ( AutomaticLockScope != NULL )
            { AutomaticLockScope -> Release(); }

        //
        //  We will release any 'View()' passed to the call.
        //
        if ( AutomaticViewScope != NULL )
            { AutomaticViewScope -> Release(); }

        return Result;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Compute the size of the tree.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::Size
            (
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Size( & AutomaticHeapScope,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size.                                                                                     */
            /*                                                                                            */
            /*  Compute the size of the tree.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::Size
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the tree.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticShareableLock2,
                    & AutomaticViewScope,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    NULL,
                    False,
                    False,
                    Address,
                    & Root -> Trunk,
                    False,
                    Root,
                    NULL,
                    Root -> Type,
                    NULL
                    };

                //
                //  We zero the 'Value' and walk the tree.
                //
                return (((*Value) = 0), (SizeNavigation( & NextLevel,Value )));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Size Navigation.                                                                          */
            /*                                                                                            */
            /*  Compute the size of the tree.                                                             */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::SizeNavigation
            (
            NAVIGATION                          *Navigation,
            FILE_SIZE                           *Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Count;

                        //
                        //  We 'Walk()' every 'Stem' in the 'Branch' in sorted order.
                        //
                        for ( Count=0;Count < Branch -> Used;Count ++ )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticShareableLock,
                                & AutomaticViewScope,
                                NULL,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                NULL,
                                False,
                                False,
                                Address,
                                & Branch -> Stem[ Order[ Count ].IndexToValue ].Address,
                                False,
                                Navigation -> Root,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                NULL
                                };

                            if ( ! SizeNavigation( & NextLevel,Value ) )
                                { return False; }
                            }

                        return True;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        //
                        //  We compute the size of the 'Twig' and add it to the 'Value'.
                        //
                        (*Value) += Twig -> Used;

                        return True;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return True; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Split Twig.                                                                               */
            /*                                                                                            */
            /*  Split a twig when it becomes too full.                                                    */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::SplitTwig
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            BRANCH                              *Branch,
            TWIG                                *Twig
            )
        {
        //
        //  We will not be able to split the 'Twig' if the 'Branch' is full.
        //
        if ( Branch -> Used < Branch -> MaxSize )
            {
            STATIC CONST SBIT32 MaxTwigs = 2;
            STATIC CONST SBIT32 SplitTwig = (MaxTwigs-1);

            //
            //  We do not need to do anything unless the 'Twig' is full.
            //
            if ( Twig -> Used == Twig -> MaxSize )
                {
                AUTO FILE_ADDRESS Address[ MaxTwigs ] = { Twig -> Address,NULL };
                AUTO TWIG *Twigs[ MaxTwigs ] = { Twig,NULL };

                //
                //  Create a new empty 'Twig' and attach it to the 'Branch'.
                //
                if ( NewTwig( & Address[ SplitTwig ],AutomaticHeapScope,& Twigs[ SplitTwig ] ) )
                    {
                    AUTO SBIT32 NewLeaves = (Twigs[0] -> Used);
                    AUTO LEAF *Leaves;

                    //
                    //  We need an array to store all of the 'Leaves' in the 'Twig'.  We copy
                    //  all of the 'Leaves' into this array in sorted order to simplify copying
                    //  them back later.
                    //
                    if ( ROCKALL_NEW<LEAF>::New( AutomaticHeapScope,NULL,& Leaves,NewLeaves,NULL,False,True ) )
                        {
                        AUTO SBIT32 IdealLeaves = ((Twig -> Used + 1) / 2);
                        AUTO SBIT32 Count1;
                        AUTO SBIT32 Count2;
                        AUTO SBIT32 Count3;
                        AUTO SBIT32 Count4;

                        //
                        //  We copy every 'Leaf' from the 'Twig' into the array in sorted order.
                        //
                        for ( Count1=0,Count3=0;Count1 < MaxTwigs;Count1 ++ )
                            {
                            AUTO TWIG *Current1 = Twigs[ Count1 ];
                            AUTO ORDER *Order = (ComputeOrderAddress( Current1 ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Current1 -> Leaf);

                            //
                            //  We copy every 'Leaf' from a twig into the array in sorted order.
                            //
                            for ( Count2=0;Count2 < Current1 -> Used;Count2 ++ )
                                { Leaves[ (Count3 ++) ] = Current1 -> Leaf[ Order[ Count2 ].IndexToValue ]; }
                            }

                        //
                        //  We have a great opportunity to verify that 'Key' values are in
                        //  sorted order.
                        //
                        for ( Count1=1;Count1 < Count3;Count1 ++ )
                            {
                            if ( GreaterOrEqual( & Leaves[ (Count1-1) ].Key,& Leaves[ Count1 ].Key ) )
                                { FAILURE( "Leaves not in sorted order" ); }
                            }

                        //
                        //  We now copy each 'Leaf' in our array back into a suitable 'Twig' so
                        //  as to balance and redistribute all the 'Leaves' amoung the 'Twigs'.
                        //
                        for ( Count1=0,Count4=0;Count1 < 2;Count1 ++ )
                            {
                            AUTO TWIG *Current1 = Twigs[ Count1 ];
                            AUTO ORDER *Order = (ComputeOrderAddress( Current1 ));
                            AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                            AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Current1 -> Leaf);

                            //
                            //  We copy the 'IdealLeaves' into each twig.  If we run out of
                            //  'Leaves' we will leave the final 'Twigs' empty (or at least
                            //  less full).
                            //
                            for ( Count2=0;(Count2 < IdealLeaves) && (Count3 > Count4);Count2 ++ )
                                {
                                AUTO ORDER *Current2 = & Order[ Count2 ];

                                Current1 -> Leaf[ Count2 ] = Leaves[ (Count4 ++) ];

                                Current2 -> IndexToValue = Count2;
                                Current2 -> ValueToOrder = Count2;
                                }

                            //
                            //  We zero any unused 'Leaf' just to be tidy.
                            //
                            if ( Count2 < Current1 -> Used )
                                {
                                AUTO SBIT16 Spare = (Current1 -> Used - Count2);

                                MemoryFunctions.Zero( & Order[ Count2 ],(Spare * sizeof(ORDER)) );
                                ZeroLeaves( & Current1 -> Leaf[ Count2 ],Spare );
                                }

                            //
                            //  Finally, we update the number of 'Leaves' in the twig.
                            //
                            Current1 -> Used = Count2;
                            }

                        //
                        //  We have now split the 'Twig' so all that is left is to put the
                        //  new 'Twig' in the 'Branch'.  The first 'Twig' should already be
                        //  present so we just need to add the new 'Twig'.
                        //
                        if ( ! NewStem( Address[ SplitTwig ],Branch,& Twigs[ SplitTwig ] -> Leaf[0].Key ) )
                            { FAILURE( "Unable to put twig into a branch" ); }

                        //
                        //  We can now delete the array be created earlier.
                        //
                        if ( ! ROCKALL_DELETE<LEAF>::Delete( AutomaticHeapScope,Leaves,NewLeaves ) )
                            { FAILURE( "Unable to delete leaves" ); }

                        return True;
                        }
                    }
                }
            else
                {
                //
                //  We do not need to split the 'Twig' if it is not full.
                //
                if ( Twig -> Used < Twig -> MaxSize )
                    { return True; }
                else
                    { FAILURE( "Unable to split twig" ); }
                }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update Key.                                                                               */
            /*                                                                                            */
            /*  Update a key value in the tree.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::UpdateKey
            (
            CONST KEY                           & Key,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (UpdateKey( & AutomaticHeapScope,Key,Value ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update Key.                                                                               */
            /*                                                                                            */
            /*  Update a key value in the tree.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::UpdateKey
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            CONST KEY                           & Key,
            CONST VALUE                         & Value
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from changing the tree.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticShareableLock2,
                    & AutomaticViewScope,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    ((KEY*) & Key),
                    True,
                    False,
                    Address,
                    & Root -> Trunk,
                    True,
                    Root,
                    NULL,
                    Root -> Type,
                    ((VALUE*) & Value)
                    };

                //
                //  Find a 'Key' and the associated 'Value' in the tree.
                //
                if  ( UpdateKeyNavigation( & NextLevel ) )
                    { return True; }
                else
                    { return (AutomaticHeapScope -> SetError( ErrorKeyNotFound )); }
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Update Key Navigation.                                                                    */
            /*                                                                                            */
            /*  Update a key value in the tree.                                                           */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::UpdateKeyNavigation
            (
            NAVIGATION                          *Navigation
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Offset;

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  When we have a 'Branch' we find closest 'Stem' which is less than
                        //  the 'Key'.
                        //
                        if ( FindStem( Branch,Navigation -> Key,True,& Offset ) )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticShareableLock,
                                & AutomaticViewScope,
                                Branch,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                Navigation -> Key,
                                False,
                                False,
                                Address,
                                & Branch -> Stem[ Offset ].Address,
                                True,
                                NULL,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                Navigation -> Value
                                };

                            //
                            //  Search the next level down in the tree.
                            //
                            return (UpdateKeyNavigation( & NextLevel ));
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.ExclusiveView( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                        AUTO SBIT16 Offset;

                        //
                        //  We only need to hold a 'View()' and a lock on our 'Parent' until we
                        //  get a 'View()' and a lock on the next level down in the tree.  We can
                        //  now safely release the 'View()' and the lock on our 'Parent' as we now
                        //  hold a 'View()' and a lock on the next level down.
                        //
                        ReleaseParent( Navigation,True );

                        //
                        //  When we finally reach a 'Twig' we find the 'Leaf' which contains
                        //  the supplied 'Key' and return the associated 'Value'.
                        //
                        if ( FindLeaf( Navigation -> Key,& Offset,False,Twig ) )
                            {
                            //
                            //  We will need to 'Update()' the 'Twig'.
                            //
                            if ( AutomaticHeapScope -> Update( Address,((VOID**) & Twig) ) )
                                {
                                Twig -> Leaf[ Offset ].Value = (*(Navigation -> Value));

                                return True;
                                }
                            }
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return False; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk every node in the tree.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::Walk
            (
            TREE_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;

        return (Walk( & AutomaticHeapScope,WalkFunction,UserValue ));
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk.                                                                                     */
            /*                                                                                            */
            /*  Walk every node in the tree.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::Walk
            (
            AUTOMATIC_HEAP_SCOPE<HEAP>          *AutomaticHeapScope,
            TREE_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( AutomaticHeapScope );
        AUTO TRANSACTIONAL_LOCK TransactionLock( AutomaticHeapScope );

        //
        //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
        //  we have a transaction.  The function calls to 'BeginTransaction()' and 'View()'
        //  should be automatically optimized away for heaps that do not support this
        //  functionality by any decent compiler, due to the way the code has been
        //  written.
        //
        if
                (
                (Active)
                    &&
                (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                )
            {
            AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticShareableLock1( & TransactionLock );
            AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
            AUTO ROOT *Root;

            //
            //  We 'View()' the 'Root' as we need to examine it and check the 'Signature'
            //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
            //  above to prevent threads in the same transaction from accessing the tree.
            //
            if
                    (
                    (AutomaticViewScope.View( Address,((VOID**) & Root) ))
                        &&
                    (Root -> Signature == Signature)
                    )
                {
                AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock2( & Root -> ClassLock );
                AUTO NAVIGATION NextLevel =
                    {
                    AutomaticHeapScope,
                    & AutomaticShareableLock2,
                    & AutomaticViewScope,
                    NULL,
                    0,
                    Root -> MaxDepth,
                    NULL,
                    False,
                    False,
                    Address,
                    & Root -> Trunk,
                    False,
                    Root,
                    NULL,
                    Root -> Type,
                    NULL
                    };

                return (WalkNavigation( & NextLevel,WalkFunction,UserValue ));
                }
            else
                { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }
            }
        else
            { return (AutomaticHeapScope -> SetError( ErrorUnableToViewRoot )); }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Walk Navigation.                                                                          */
            /*                                                                                            */
            /*  Walk every node in the tree.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        BOOLEAN TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::WalkNavigation
            (
            NAVIGATION                          *Navigation,
            TREE_WALK_FUNCTION                  WalkFunction,
            VOID                                *UserValue
            )
        {
        AUTO AUTOMATIC_HEAP_SCOPE<HEAP> *AutomaticHeapScope = (Navigation -> AutomaticHeapScope);
        AUTO FILE_ADDRESS Address = (*(Navigation -> Reference));

        //
        //  An empty tree has a 'Trunk' value of 'NULL' so we can exit immediately if this is
        //  the case.  This should not occur in any other situation as any 'NULL' reference
        //  should be automatically deleted.
        //
        if ( (Address != NULL) && (Navigation -> Depth > 0) )
            {
            //
            //  We know the 'Type' of tree node to expect based on the 'Depth'.  This is very
            //  helpful and allows us to claim the correct type of access and locks.  If the
            //  'Depth' is greater than one it must be a 'Branch', equal to one it must be
            //  a 'Twig' and zero the tree must be empty.
            //
            if ( Navigation -> Depth > 1 )
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO BRANCH *Branch;

                //
                //  The 'Type' of tree node should be a 'Branch' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Branch) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Branch -> BranchLock );

                    //
                    //  Let's ensure that we have found the correct 'Address', 'Depth' and
                    //  'Type' of node in the tree.
                    //
                    if
                            (
                            (Branch -> Address == Address)
                                &&
                            (Branch -> Depth == Navigation -> Depth)
                                &&
                            (Branch -> Type == TreeBranch)
                                &&
                            (Branch -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Branch ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO STEM (*View2)[ ViewSize ] = ((STEM(*)[ ViewSize ]) Branch -> Stem);
                        AUTO SBIT16 Count;

                        //
                        //  We 'Walk()' every 'Stem' in the 'Branch' in sorted order.
                        //
                        for ( Count=0;Count < Branch -> Used;Count ++ )
                            {
                            AUTO NAVIGATION NextLevel =
                                {
                                Navigation -> AutomaticHeapScope,
                                & AutomaticShareableLock,
                                & AutomaticViewScope,
                                NULL,
                                (Navigation -> Calls+1),
                                (Navigation -> Depth-1),
                                NULL,
                                False,
                                False,
                                Address,
                                & Branch -> Stem[ Order[ Count ].IndexToValue ].Address,
                                False,
                                Navigation -> Root,
                                Navigation -> SharedNavigation,
                                Branch -> Type,
                                NULL
                                };

                            if ( ! WalkNavigation( & NextLevel,WalkFunction,UserValue ) )
                                { return False; }
                            }

                        return True;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            else
                {
                AUTO AUTOMATIC_VIEW_SCOPE<HEAP> AutomaticViewScope( AutomaticHeapScope );
                AUTO TWIG *Twig;

                //
                //  The 'Type' of tree node should be a 'Twig' so claim the correct type
                //  of access.
                //
                if ( AutomaticViewScope.View( Address,((VOID**) & Twig) ) )
                    {
                    AUTO AUTOMATIC_SHAREABLE_LOCK_SCOPE<LOCK> AutomaticShareableLock( & Twig -> TwigLock );

                    //
                    //  Let's ensure that we have found the correct 'Address' and 'Type'
                    //  of node in the tree.
                    //
                    if
                            (
                            (Twig -> Address == Address)
                                &&
                            (Twig -> Type == TreeTwig)
                                &&
                            (Twig -> Used > 0)
                            )
                        {
                        AUTO ORDER *Order = (ComputeOrderAddress( Twig ));
                        AUTO ORDER (*View1)[ ViewSize ] = ((ORDER(*)[ ViewSize ]) Order);
                        AUTO LEAF (*View2)[ ViewSize ] = ((LEAF(*)[ ViewSize ]) Twig -> Leaf);
                        AUTO SBIT16 Count;

                        //
                        //  We 'Walk()' every 'Leaf' in the 'Twig' in sorted order.
                        //
                        for ( Count=0;Count < Twig -> Used;Count ++ )
                            {
                            AUTO LEAF *Current = & Twig -> Leaf[ Order[ Count ].IndexToValue ];

                            //
                            //  We simply don't trust user supplied code so we prepare for trouble.
                            //
#ifndef DISABLE_CRT_FEATURES
                            try
                                {
#endif
                                //
                                //  Execute the user supplied walk function.
                                //
                                if ( ! WalkFunction( Current -> Key,Current -> Value,UserValue ) )
                                    { return False; }
#ifndef DISABLE_CRT_FEATURES
                                }
                            catch ( ... )
                                {
                                //
                                //  We will catch and skip all errors detected in the user supplied
                                //  walk function.
                                //
                                return (AutomaticHeapScope -> SetError( ErrorExceptionCaught ));
                                }
#endif
                            }

                        return True;
                        }
                    else
                        { FAILURE( "Malformed tree" ); }
                    }
                }
            }
        else
            {
            //
            //  Let's ensure that we have found the correct 'Address', 'Depth' and 'Type'
            //  of node in the tree.
            //
            if ( (Address == NULL) && (Navigation -> Depth == 0) && (Navigation -> Type == TreeRoot)  )
                { return True; }
            else
                { FAILURE( "Malformed tree" ); }
            }

        return False;
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Leaves.                                                                              */
            /*                                                                                            */
            /*  Zero a collection of leaves.                                                              */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::ZeroLeaves
            (
            LEAF                                *Leaves,
            SBIT32                              Size
            )
        {
        AUTO SBIT32 Count;

        //
        //  We have to be careful as a 'LEAF' and a 'STEM' contain user types.  It is very hard
        //  to zero these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Leaves[ Count ],LEAF ); }

        MemoryFunctions.Zero( Leaves,(Size * sizeof(LEAF)) );

        for ( Count=0;Count < Size;Count ++ )
            { PLACEMENT_NEW( & Leaves[ Count ],LEAF ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Zero Stems.                                                                               */
            /*                                                                                            */
            /*  Zero a collection of stems.                                                               */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
        VOID TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::ZeroStems
            (
            STEM                                *Stems,
            SBIT32                              Size
            )
        {
        AUTO SBIT32 Count;

        //
        //  We have to be careful as a 'LEAF' and a 'STEM' contain user types.  It is very hard
        //  to zero these as we would not wish to damage them.  So just to be safe we call the
        //  destructor, then we can zero them and then finally we call the constructor.
        //
        for ( Count=(Size-1);Count >= 0;Count -- )
            { PLACEMENT_DELETE( & Stems[ Count ],STEM ); }

        MemoryFunctions.Zero( Stems,(Size * sizeof(STEM)) );

        for ( Count=0;Count < Size;Count ++ )
            { PLACEMENT_NEW( & Stems[ Count ],STEM ); }
        }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Destructor for 'TREE'.                                                                    */
            /*                                                                                            */
            /*  The destructor for 'TREE' cleans up the class.                                            */
            /*                                                                                            */
            /**********************************************************************************************/

    template <class HEAP,class LOCK,class KEY,class VALUE,INT LEAVES,INT STEMS>
            TREE<HEAP,LOCK,KEY,VALUE,LEAVES,STEMS>::~TREE( VOID )
        {
        //
        //  We need to ensure that the class is 'Active'.
        //
        if ( Active )
            {
            AUTO AUTOMATIC_HEAP_SCOPE<HEAP> AutomaticHeapScope;
            AUTO AUTOMATIC_TRANSACTION_SCOPE<HEAP> AutomaticTransactionScope( & AutomaticHeapScope );
            AUTO TRANSACTIONAL_LOCK TransactionLock( & AutomaticHeapScope );

            //
            //  We ensure that we are 'Active' and then 'BeginTransaction()' to ensure that
            //  we have a transaction.  The function calls to 'BeginTransaction()' and 'Update()'
            //  should be automatically optimized away for heaps that do not support this
            //  functionality by any decent compiler, due to the way the code has been written.
            //
            if
                    (
                    (Active)
                        &&
                    (AutomaticTransactionScope.BeginTransaction( NULL,& TransactionLock ))
                    )
                {
                AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<TRANSACTIONAL_LOCK> AutomaticExclusiveLock1( & TransactionLock );
                AUTO ROOT *Root;

                //
                //  We 'Update()' the 'Root' as we need to modify it and check the 'Signature'
                //  to ensure the 'Root' is of the correct type.  We claimed the 'TransactionLock'
                //  above to prevent threads in the same transaction from accessing the tree.
                //
                if
                        (
                        (AutomaticHeapScope.Update( Address,((VOID**) & Root) ))
                            &&
                        (Root -> Signature == Signature)
                        )
                    {
                    AUTO AUTOMATIC_EXCLUSIVE_LOCK_SCOPE<LOCK> AutomaticExclusiveLock2( & Root -> ClassLock );
                    AUTO NAVIGATION NextLevel =
                        {
                        & AutomaticHeapScope,
                        & AutomaticExclusiveLock2,
                        NULL,
                        NULL,
                        0,
                        Root -> MaxDepth,
                        NULL,
                        True,
                        False,
                        Address,
                        & Root -> Trunk,
                        False,
                        Root,
                        NULL,
                        Root -> Type,
                        NULL
                        };

                    //
                    //  Delete the entire tree.
                    //
                    if ( DeleteAllNavigation( & NextLevel ) )
                        {
                        //
                        //  Zero the 'Root' structure.
                        //
                        Root -> Trunk = NULL;
                        Root -> Signature = NULL;
                        Root -> MaxDepth = 0;
                        Root -> Type = ((TREE_TYPE) NULL);
                        Root -> Address = NULL;

                        //
                        //  We are no longer 'Active'.
                        //
                        Active = False;

                        //
                        //  Release the lock on the 'Root'.
                        //
                        AutomaticExclusiveLock2.Release();

                        //
                        //  Delete the 'Root' structure.
                        //
                        if ( ROCKALL_DELETE<ROOT>::Delete( & AutomaticHeapScope,Address ) )
                            {
                            //
                            //  Zero class members.
                            //
                            Address = NULL;
                            Active = False;
                            }
                        else
                            { FAILURE( "Unable to delete the root" ); }
                        }
                    else
                        { FAILURE( "Unable to delete the root structure" ); }
                    }
                else
                    { FAILURE( "Unable to update the root structure" ); }
                }
            else
                { FAILURE( "Unable to update the root structure" ); }
            }
        }
    }

            /**********************************************************************************************/
            /*                                                                                            */
            /*  Rockall Global Types.                                                                     */
            /*                                                                                            */
            /*  The rockall global types allow the 'ROCKALL' namespace to be optionally omitted in        */
            /*  most situations but can be disabled if necessary.                                         */
            /*                                                                                            */
            /**********************************************************************************************/

#ifndef DISABLE_ROCKALL_GLOBAL_TYPES
#define TREE ROCKALL::TREE
#endif
#endif